use crate::descriptor_config::{SourceDescriptor, SourceDescriptorInner, VirtualSourceKind};
use crate::pointer_names::{get_pointer_name_at, POINTER_NAME_DATETIME_SPLITTER};
use bytesize::ByteSize;
use chrono::{DateTime, Utc};
use clap::Args;
use dashmap::DashSet;
use eyre::{bail, ensure, eyre, Context, ContextCompat};
use indicatif::ProgressStyle;
use patricia_tree::PatriciaMap;
use std::cmp::max;
use std::collections::{BTreeMap, BTreeSet, HashMap};
use std::io::Write;
use std::path::PathBuf;
use std::process::{Child, Command, Stdio};
use std::sync::Arc;
use std::time::{Instant, SystemTime, UNIX_EPOCH};
use tokio::runtime::Handle;
use tokio::task::JoinSet;
use tracing::{debug, info, info_span, Instrument, Span};
use tracing_indicatif::span_ext::IndicatifSpanExt;
use users::{get_current_gid, get_current_uid};
use yama::pile_with_cache::{PileWithCache, PointerIntegrationStatistics};
use yama::scan::{create_uidgid_lookup_tables, limit_scan_entry_map_to_size};
use yama::storing::{
    assemble_and_write_indices, StoragePipeline, StoringBloblogWriters, StoringState,
};
use yama::{scan, PROGRESS_BAR_STYLE};
use yama_midlevel_crypto::chunk_id::ChunkId;
use yama_pile::definitions::{BlobLocator, BloblogId, IndexBloblogEntry, RecursiveChunkRef};
use yama_pile::pointers::Pointer;
use yama_pile::tree::unpopulated::ScanEntry;
use yama_pile::tree::{
    assemble_tree_from_scan_entries, differentiate_node_in_place, FilesystemOwnership,
    FilesystemPermissions, RootTreeNode, TreeNode,
};
use yama_wormfile::boxed::BoxedWormFileProvider;

#[derive(Args, Clone, Debug)]
pub struct BackupOptions {
    /// Number of bytes to back up in one go. Intended for gradually getting a backup started.
    /// Supports suffixes like MiB and MB.
    /// Applies per-source. Does not apply to virtual sources.
    #[clap(long)]
    gradual: Option<ByteSize>,
}

pub async fn backup(
    pwc: Arc<PileWithCache<BoxedWormFileProvider>>,
    sources_to_backup: BTreeMap<String, SourceDescriptor>,
    options: &BackupOptions,
) -> eyre::Result<()> {
    // Locate suitable parent pointers
    let timestart = Instant::now();
    let parents_to_use = find_suitable_parent_pointers(&pwc, &sources_to_backup)
        .await
        .context("failed to look for suitable parent pointers")?;
    debug!(
        "find_suitable_parent_pointers: {:?}",
        Instant::now() - timestart
    );
    let now = Utc::now();

    // (dirtrees) Scan
    let timestart = Instant::now();
    let dir_sources = scan_dir_sources(&sources_to_backup, parents_to_use, now, options)
        .await
        .context("failed to scan directory sources")?;
    debug!("scan_dir_sources: {:?}", Instant::now() - timestart);

    let new_unflushed_chunks: Arc<DashSet<ChunkId>> = Arc::new(Default::default());

    // (dirtrees) Start a storage pipeline and submit jobs to it
    let task_store_dirs = {
        let new_unflushed_chunks = new_unflushed_chunks.clone();
        let pwc = pwc.clone();

        let bds_span = info_span!("storing");
        tokio::spawn(async_backtrace::frame!(async move {
            backup_dir_sources(dir_sources, pwc, new_unflushed_chunks)
                .await
                .context("whilst backing up dir sources")
        }
        .instrument(bds_span)))
    };

    // (virtual source streams) Store to bloblog writers
    let task_store_virtuals = {
        let bvs_span = info_span!("storing_virts");
        let new_unflushed_chunks = new_unflushed_chunks.clone();
        let pwc = pwc.clone();

        tokio::spawn(async_backtrace::frame!(async move {
            backup_virtual_sources(&sources_to_backup, now, pwc, new_unflushed_chunks)
                .await
                .context("whilst backing up virtual sources")
        }
        .instrument(bvs_span)))
    };

    let timestart = Instant::now();
    let (dir_sources_and_chunkmaps, virt_sources) =
        tokio::join!(task_store_dirs, task_store_virtuals);
    debug!(
        "join!(task_store_dirs, task_store_virtuals): {:?}",
        Instant::now() - timestart
    );
    let dir_sources_and_chunkmaps: BackupDirSourcesReturn = dir_sources_and_chunkmaps??;
    let mut virt_sources: Vec<VirtualSourceReturn> = virt_sources??;

    let mut chunkmaps = dir_sources_and_chunkmaps.chunkmaps;
    for source in &mut virt_sources {
        chunkmaps.extend(
            std::mem::take(&mut source.chunkmaps)
                .into_iter()
                .map(|(k, nb)| {
                    (
                        k,
                        IndexBloblogEntry {
                            chunks: nb,
                            forgotten_bytes: 0,
                        },
                    )
                }),
        );
    }

    // Chunkmaps, indices and write pointers
    assemble_and_write_indices(&pwc, chunkmaps)
        .await
        .context("failed to assemble and write indices")?;

    info!("All indices stored, writing pointer...");

    for (dir_source_prep, chunk_file_map) in dir_sources_and_chunkmaps.dir_source_returns {
        // Assemble and write a pointer
        let mut tree = assemble_tree_from_scan_entries(
            dir_source_prep.new_scan_entry_map,
            dir_source_prep.reused_scan_entry_map,
            chunk_file_map,
        )
        .context("failed to assemble tree")?;
        let (uids, gids) =
            create_uidgid_lookup_tables(&tree).context("failed to create uid/gid tables")?;

        if let Some(ref parent_node) = dir_source_prep.parent {
            differentiate_node_in_place(&mut tree, &parent_node.root.node)
                .context("failed to differentiate?")?;
        }

        pwc.pile
            .write_pointer(
                &dir_source_prep.new_pointer_name,
                false,
                &Pointer {
                    parent: dir_source_prep.parent_name.clone(),
                    root: RootTreeNode {
                        name: dir_source_prep
                            .path
                            .file_name()
                            .map(|oss| oss.to_str())
                            .flatten()
                            .unwrap_or("")
                            .to_owned(),
                        node: tree,
                    },
                    uids,
                    gids,
                },
            )
            .await
            .context("failed to write pointer")?;
    }

    for virtual_source in virt_sources {
        pwc.pile
            .write_pointer(&virtual_source.pointer_name, false, &virtual_source.pointer)
            .await
            .context("failed to write pointer")?;
    }

    Arc::try_unwrap(pwc)
        .map_err(|_| eyre!("pwc still in use; can't close down gracefully"))?
        .close()
        .await?;

    Ok(())
}

/// Given access to a PWC and a map of sources to back up, returns a map of pointer names to use as parents.
/// For virtual sources, no parent is chosen.
/// For directory sources, the most recent pointer from the same source is chosen as a parent.
async fn find_suitable_parent_pointers(
    pwc: &PileWithCache<BoxedWormFileProvider>,
    sources_to_backup: &BTreeMap<String, SourceDescriptor>,
) -> eyre::Result<BTreeMap<String, (String, Pointer)>> {
    let mut result = BTreeMap::new();

    let pointers = pwc
        .pile
        .list_pointers()
        .await
        .context("failed to list pointers")?;

    for (source_name, source) in sources_to_backup.iter() {
        if source.is_directory_source() {
            let starter = format!("{source_name}{POINTER_NAME_DATETIME_SPLITTER}");
            if let Some(most_recent_pointer) = pointers
                .iter()
                .rev()
                .filter(|pn| pn.starts_with(&starter))
                .next()
            {
                debug!("for {source_name:?}, using parent {most_recent_pointer:?}");
                let mut stats = PointerIntegrationStatistics::default();
                let pointer = pwc
                    .read_pointer_fully_integrated(&most_recent_pointer, &mut stats)
                    .await
                    .context("failed to read parent pointer")?
                    .context("no parent pointer despite having just listed it")?;
                debug!("when loading parent, stats = {stats:?}");
                result.insert(
                    source_name.to_owned(),
                    (most_recent_pointer.clone(), pointer),
                );
            }
        }
    }

    Ok(result)
}

struct DirSourcePrep {
    /// New entries only.
    new_scan_entry_map: PatriciaMap<ScanEntry>,
    /// Files: Reused entries only. Directories: can be partially changed but there's no chunking to be done.
    reused_scan_entry_map: PatriciaMap<ScanEntry>,
    parent_name: Option<String>,
    parent: Option<Pointer>,
    path: PathBuf,
    new_pointer_name: String,
    chunk_file_map: PatriciaMap<Option<(RecursiveChunkRef, u64)>>,
}

async fn scan_dir_sources(
    sources_to_backup: &BTreeMap<String, SourceDescriptor>,
    mut parents: BTreeMap<String, (String, Pointer)>,
    now: DateTime<Utc>,
    options: &BackupOptions,
) -> eyre::Result<Vec<DirSourcePrep>> {
    let mut joinset = JoinSet::new();

    for (source_name, source) in sources_to_backup {
        if let SourceDescriptorInner::DirectorySource {
            path,
            cross_filesystems,
            ignore,
        } = &source.inner
        {
            let path = path.to_owned();
            let ignore = ignore.to_owned();
            let (parent_name, parent) = parents.remove(source_name).unzip();
            let new_pointer_name = get_pointer_name_at(&source_name, now);
            let options = options.clone();
            joinset.spawn_blocking(move || -> eyre::Result<DirSourcePrep> {
                let timestart = Instant::now();
                let scan_entry_map = scan::scan(&path, &ignore).context("Failed to scan")?;
                debug!("scan: {:?}", Instant::now() - timestart);
                info!(
                    "size estimate for {:?} (full scan): {}",
                    path,
                    summarise_scan_entry_map_size(&scan_entry_map)
                );

                // TODO This whole section is messy.
                //      Maybe we should consider combining prepopulate_unmodified and limit_scan_entry_map_to_size
                //      as the latter might benefit from being able to see what is in the parent pointer...

                let (chunk_file_map, pruned_scan_entry_map, prepopulated_scan_entry_map) =
                    if let Some(ref parent_node) = parent {
                        let (cfm, pruned, prepopulated) =
                            scan::prepopulate_unmodified(&parent_node.root.node, &scan_entry_map);

                        // debug
                        let pruned_keys: BTreeSet<String> = pruned
                            .keys()
                            .map(|b| String::from_utf8(b).unwrap())
                            .collect();
                        let prepop_keys: BTreeSet<String> = prepopulated
                            .keys()
                            .map(|b| String::from_utf8(b).unwrap())
                            .collect();
                        let ix_keys: BTreeSet<&String> =
                            pruned_keys.intersection(&prepop_keys).collect();
                        if !ix_keys.is_empty() {
                            bail!("bug: intersecting prepop and prune keys: {ix_keys:?}");
                        }

                        info!(
                            "size estimate for {:?} (differential): {}",
                            path,
                            summarise_scan_entry_map_size(&pruned)
                        );

                        (cfm, pruned, prepopulated)
                    } else {
                        (
                            PatriciaMap::<Option<(RecursiveChunkRef, u64)>>::new(),
                            scan_entry_map,
                            PatriciaMap::new(),
                        )
                    };

                let pruned_scan_entry_map = match options.gradual {
                    Some(gradual_size_limit) => {
                        let limited = limit_scan_entry_map_to_size(
                            pruned_scan_entry_map,
                            gradual_size_limit.as_u64(),
                        );
                        info!(
                            "size estimate for {:?} (gradual/limited): {}",
                            path,
                            summarise_scan_entry_map_size(&limited)
                        );
                        limited
                    }
                    None => pruned_scan_entry_map,
                };

                Ok(DirSourcePrep {
                    chunk_file_map,
                    new_scan_entry_map: pruned_scan_entry_map,
                    reused_scan_entry_map: prepopulated_scan_entry_map,
                    parent_name,
                    parent,
                    path,
                    new_pointer_name,
                })
            });
        }
    }

    let mut result = Vec::new();

    while let Some(dsp_res_res) = joinset.join_next().await {
        result.push(dsp_res_res??);
    }
    Ok(result)
}

fn summarise_scan_entry_map_size(scan_entry_map: &PatriciaMap<ScanEntry>) -> String {
    let mut num_bytes = 0u64;

    for (_, entry) in scan_entry_map.iter() {
        num_bytes += match entry {
            ScanEntry::NormalFile { size, .. } => max(*size, 4096),
            _ => 4096,
        };
    }

    let num_files = scan_entry_map.len();

    format!(
        "{num_files} files ({})",
        ByteSize(num_bytes).to_string_as(true)
    )
}

struct BackupDirSourcesReturn {
    pub chunkmaps: BTreeMap<BloblogId, IndexBloblogEntry>,
    pub dir_source_returns: Vec<(DirSourcePrep, PatriciaMap<Option<(RecursiveChunkRef, u64)>>)>,
}

#[async_backtrace::framed]
async fn backup_dir_sources(
    mut dir_sources: Vec<DirSourcePrep>,
    pwc: Arc<PileWithCache<BoxedWormFileProvider>>,
    new_unflushed_chunks: Arc<DashSet<ChunkId>>,
) -> eyre::Result<BackupDirSourcesReturn> {
    let mut chunk_file_maps = Vec::new();
    let mut pruned_scan_entry_maps = Vec::new();

    // First collect all that stuff together...
    for dir_source in &mut dir_sources {
        chunk_file_maps.push(std::mem::take(&mut dir_source.chunk_file_map));
    }
    for dir_source in &dir_sources {
        pruned_scan_entry_maps.push(&dir_source.new_scan_entry_map);
    }

    let store_span = Span::current();
    // store_span.pb_set_style(&ProgressStyle::default_bar());
    store_span.pb_set_style(
        &ProgressStyle::default_bar()
            .template(PROGRESS_BAR_STYLE)
            .unwrap(),
    );
    store_span.pb_set_message("storing files");
    store_span.pb_set_length(
        pruned_scan_entry_maps
            .iter()
            .map(|pruned_scan_entry_map| {
                pruned_scan_entry_map
                    .values()
                    .filter(|v| matches!(v, ScanEntry::NormalFile { .. }))
                    .count() as u64
            })
            .sum(),
    );

    //
    let (pipeline, pipeline_job_tx) =
        StoragePipeline::launch_new(4, pwc.clone(), new_unflushed_chunks).await?;

    let dir_sources2 = &dir_sources;
    let mut submitted = 0;
    let mut completed = 0;
    let submitted_mut = &mut submitted;
    let (submitter_task, receiver_task) = tokio::join!(
        async move {
            let pipeline_job_tx = pipeline_job_tx;
            for (dir_source_idx, dir_source) in dir_sources2.iter().enumerate() {
                for (name_bytes, scan_entry) in pruned_scan_entry_maps[dir_source_idx].iter() {
                    if let ScanEntry::NormalFile { .. } = scan_entry {
                        let name = std::str::from_utf8(name_bytes.as_slice())
                            .context("name is not str")?;
                        // TODO(bug): if source name is a file, this doesn't work (.join(""))
                        pipeline_job_tx
                            .send_async((
                                (dir_source_idx, name.to_owned()),
                                dir_source.path.join(name),
                            ))
                            .await
                            .map_err(|_| eyre!("unable to send to pipeline."))?;
                        *submitted_mut += 1;
                    }
                }
            }

            drop(pipeline_job_tx);
            Ok::<_, eyre::Report>(())
        },
        async {
            while let Ok(((dir_source_idx, job_id), file_store_opt)) = pipeline.next_result().await
            {
                chunk_file_maps[dir_source_idx].insert(&job_id, file_store_opt);
                completed += 1;
                Span::current().pb_inc(1);
            }
            // eprintln!("fin rec");
            Ok::<_, eyre::Report>(())
        }
    );

    submitter_task?;
    receiver_task?;

    ensure!(
        completed == submitted,
        "completed: {completed:?} != submitted {submitted:?}"
    );

    assert_eq!(dir_sources.len(), chunk_file_maps.len());

    let chunkmaps = pipeline.finish_into_chunkmaps().await?;

    Ok(BackupDirSourcesReturn {
        chunkmaps,
        dir_source_returns: dir_sources
            .into_iter()
            .zip(chunk_file_maps.into_iter())
            .collect(),
    })
}

async fn backup_virtual_sources(
    sources: &BTreeMap<String, SourceDescriptor>,
    now: DateTime<Utc>,
    pwc: Arc<PileWithCache<BoxedWormFileProvider>>,
    new_unflushed_chunks: Arc<DashSet<ChunkId>>,
) -> eyre::Result<Vec<VirtualSourceReturn>> {
    let mut joinset: JoinSet<eyre::Result<VirtualSourceReturn>> = JoinSet::new();

    for (source_name, source) in sources {
        if source.is_virtual_source() {
            joinset.spawn(backup_virtual_source(
                get_pointer_name_at(source_name, now),
                source.clone(),
                pwc.clone(),
                new_unflushed_chunks.clone(),
            ));
        }
    }

    let mut results = Vec::new();
    while let Some(result_res_res) = joinset.join_next().await {
        results.push(result_res_res??);
    }

    Ok(results)
}

struct VirtualSourceReturn {
    pub pointer_name: String,
    pub pointer: Pointer,
    pub chunkmaps: Vec<(BloblogId, BTreeMap<ChunkId, BlobLocator>)>,
}

async fn backup_virtual_source(
    pointer_name: String,
    source: SourceDescriptor,
    pwc: Arc<PileWithCache<BoxedWormFileProvider>>,
    new_unflushed_chunks: Arc<DashSet<ChunkId>>,
) -> eyre::Result<VirtualSourceReturn> {
    let SourceDescriptorInner::VirtualSource(virtual_source) = &source.inner else {
        bail!("bug: non-VS SDI passed to BVS");
    };

    let mut storing_state = StoringState::new(pwc.clone(), new_unflushed_chunks, Handle::current())
        .await
        .context("failed to create storing state")?;
    let mut sbw = StoringBloblogWriters::default();
    let ((chunkref, size), mut sbw, mut storing_state) = tokio::task::spawn_blocking({
        let virtual_source = virtual_source.clone();
        move || -> eyre::Result<((RecursiveChunkRef, u64), StoringBloblogWriters, StoringState)> {
            let child = open_stdout_backup_process(&virtual_source.extra_args, &virtual_source.helper)?;
            Ok((storing_state.store_full_stream(child.stdout.unwrap(), &mut sbw).context("Failed to store stream into Yama pile")?, sbw, storing_state))
        }
    }).await??;

    sbw.finish_bloblogs(&mut storing_state)
        .await
        .context("Failed to finish bloblogs")?;
    let chunkmaps = storing_state.new_bloblogs;

    // Assemble and write a pointer
    let uid = get_current_uid() as u16;
    let gid = get_current_gid() as u16;
    let tree = TreeNode::NormalFile {
        mtime: SystemTime::now()
            .duration_since(UNIX_EPOCH)
            .map(|d| d.as_millis() as u64)
            .unwrap_or(0),
        ownership: FilesystemOwnership { uid, gid },
        permissions: FilesystemPermissions { mode: 0o600 },
        size,
        content: chunkref,
    };
    let (uids, gids) =
        create_uidgid_lookup_tables(&tree).context("failed to create uid/gid tables")?;

    let VirtualSourceKind::Stdout { filename } = &virtual_source.kind;

    Ok(VirtualSourceReturn {
        pointer_name,
        pointer: Pointer {
            parent: None,
            root: RootTreeNode {
                name: filename.clone(),
                node: tree,
            },
            uids,
            gids,
        },
        chunkmaps,
    })
}

pub fn open_stdout_backup_process(
    extra_args: &HashMap<String, toml::Value>,
    program_name: &str,
) -> eyre::Result<Child> {
    let mut child = Command::new(format!("datman-helper-{}-backup", program_name))
        .stdout(Stdio::piped())
        .stderr(Stdio::inherit())
        .stdin(Stdio::piped())
        .spawn()?;
    let mut child_stdin = child.stdin.as_mut().unwrap();
    serde_json::to_writer(&mut child_stdin, extra_args)?;
    child_stdin.flush()?;
    // close stdin!
    child.stdin = None;
    Ok(child)
}
