/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/

use clap::{Parser, Subcommand};
use datman::backup::{backup, BackupOptions};
use datman::descriptor_config::{
    load_descriptor, Descriptor, PilePathOrConnector, SourceDescriptor,
};
use datman::extract::{
    extract, load_pointers_for_extraction, merge_roots_for_batch_extract, select_to_extract,
};
use eyre::{bail, Context, ContextCompat};
use std::collections::{BTreeMap, BTreeSet};
use std::path::PathBuf;
use std::str::FromStr;
use std::sync::Arc;
use tracing::info;
use tracing_indicatif::IndicatifLayer;
use tracing_subscriber::filter::filter_fn;
use tracing_subscriber::layer::SubscriberExt;
use tracing_subscriber::util::SubscriberInitExt;
use tracing_subscriber::Layer;
use yama::debugging::register_sigusr1_backtrace_helper;
use yama::get_hostname;
use yama::open::{open_lock_and_update_cache, open_lock_and_update_cache_with_connector};
use yama::pile_with_cache::PileWithCache;
use yama_wormfile::boxed::BoxedWormFileProvider;

#[derive(Clone, Debug)]
pub struct PileAndPointer {
    pub pile_path: Option<PathBuf>,
    pub pointer: PointerName,
}

#[derive(Clone, Debug)]
#[repr(transparent)]
pub struct PointerName(String);

impl FromStr for PointerName {
    type Err = eyre::Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        if !s
            .chars()
            .all(|c| c.is_alphanumeric() || ['_', '+', '-', ':'].contains(&c))
        {
            bail!("Bad pointer name: {s:?}");
        }
        Ok(PointerName(s.to_owned()))
    }
}

impl FromStr for PileAndPointer {
    type Err = eyre::Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s.split_once(":") {
            None => Ok(PileAndPointer {
                pile_path: None,
                pointer: PointerName::from_str(s)?,
            }),
            Some((pile_path, pointer)) => Ok(PileAndPointer {
                pile_path: Some(PathBuf::from(pile_path)),
                pointer: PointerName::from_str(pointer)?,
            }),
        }
    }
}

#[derive(Clone, Debug)]
pub struct PileAndPointerWithSubTree {
    pub pile_path: Option<PathBuf>,
    pub pointer: PointerName,
    // TODO how to represent...
    pub sub_tree: String,
}

impl FromStr for PileAndPointerWithSubTree {
    type Err = eyre::Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let (pile_path, pointer_and_subtree) = match s.split_once(":") {
            None => (None, s),
            Some((pile_path, pointer)) => (Some(PathBuf::from(pile_path)), pointer),
        };

        if let Some(slash) = pointer_and_subtree.find('/') {
            Ok(PileAndPointerWithSubTree {
                pile_path,
                pointer: PointerName::from_str(&pointer_and_subtree[0..slash])?,
                sub_tree: pointer_and_subtree[slash + 1..].to_owned(),
            })
        } else {
            Ok(PileAndPointerWithSubTree {
                pile_path,
                pointer: PointerName::from_str(&pointer_and_subtree)?,
                sub_tree: String::new(),
            })
        }
    }
}

#[derive(Parser, Clone, Debug)]
pub struct DatmanArgs {
    #[arg(long, env = "DATMAN_CONFIG", default_value = "datman.toml")]
    config: PathBuf,

    #[command(subcommand)]
    command: DatmanCommand,
}

#[derive(Subcommand, Clone, Debug)]
pub enum DatmanCommand {
    BackupOne {
        source_name: String,
        pile_name: String,
        #[clap(flatten)]
        options: BackupOptions,
    },

    BackupAll {
        pile_name: String,
        #[clap(flatten)]
        options: BackupOptions,
    },

    ExtractOne {
        pile_name: String,
        source_name: String,
        destination: PathBuf,
    },

    ExtractAll {
        pile_name: String,
        destination: PathBuf,
    },
}

const PROGRESS_SPANS: &'static [&'static str] = &[
    "store_file",
    "storing",
    "unpack_files",
    "expand_chunkrefs",
    "extract_files",
    "check_all_chunks",
];

#[tokio::main]
pub async fn main() -> eyre::Result<()> {
    let indicatif_layer = IndicatifLayer::new();
    let stderr_writer = indicatif_layer.get_stderr_writer();
    let indicatif_layer = indicatif_layer.with_filter(filter_fn(|span_metadata| {
        (span_metadata.target().starts_with("yama") || span_metadata.target().starts_with("datman"))
            && PROGRESS_SPANS.contains(&span_metadata.name())
    }));

    tracing_subscriber::registry()
        .with(
            tracing_subscriber::EnvFilter::try_from_default_env().unwrap_or_else(|_| {
                "sqlx=warn,yama=debug,datman=debug,yama_wormfile_sftp=debug,info".into()
            }),
        )
        .with(tracing_subscriber::fmt::layer().with_writer(stderr_writer))
        .with(indicatif_layer)
        .init();

    register_sigusr1_backtrace_helper();

    let args: DatmanArgs = dbg!(DatmanArgs::parse());

    let descriptor = load_descriptor(&args.config)
        .await
        .context("failed to load Datman descriptor")?;
    dbg!(&descriptor);

    match args.command {
        DatmanCommand::BackupOne {
            source_name,
            pile_name,
            options,
        } => {
            let lock_name = format!("{} datman backup {:?}", get_hostname(), source_name);
            let pwc = open_destination(&descriptor, &pile_name, lock_name).await?;

            let source = descriptor
                .sources
                .get(&source_name)
                .context("no source by that name")?;
            let my_hostname = get_hostname();
            if &source.host != &my_hostname {
                bail!(
                    "Current hostname is {:?}, not {:?} as expected for this source.",
                    my_hostname,
                    source.host
                );
            }

            let mut sources_to_backup = BTreeMap::new();
            sources_to_backup.insert(source_name.clone(), source.clone());

            backup(pwc, sources_to_backup, &options).await?;
        }
        DatmanCommand::BackupAll { pile_name, options } => {
            let lock_name = format!("{} datman backupall", get_hostname());
            let pwc = open_destination(&descriptor, &pile_name, lock_name).await?;

            let my_hostname = get_hostname();
            let sources_to_backup: BTreeMap<String, SourceDescriptor> = descriptor
                .sources
                .clone()
                .into_iter()
                .filter(|(_, source)| &source.host == &my_hostname)
                .collect();

            if sources_to_backup.len() == 0 {
                bail!(
                    "No sources to back up! The current hostname is {:?}; is it correct?",
                    my_hostname
                );
            }

            info!(
                "Backing up the following {} sources: {:?}",
                sources_to_backup.len(),
                sources_to_backup.keys().collect::<Vec<_>>()
            );

            backup(pwc, sources_to_backup, &options).await?;
        }
        DatmanCommand::ExtractOne {
            pile_name,
            source_name,
            destination,
        } => {
            let lock_name = format!("{} datman extract {:?}", get_hostname(), source_name);
            let pwc = open_destination(&descriptor, &pile_name, lock_name).await?;

            let mut sources = BTreeSet::new();
            sources.insert(source_name.clone());
            let selected = select_to_extract(&pwc, sources, None, None, false).await?;
            let mut for_extraction = load_pointers_for_extraction(pwc.clone(), selected).await?;
            assert_eq!(for_extraction.len(), 1);
            let root_node = for_extraction.remove(&source_name).unwrap();
            extract(pwc, root_node.node, &destination).await?;
        }
        DatmanCommand::ExtractAll {
            pile_name,
            destination,
        } => {
            let lock_name = format!("{} datman extractall", get_hostname());
            let pwc = open_destination(&descriptor, &pile_name, lock_name).await?;

            let sources = descriptor.sources.keys().cloned().collect();
            let selected = select_to_extract(&pwc, sources, None, None, false).await?;
            let for_extraction = load_pointers_for_extraction(pwc.clone(), selected).await?;
            let merged_node = merge_roots_for_batch_extract(for_extraction);
            extract(pwc, merged_node, &destination).await?;
        }
    }

    Ok(())
}

async fn open_destination(
    descriptor: &Descriptor,
    pile_name: &str,
    lock_name: String,
) -> eyre::Result<Arc<PileWithCache<BoxedWormFileProvider>>> {
    let path_or_connector = descriptor
        .piles
        .get(pile_name)
        .context("no pile by that name")?;
    match path_or_connector {
        PilePathOrConnector::PilePath(path) => {
            open_lock_and_update_cache(path.clone(), lock_name).await
        }
        PilePathOrConnector::PileConnector { scheme, yamakey } => {
            open_lock_and_update_cache_with_connector(scheme, pile_name, yamakey, lock_name).await
        }
    }
}
