/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/

use eyre::{Context, ContextCompat};
use serde::{Deserialize, Serialize};
use std::collections::HashMap;
use std::path::{Path, PathBuf};
use yama::pile_connector::PileConnectionScheme;

// TODO how do we handle?:
//      - (important) yama push of one pile to another
//      - backup policy stuff like 'minimum backup frequency' ... show when it's not been done
//      - backup policy stuff like 'minimum on two different disks, not powered at the same time...'

#[derive(Clone, Serialize, Deserialize, Debug)]
pub struct Descriptor {
    /// Sources
    pub sources: HashMap<String, SourceDescriptor>,

    /// Paths to destination Yama Piles. Remote Piles need a local virtual pile to specify the layers.
    pub piles: HashMap<String, PilePathOrConnector>,

    #[serde(default)]
    #[serde(skip_serializing_if = "Option::is_none")]
    pub retention: Option<RetentionPolicyConfig>,
}

#[derive(Clone, Serialize, Deserialize, Debug)]
#[serde(untagged)]
pub enum PilePathOrConnector {
    PilePath(PathBuf),
    PileConnector {
        #[serde(flatten)]
        scheme: PileConnectionScheme,
        yamakey: PathBuf,
    },
}

#[derive(Clone, Serialize, Deserialize, Debug)]
pub struct RetentionPolicyConfig {
    pub daily: u32,
    pub weekly: u32,
    pub monthly: u32,
    pub yearly: u32,
}

#[derive(Clone, Serialize, Deserialize, Debug)]
pub struct SourceDescriptor {
    /// The host to run this backup task on.
    pub host: String,
    #[serde(flatten)]
    pub inner: SourceDescriptorInner,
}

impl SourceDescriptor {
    pub fn is_directory_source(&self) -> bool {
        matches!(&self.inner, &SourceDescriptorInner::DirectorySource { .. })
    }

    pub fn is_virtual_source(&self) -> bool {
        matches!(&self.inner, &SourceDescriptorInner::VirtualSource { .. })
    }
}

#[derive(Clone, Serialize, Deserialize, Debug)]
#[serde(untagged)]
pub enum SourceDescriptorInner {
    DirectorySource {
        path: PathBuf,
        #[serde(default)]
        cross_filesystems: bool,

        /// TODO Paths to ignore
        #[serde(default)]
        ignore: Vec<String>,
    },

    VirtualSource(VirtualSource),
}

#[derive(Clone, Serialize, Deserialize, Debug)]
pub struct VirtualSource {
    /// The name of the helper program that will be used to do this backup.
    pub helper: String,

    /// The kind of virtual source (how it operates).
    pub kind: VirtualSourceKind,

    #[serde(flatten)]
    pub extra_args: HashMap<String, toml::Value>,
}

#[derive(Clone, Serialize, Deserialize, Debug)]
#[serde(untagged)]
pub enum VirtualSourceKind {
    Stdout {
        #[serde(rename = "stdout")]
        filename: String,
    },
    // TODO(feature) TempDir
}

/// Loads a descriptor and resolves relative paths contained within.
pub async fn load_descriptor(path: &Path) -> eyre::Result<Descriptor> {
    let text = tokio::fs::read_to_string(path).await?;
    let mut descriptor: Descriptor = toml::de::from_str(&text)?;

    let dir = path
        .parent()
        .context("there must be a parent path for the descriptor file")?;

    // Absolutise pile paths
    for (_, pile_path_or_connector) in descriptor.piles.iter_mut() {
        match pile_path_or_connector {
            PilePathOrConnector::PilePath(pile_path) => {
                *pile_path = dir
                    .join(&*pile_path)
                    .canonicalize()
                    .context("Failed to canonicalise path in descriptor")?;
            }
            PilePathOrConnector::PileConnector {
                scheme:
                    PileConnectionScheme::Local {
                        directory: pile_path,
                    },
                ..
            } => {
                *pile_path = dir
                    .join(&*pile_path)
                    .canonicalize()
                    .context("Failed to canonicalise path in descriptor")?;
            }
            PilePathOrConnector::PileConnector { .. } => { /* nop */ }
        }
    }

    Ok(descriptor)
}
