use crate::datetime::HumanDateTime;
use crate::pointer_names::split_pointer_name;
use chrono::{DateTime, Utc};
use eyre::{bail, eyre, Context, ContextCompat};
use std::collections::btree_map::Entry;
use std::collections::{BTreeMap, BTreeSet};
use std::path::Path;
use std::sync::Arc;
use tracing::{info_span, warn, Instrument};
use yama::extract;
use yama::extract::flatten_treenode;
use yama::pile_with_cache::{PileWithCache, PointerIntegrationStatistics};
use yama_pile::tree::{FilesystemOwnership, FilesystemPermissions, RootTreeNode, TreeNode};
use yama_wormfile::boxed::BoxedWormFileProvider;

/// Given a list of source names and conditions to find pointers within,
/// returns a mapping of source names to pointers.
pub async fn select_to_extract(
    pwc: &PileWithCache<BoxedWormFileProvider>,
    sources: BTreeSet<String>,
    before: Option<HumanDateTime>,
    after: Option<HumanDateTime>,
    accept_partial: bool,
) -> eyre::Result<BTreeMap<String, String>> {
    let before = before.map(|dt| dt.0.with_timezone(&Utc));
    let after = after.map(|dt| dt.0.with_timezone(&Utc));
    let pointers_list = pwc
        .pile
        .list_pointers()
        .await
        .context("failed to list pointers")?;

    select_to_extract_impl(pointers_list, sources, before, after, accept_partial)
}

/// Given a list of source names and conditions to find pointers within,
/// returns a mapping of source names to pointers.
fn select_to_extract_impl(
    pointers_list: Vec<String>,
    sources: BTreeSet<String>,
    before: Option<DateTime<Utc>>,
    after: Option<DateTime<Utc>>,
    accept_partial: bool,
) -> eyre::Result<BTreeMap<String, String>> {
    if after.is_some() && before.is_some() {
        bail!("Can't specify both before and after!");
    }

    let mut pointers_by_source: BTreeMap<String, String> = BTreeMap::new();

    for pointer in pointers_list {
        if let Some((source_name, pointer_datetime)) = split_pointer_name(&pointer) {
            if !sources.contains(&source_name) {
                // Not a source that we're interested in.
                continue;
            }
            if let Some(before) = before {
                if before < pointer_datetime {
                    // datetime is after the 'before' time
                    continue;
                }
            } else if let Some(after) = after {
                if pointer_datetime < after {
                    // datetime is before the 'after' time
                    continue;
                }
            }

            match pointers_by_source.entry(source_name) {
                Entry::Vacant(ve) => {
                    ve.insert(pointer);
                }
                Entry::Occupied(mut oe) => {
                    let current_choice = oe.get_mut();
                    let (_, current_datetime) = split_pointer_name(&current_choice).unwrap();
                    let should_replace = if after.is_some() {
                        // if we want the first one after a time, we want the earliest option!
                        // so replace if new datetime is earlier than current
                        pointer_datetime < current_datetime
                    } else {
                        // replace if new datetime is after current datetime
                        current_datetime < pointer_datetime
                    };
                    if should_replace {
                        *current_choice = pointer;
                    }
                }
            }
        };
    }

    if pointers_by_source.is_empty() {
        bail!("No pointers selected for ANY of the sources: {sources:?}");
    }

    let missing: Vec<&String> = sources
        .iter()
        .filter(|src| !pointers_by_source.contains_key(*src))
        .collect();
    if !missing.is_empty() {
        if accept_partial {
            warn!("Some sources didn't have any pointers selected: {missing:?}. Continuing because --accept-partial passed.");
        } else {
            bail!("Some sources didn't have any pointers selected: {missing:?}. Pass --accept-partial if this is intended anyway.");
        }
    }

    Ok(pointers_by_source)
}

pub async fn load_pointers_for_extraction(
    pwc: Arc<PileWithCache<BoxedWormFileProvider>>,
    what_to_extract: BTreeMap<String, String>,
) -> eyre::Result<BTreeMap<String, RootTreeNode>> {
    let mut result = BTreeMap::new();
    for (source_name, pointer_name) in &what_to_extract {
        let mut stats = PointerIntegrationStatistics::default();
        let pointer = pwc
            .read_pointer_fully_integrated(&pointer_name, &mut stats)
            .await?
            .context("pointer doesn't exist??")?;
        // TODO(ownership): adapt uid/gids here
        result.insert(source_name.clone(), pointer.root);
    }
    Ok(result)
}

pub fn merge_roots_for_batch_extract(extracts: BTreeMap<String, RootTreeNode>) -> TreeNode {
    let mut children = BTreeMap::new();

    for (name, entry) in extracts {
        if matches!(entry.node, TreeNode::NormalFile { .. }) {
            let mut children2 = BTreeMap::new();
            children2.insert(entry.name, entry.node);
            children.insert(
                name,
                TreeNode::Directory {
                    ownership: FilesystemOwnership {
                        // TODO(ownership): populate this correctly (current user?)
                        uid: 0,
                        gid: 0,
                    },
                    permissions: FilesystemPermissions { mode: 0o700 },
                    children: children2,
                },
            );
        } else {
            children.insert(name, entry.node);
        }
    }

    TreeNode::Directory {
        ownership: FilesystemOwnership {
            // TODO(ownership): populate this correctly (current user?)
            uid: 0,
            gid: 0,
        },
        permissions: FilesystemPermissions { mode: 0o700 },
        children,
    }
}

pub async fn extract(
    pwc: Arc<PileWithCache<BoxedWormFileProvider>>,
    node: TreeNode,
    destination: &Path,
) -> eyre::Result<()> {
    let flat = flatten_treenode(&node)?;
    drop(node);

    extract::unpack_nonfiles(destination, &flat.nonfiles, false, true).await?;

    let extract_span = info_span!("extract_files");
    extract::unpack_files(&pwc, destination, &flat.files, false, true)
        .instrument(extract_span)
        .await?;

    Arc::try_unwrap(pwc)
        .map_err(|_| eyre!("pwc still in use; can't close down gracefully"))?
        .close()
        .await?;
    Ok(())
}
