/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/

use clap::{Parser, Subcommand};
use eyre::{bail, eyre, Context, ContextCompat};
use indicatif::ProgressStyle;
use patricia_tree::PatriciaMap;
use std::collections::BTreeSet;
use std::iter::Iterator;
use std::path::{Path, PathBuf};
use std::str::FromStr;
use std::sync::Arc;
use std::time::{SystemTime, UNIX_EPOCH};
use tokio::io::{stdin, AsyncBufReadExt, BufReader};
use tokio::runtime::Handle;
use tracing::{info, info_span, warn, Instrument, Span};
use tracing_indicatif::span_ext::IndicatifSpanExt;
use tracing_indicatif::IndicatifLayer;
use tracing_subscriber::filter::filter_fn;
use tracing_subscriber::layer::SubscriberExt;
use tracing_subscriber::util::SubscriberInitExt;
use tracing_subscriber::Layer;
use users::{get_current_gid, get_current_uid};
use yama::debugging::register_sigusr1_backtrace_helper;
use yama::extract::flatten_treenode;
use yama::init::pack_keyring;
use yama::open::{
    open_keyring_interactive, open_pile, pre_open_keyring, pre_open_keyring_at_path, update_cache,
};
use yama::pile_connector::PileConnectionScheme;
use yama::pile_with_cache::PointerIntegrationStatistics;
use yama::scan::create_uidgid_lookup_tables;
use yama::storing::{
    assemble_and_write_indices, StoragePipeline, StoringBloblogWriters, StoringState,
};
use yama::vacuum::forget_chunks::{find_forgettable_chunks, forget_chunks};
use yama::vacuum::merge_indices::{MERGE_TARGET_SIZE, MERGE_THRESHOLD_SIZE};
use yama::vacuum::repack_bloblogs_and_indices::{
    get_bloblogs_stats, perform_repack, select_bloblogs_for_repack,
};
use yama::{check, extract, get_hostname, init, scan, vacuum, PROGRESS_BAR_STYLE};
use yama_midlevel_crypto::byte_layer::{ByteLayer, CborSerde};
use yama_midlevel_crypto::chunk_id::ChunkIdKey;
use yama_pile::definitions::{
    IndexBloblogEntry, PackedPileConfig, PileConfig, RecursiveChunkRef, SUPPORTED_YAMA_PILE_VERSION,
};
use yama_pile::keyring::{Keyring, INDIVIDUAL_KEYS, KEY_GROUPS};
use yama_pile::locks::LockKind;
use yama_pile::pointers::Pointer;
use yama_pile::tree::unpopulated::ScanEntry;
use yama_pile::tree::{
    assemble_tree_from_scan_entries, differentiate_node_in_place, FilesystemOwnership,
    FilesystemPermissions, RootTreeNode, TreeNode,
};
use yama_pile::FILE_YAMA_CONNECTOR;

#[derive(Clone, Debug)]
pub struct PileAndPointer {
    pub pile_path: Option<PathBuf>,
    pub pointer: PointerName,
}

#[derive(Clone, Debug)]
#[repr(transparent)]
pub struct PointerName(String);

impl FromStr for PointerName {
    type Err = eyre::Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        if !s
            .chars()
            .all(|c| c.is_alphanumeric() || ['_', '+', '-', ':'].contains(&c))
        {
            bail!("Bad pointer name: {s:?}");
        }
        Ok(PointerName(s.to_owned()))
    }
}

impl FromStr for PileAndPointer {
    type Err = eyre::Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s.split_once(":") {
            None => Ok(PileAndPointer {
                pile_path: None,
                pointer: PointerName::from_str(s)?,
            }),
            Some((pile_path, pointer)) => Ok(PileAndPointer {
                pile_path: Some(PathBuf::from(pile_path)),
                pointer: PointerName::from_str(pointer)?,
            }),
        }
    }
}

#[derive(Clone, Debug)]
pub struct PileAndPointerWithSubTree {
    pub pile_path: Option<PathBuf>,
    pub pointer: PointerName,
    // TODO how to represent...
    pub sub_tree: String,
}

impl FromStr for PileAndPointerWithSubTree {
    type Err = eyre::Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let (pile_path, pointer_and_subtree) = match s.split_once(":") {
            None => (None, s),
            Some((pile_path, pointer)) => (Some(PathBuf::from(pile_path)), pointer),
        };

        if let Some(slash) = pointer_and_subtree.find('/') {
            Ok(PileAndPointerWithSubTree {
                pile_path,
                pointer: PointerName::from_str(&pointer_and_subtree[0..slash])?,
                sub_tree: pointer_and_subtree[slash + 1..].to_owned(),
            })
        } else {
            Ok(PileAndPointerWithSubTree {
                pile_path,
                pointer: PointerName::from_str(&pointer_and_subtree)?,
                sub_tree: String::new(),
            })
        }
    }
}

#[derive(Parser, Clone, Debug)]
pub enum YamaCommand {
    /// Create a new Yama pile.
    Init {
        #[arg(long)]
        sftp: bool,
        #[arg(long)]
        s3: bool,
        #[arg(long)]
        no_password: bool,

        #[arg(long)]
        zstd_dict: Option<PathBuf>,
        #[arg(long)]
        no_zstd_dict: bool,
    },

    /// Keyring management commands; see `yama keyring --help`.
    #[command(subcommand)]
    Keyring(KeyringCommand),

    /// Store a file or directory to a Yama pile.
    Store {
        source: PathBuf,
        destination: PileAndPointer,

        #[arg(long)]
        overwrite: bool,

        /// Name of a parent pointer to use, if storing an incremental update.
        /// Although optional, this can make storing time much faster as it can prevent unchanged
        /// files from being rescanned.
        #[arg(short = 'p', long)]
        parent: Option<PointerName>,
    },

    /// Store an input stream to the Yama pile.
    StoreStdin {
        destination: PileAndPointer,

        #[arg(long)]
        overwrite: bool,

        /// A name to give to the file that the stream is stored as. Otherwise it will just be called
        /// 'stream'.
        #[arg(short = 'n', long)]
        name: Option<String>,
    },

    /// Extract a file, directory or output stream from a Yama pile.
    Extract {
        source: PileAndPointerWithSubTree,
        destination: PathBuf,
    },

    /// Extract an output stream from a Yama pile.
    ExtractStdout { source: PileAndPointerWithSubTree },

    // TODO Mount { ... },
    Check {
        #[arg(long, short = '1')]
        pointers: bool,

        #[arg(long, short = '2')]
        shallow: bool,

        #[arg(long, short = '9')]
        intensive: bool,
    },
    // TODO lsp, rmp
    /// Perform maintenance tasks, usually freeing up space or clumping together files to reduce
    /// clutter.
    Vacuum {
        /// Perform all maintenance and space-saving tasks.
        #[arg(long, short = 'a')]
        all: bool,

        /// Merge indices together. Implied by -a.
        #[arg(long, short = 'm')]
        merge: bool,

        /// Forget chunks from indices. Implied by -a.
        /// This process is slow because it involves walking all pointers to see which chunks can be
        /// forgotten.
        #[arg(long, short = 'f')]
        forget: bool,

        /// Repack bloblogs and corresponding indices. Implied by -a.
        #[arg(long, short = 'r')]
        repack: bool,

        /// Delete unreferenced bloblogs. Implied by -a.
        #[arg(long, short = 'd')]
        delete_unrefd_bloblogs: bool,
    }, // TODO `locks` to inspect locks
}

#[derive(Subcommand, Clone, Debug)]
pub enum KeyringCommand {
    Inspect {
        file: PathBuf,
    },
    Create {
        /// Where to put the new keyring.
        new: PathBuf,

        /// What keyring to use to copy keys from. Default is `master.yamakeyring`.
        #[arg(long)]
        from: Option<PathBuf>,

        /// Which keys to put into the new keyring.
        #[arg(long)]
        with: String,

        /// Don't password-protect the new keyring.
        /// It goes without saying that you should never disclose an unprotected keyring to anyone.
        #[arg(long)]
        no_password: bool,

        /// Use weaker, faster password protection for the new keyring.
        /// You should use different passwords for weak and strong keyrings!
        /// This may be suitable if the keyring is kept in a moderately safe environment, but
        /// prefer a strongly-protected keyring if the file might be reasonably accessible by
        /// enemies.
        #[arg(long)]
        weak: bool,
    }, // TODO ChangePassword
}

const PROGRESS_SPANS: &'static [&'static str] = &[
    "store_file",
    "storing",
    "unpack_files",
    "expand_chunkrefs",
    "extract_files",
    "check_all_chunks",
];

#[tokio::main]
async fn main() -> eyre::Result<()> {
    let indicatif_layer = IndicatifLayer::new();
    let stderr_writer = indicatif_layer.get_stderr_writer();
    let indicatif_layer = indicatif_layer.with_filter(filter_fn(|span_metadata| {
        span_metadata.target().starts_with("yama") && PROGRESS_SPANS.contains(&span_metadata.name())
    }));

    tracing_subscriber::registry()
        .with(
            tracing_subscriber::EnvFilter::try_from_default_env()
                .unwrap_or_else(|_| "sqlx=warn,yama=debug,info".into()),
        )
        .with(tracing_subscriber::fmt::layer().with_writer(stderr_writer))
        .with(indicatif_layer)
        .init();

    register_sigusr1_backtrace_helper();

    match dbg!(YamaCommand::parse()) {
        YamaCommand::Init {
            sftp,
            s3,
            no_password,
            zstd_dict,
            no_zstd_dict,
        } => {
            if zstd_dict.is_some() && no_zstd_dict {
                bail!("Can only choose one of --zstd-dict or --no-zstd-dict!");
            }
            if zstd_dict.is_none() && !no_zstd_dict {
                bail!("Must choose one of --zstd-dict or --no-zstd-dict!");
            }

            // Begin. Create a pile in the current directory.
            let pile_path = PathBuf::from(".");

            // Check we are happy to make a pile or connector here.
            init::pre_init_check(&pile_path).await?;

            let connection_scheme = set_up_connection(sftp, s3, &pile_path).await?;
            let connection = connection_scheme
                .connect_to_wormfileprovider()
                .await
                .context("Failed to open pile destination")?;

            info!("Pile destination seems OK; can proceed to set up pile.");

            let master_password = if no_password {
                warn!("Not setting a master password. The master keyring will be unprotected.");
                None
            } else {
                println!("enter master password:");
                let stdin = stdin();
                let mut stdin_br = BufReader::new(stdin);
                let mut line = String::new();
                stdin_br.read_line(&mut line).await?;
                Some(line.trim().to_owned())
            };

            let master_keyring = Keyring::generate_new_master();
            let master_key_packed = pack_keyring(
                master_keyring.clone(),
                master_password.as_ref().map(|s| s.as_ref()),
            )?;

            let zstd_dict = if no_zstd_dict {
                None
            } else {
                let zstd_dict_path = zstd_dict.unwrap();
                Some(Arc::new(
                    tokio::fs::read(&zstd_dict_path).await.with_context(|| {
                        format!("failed to read Zstd dict at {zstd_dict_path:?}")
                    })?,
                ))
            };

            let pile_config = PileConfig {
                yama_pile_version: SUPPORTED_YAMA_PILE_VERSION.to_owned(),
                chunk_id_key: ChunkIdKey::new_rand(),
                zstd_dict,
            };
            let packed_pile_config: PackedPileConfig = master_keyring
                .w_config
                .unwrap()
                .make_locked_asymbox(CborSerde::serialise(&pile_config).unwrap());

            let make_pileside_copy_of_master_key = !(sftp || s3); // TODO ask for these

            init::init_pile(
                &connection,
                packed_pile_config,
                if make_pileside_copy_of_master_key {
                    Some(master_key_packed.clone())
                } else {
                    None
                },
            )
            .await?;

            if sftp || s3 {
                // For remote piles: save a master keyring copy locally
                tokio::fs::write(
                    pile_path.join("master.yamakeyring"),
                    &master_key_packed.into_byte_vec(),
                )
                .await
                .context("failed to make local copy of master keyring")?;
                tokio::fs::write(
                    pile_path.join(FILE_YAMA_CONNECTOR),
                    toml::to_string_pretty(&connection_scheme)
                        .context("failed to serialise connector")?,
                )
                .await
                .context("failed to write connector")?;
            }
        }
        YamaCommand::Store {
            source,
            destination,
            overwrite,
            parent,
        } => {
            let pile_connector_path = destination.pile_path.unwrap_or(PathBuf::from("."));
            let keyring = pre_open_keyring(&pile_connector_path).await?;
            let keyring = open_keyring_interactive(keyring).await?;

            let pwc = open_pile(
                &pile_connector_path,
                keyring,
                LockKind::Shared,
                format!("{} store {:?}", get_hostname(), destination.pointer),
            )
            .await?;
            update_cache(&pwc).await?;

            let parent_pointer = if let Some(ref parent) = parent {
                let mut stats = PointerIntegrationStatistics::default();
                let pointer = pwc
                    .read_pointer_fully_integrated(parent.0.as_str(), &mut stats)
                    .await
                    .with_context(|| format!("whilst reading parent pointer: {parent:?}"))?
                    .with_context(|| {
                        format!("it appears that the pointer {parent:?} does not exist")
                    })?;
                assert!(pointer.parent.is_none());
                Some(pointer)
            } else {
                None
            };

            let source2 = source.clone();
            let scan_entry_map = tokio::task::spawn_blocking(move || {
                scan::scan(&source2, &Vec::new()).context("Failed to scan")
            })
            .await??;

            let pwc = Arc::new(pwc);

            let (mut chunk_file_map, pruned_scan_entry_map, prepopulated_scan_entry_map) =
                if let Some(ref parent_node) = parent_pointer {
                    let (cfm, pruned, prepopulated) =
                        scan::prepopulate_unmodified(&parent_node.root.node, &scan_entry_map);

                    (cfm, pruned, prepopulated)
                } else {
                    (
                        PatriciaMap::<Option<(RecursiveChunkRef, u64)>>::new(),
                        scan_entry_map,
                        PatriciaMap::new(),
                    )
                };

            let store_span = info_span!("storing");
            // store_span.pb_set_style(&ProgressStyle::default_bar());
            store_span.pb_set_style(
                &ProgressStyle::default_bar()
                    .template(PROGRESS_BAR_STYLE)
                    .unwrap(),
            );
            store_span.pb_set_message("storing files");
            store_span.pb_set_length(
                pruned_scan_entry_map
                    .values()
                    .filter(|v| matches!(v, ScanEntry::NormalFile { .. }))
                    .count() as u64,
            );
            let store_span_entered = store_span.enter();

            let new_unflushed_chunks = Arc::new(Default::default());
            let (pipeline, pipeline_job_tx) =
                StoragePipeline::launch_new(4, pwc.clone(), new_unflushed_chunks).await?;

            let source2 = source.clone();
            let pruned_scan_entry_map2 = &pruned_scan_entry_map;
            let (submitter_task, receiver_task) = tokio::join!(
                async move {
                    let pipeline_job_tx = pipeline_job_tx;
                    for (name_bytes, scan_entry) in pruned_scan_entry_map2.iter() {
                        if let ScanEntry::NormalFile { .. } = scan_entry {
                            let name = std::str::from_utf8(name_bytes.as_slice())
                                .context("name is not str")?;
                            let path = if name != "" {
                                source2.join(name)
                            } else {
                                source2.clone()
                            };
                            pipeline_job_tx
                                .send_async((name.to_owned(), path))
                                .await
                                .map_err(|_| eyre!("unable to send to pipeline."))?;
                        }
                    }
                    drop(pipeline_job_tx);
                    Ok::<_, eyre::Report>(())
                },
                async {
                    while let Ok((job_id, file_store_opt)) = pipeline.next_result().await {
                        chunk_file_map.insert(&job_id, file_store_opt);
                        Span::current().pb_inc(1);
                    }
                    // eprintln!("fin rec");
                    Ok::<_, eyre::Report>(())
                }
            );
            submitter_task?;
            receiver_task?;

            drop(store_span_entered);
            drop(store_span);

            info!("All files stored, writing indices...");

            // Write indices for the new bloblogs we have created. This is a prerequisite for creating a pointer.
            let chunkmaps = pipeline
                .finish_into_chunkmaps()
                .await
                .context("failed to finish into chunkmaps")?;
            assemble_and_write_indices(&pwc, chunkmaps)
                .await
                .context("failed to assemble and write indices")?;

            info!("All indices stored, writing pointer...");

            // Assemble and write a pointer
            let mut tree = assemble_tree_from_scan_entries(
                pruned_scan_entry_map,
                prepopulated_scan_entry_map,
                chunk_file_map,
            )
            .context("failed to assemble tree")?;
            let (uids, gids) =
                create_uidgid_lookup_tables(&tree).context("failed to create uid/gid tables")?;

            if let Some(ref parent_node) = parent_pointer {
                differentiate_node_in_place(&mut tree, &parent_node.root.node)
                    .context("failed to differentiate?")?;
            }

            pwc.pile
                .write_pointer(
                    destination.pointer.0.as_str(),
                    overwrite,
                    &Pointer {
                        parent: parent.as_ref().map(|p| p.0.clone()),
                        root: RootTreeNode {
                            name: source
                                .file_name()
                                .map(|oss| oss.to_str())
                                .flatten()
                                .unwrap_or("")
                                .to_owned(),
                            node: tree,
                        },
                        uids,
                        gids,
                    },
                )
                .await
                .context("failed to write pointer")?;

            Arc::try_unwrap(pwc)
                .map_err(|_| eyre!("pwc still in use; can't close down gracefully"))?
                .close()
                .await?;
        }
        YamaCommand::StoreStdin {
            destination,
            overwrite,
            name,
        } => {
            let pile_connector_path = destination.pile_path.unwrap_or(PathBuf::from("."));
            let keyring = pre_open_keyring(&pile_connector_path).await?;
            let keyring = open_keyring_interactive(keyring).await?;

            let pwc = open_pile(
                &pile_connector_path,
                keyring,
                LockKind::Shared,
                format!("{} store {:?}", get_hostname(), destination.pointer),
            )
            .await?;
            update_cache(&pwc).await?;

            let pwc = Arc::new(pwc);

            let store_span = info_span!("storing");
            // store_span.pb_set_style(&ProgressStyle::default_bar());
            // TODO INDETERMINATE PROGRESS BAR with bytes shown?
            store_span.pb_set_style(
                &ProgressStyle::default_bar()
                    .template(PROGRESS_BAR_STYLE)
                    .unwrap(),
            );
            store_span.pb_set_message("storing files");
            store_span.pb_set_length(1u64);
            // TODO Dirty
            let _store_span_entered = store_span.enter();

            let new_unflushed_chunks = Arc::new(Default::default());
            let mut storing_state =
                StoringState::new(pwc.clone(), new_unflushed_chunks, Handle::current())
                    .await
                    .context("failed to create storing state")?;
            let mut sbw = StoringBloblogWriters::default();
            let stdin = std::io::BufReader::new(
                io_streams::StreamReader::stdin().context("failed to open stdin")?,
            );
            let (chunkref, size) = storing_state
                .store_full_stream(stdin, &mut sbw)
                .context("Failed to store stream into Yama pile")?;

            sbw.finish_bloblogs(&mut storing_state)
                .await
                .context("Failed to finish bloblogs")?;

            info!("Stream stored, writing indices...");

            // Write indices for the new bloblogs we have created. This is a prerequisite for creating a pointer.
            let chunkmaps = storing_state.new_bloblogs;

            assemble_and_write_indices(
                &pwc,
                chunkmaps
                    .into_iter()
                    .map(|(k, nb)| {
                        (
                            k,
                            IndexBloblogEntry {
                                chunks: nb,
                                forgotten_bytes: 0,
                            },
                        )
                    })
                    .collect(),
            )
            .await
            .context("failed to assemble and write indices")?;

            info!("All indices stored, writing pointer...");

            // Assemble and write a pointer
            let uid = get_current_uid() as u16;
            let gid = get_current_gid() as u16;
            let tree = TreeNode::NormalFile {
                mtime: SystemTime::now()
                    .duration_since(UNIX_EPOCH)
                    .map(|d| d.as_millis() as u64)
                    .unwrap_or(0),
                ownership: FilesystemOwnership { uid, gid },
                permissions: FilesystemPermissions { mode: 0o600 },
                size,
                content: chunkref,
            };
            let (uids, gids) =
                create_uidgid_lookup_tables(&tree).context("failed to create uid/gid tables")?;

            pwc.pile
                .write_pointer(
                    destination.pointer.0.as_str(),
                    overwrite,
                    &Pointer {
                        parent: None,
                        root: RootTreeNode {
                            name: name.unwrap_or_else(|| String::from("stream")),
                            node: tree,
                        },
                        uids,
                        gids,
                    },
                )
                .await
                .context("failed to write pointer")?;

            Arc::try_unwrap(pwc)
                .map_err(|_| eyre!("pwc still in use; can't close down gracefully"))?
                .close()
                .await?;
        }
        YamaCommand::Extract {
            source,
            destination,
        } => {
            let pile_connector_path = source
                .pile_path
                .as_ref()
                .map(|p| p.as_ref())
                .unwrap_or(Path::new("."));
            let keyring = pre_open_keyring(&pile_connector_path).await?;
            let keyring = open_keyring_interactive(keyring).await?;

            let pwc = Arc::new(
                open_pile(
                    &pile_connector_path,
                    keyring,
                    LockKind::Shared,
                    format!("{} store {:?}", get_hostname(), source.pointer),
                )
                .await?,
            );
            update_cache(&pwc).await?;

            let mut stats = PointerIntegrationStatistics::default();
            let pointer = pwc
                .read_pointer_fully_integrated(source.pointer.0.as_str(), &mut stats)
                .await
                .context("failed to read pointer")?
                .with_context(|| {
                    format!(
                        "it appears that the pointer {:?} does not exist",
                        source.pointer
                    )
                })?;
            assert!(pointer.parent.is_none());

            let node = if source.sub_tree.is_empty() {
                &pointer.root.node
            } else {
                let mut current = &pointer.root.node;
                for subpath in source.sub_tree.split('/') {
                    if let TreeNode::Directory { children, .. } = current {
                        current = children.get(subpath).with_context(|| {
                            format!("can't descend into {subpath}: doesn't exist in directory.")
                        })?;
                    } else {
                        bail!("can't descend into {subpath}; parent isn't a directory...");
                    }
                }
                current
            };

            let flat = flatten_treenode(&node)?;

            extract::unpack_nonfiles(&destination, &flat.nonfiles, false, true).await?;

            let extract_span = info_span!("extract_files");
            extract::unpack_files(&pwc, &destination, &flat.files, false, true)
                .instrument(extract_span)
                .await?;

            Arc::try_unwrap(pwc)
                .map_err(|_| eyre!("pwc still in use; can't close down gracefully"))?
                .close()
                .await?;
        }
        YamaCommand::ExtractStdout { source } => {
            let pile_connector_path = source
                .pile_path
                .as_ref()
                .map(|p| p.as_ref())
                .unwrap_or(Path::new("."));
            let keyring = pre_open_keyring(&pile_connector_path).await?;
            let keyring = open_keyring_interactive(keyring).await?;

            let pwc = Arc::new(
                open_pile(
                    &pile_connector_path,
                    keyring,
                    LockKind::Shared,
                    format!("{} store {:?}", get_hostname(), source.pointer),
                )
                .await?,
            );
            update_cache(&pwc).await?;

            let mut stats = PointerIntegrationStatistics::default();
            let pointer = pwc
                .read_pointer_fully_integrated(source.pointer.0.as_str(), &mut stats)
                .await
                .context("failed to read pointer")?
                .with_context(|| {
                    format!(
                        "it appears that the pointer {:?} does not exist",
                        source.pointer
                    )
                })?;
            assert!(pointer.parent.is_none());

            let node = if source.sub_tree.is_empty() {
                &pointer.root.node
            } else {
                let mut current = &pointer.root.node;
                for subpath in source.sub_tree.split('/') {
                    if let TreeNode::Directory { children, .. } = current {
                        current = children.get(subpath).with_context(|| {
                            format!("can't descend into {subpath}: doesn't exist in directory.")
                        })?;
                    } else {
                        bail!("can't descend into {subpath}; parent isn't a directory...");
                    }
                }
                current
            };

            let chunkref = match node {
                TreeNode::NormalFile { content, .. } => content,
                TreeNode::Directory { .. } => {
                    bail!("Can't extract `Directory` to stdout!");
                }
                TreeNode::SymbolicLink { .. } => {
                    bail!("Can't extract `SymbolicLink` to stdout!");
                }
                TreeNode::Deleted => {
                    bail!("Can't extract `Deleted` to stdout!");
                }
            };

            let extract_span = info_span!("extract_files");
            let stream = std::io::BufWriter::new(
                io_streams::StreamWriter::stdout().context("failed to open stdout")?,
            );
            extract::unpack_sync_stream(&pwc, *chunkref, stream)
                .instrument(extract_span)
                .await?;

            Arc::try_unwrap(pwc)
                .map_err(|_| eyre!("pwc still in use; can't close down gracefully"))?
                .close()
                .await?;
        }
        YamaCommand::Check {
            pointers,
            shallow,
            intensive,
        } => {
            if !pointers && !shallow && !intensive {
                bail!("Check level not chosen. Try -2");
            }
            if pointers {
                bail!("pointers check not implemented yet. Try -2");
            }

            let pile_connector_path = Path::new(".");
            let keyring = pre_open_keyring(&pile_connector_path).await?;
            let keyring = open_keyring_interactive(keyring).await?;

            let pwc = Arc::new(
                open_pile(
                    &pile_connector_path,
                    keyring,
                    LockKind::Shared,
                    format!("{} check", get_hostname()),
                )
                .await?,
            );
            update_cache(&pwc).await?;

            if shallow {
                check::check_pointers_point_to_indexed_chunks(&pwc)
                    .await
                    .context("shallow check failed")?;
            }

            if intensive {
                let all_bloblogs = pwc.localcache.read().await?.list_bloblogs().await?;

                check::check_all_chunks_in_bloblogs(&pwc, &all_bloblogs)
                    .await
                    .context("intensive check failed")?;
            }

            Arc::try_unwrap(pwc)
                .map_err(|_| eyre!("pwc still in use; can't close down gracefully"))?
                .close()
                .await?;
        }
        YamaCommand::Vacuum {
            all,
            merge,
            forget,
            repack,
            delete_unrefd_bloblogs,
        } => {
            let pile_connector_path = Path::new(".");
            let keyring = pre_open_keyring(&pile_connector_path).await?;
            let keyring = open_keyring_interactive(keyring).await?;

            // TODO figure out how we use the pendingexclusive thing again...
            let pwc = Arc::new(
                open_pile(
                    &pile_connector_path,
                    keyring,
                    LockKind::Exclusive,
                    format!("{} vacuum", get_hostname()),
                )
                .await?,
            );
            update_cache(&pwc).await?;

            if all || merge {
                let to_merge = vacuum::merge_indices::select_indices_for_merge(
                    &pwc,
                    MERGE_TARGET_SIZE,
                    MERGE_THRESHOLD_SIZE,
                )
                .await
                .context("failed to select indices for merge")?;
                vacuum::merge_indices::merge_indices(&pwc, to_merge)
                    .await
                    .context("failed to merge indices")?;
                update_cache(&pwc).await?;
            }

            if all || forget {
                // TODO: allow running on smaller sets of indices than all of them
                let all_indices = {
                    let mut cache_conn = pwc.localcache.read().await?;
                    cache_conn.list_indices().await?
                };

                let forgettable_chunks = find_forgettable_chunks(&pwc, all_indices.clone()).await?;
                info!("{} chunks can be forgotten", forgettable_chunks.len());
                forget_chunks(&pwc, all_indices, forgettable_chunks).await?;
                update_cache(&pwc).await?;
            }

            if all || repack {
                let bloblog_stats = get_bloblogs_stats(&pwc).await?;
                let to_repack = select_bloblogs_for_repack(bloblog_stats).await?;
                info!("{} repack groups to be processed.", to_repack.len());
                perform_repack(pwc.clone(), to_repack).await?;
                update_cache(&pwc).await?;
            }

            if all || delete_unrefd_bloblogs {
                todo!();
            }

            Arc::try_unwrap(pwc)
                .map_err(|_| eyre!("pwc still in use; can't close down gracefully"))?
                .close()
                .await?;
        }
        YamaCommand::Keyring(KeyringCommand::Inspect { file }) => {
            let keyring = pre_open_keyring_at_path(&file).await?;
            let _keyring = open_keyring_interactive(keyring).await?;

            todo!();
        }
        YamaCommand::Keyring(KeyringCommand::Create {
            new,
            from,
            with,
            no_password,
            weak,
        }) => {
            let mut keys_to_copy = BTreeSet::new();

            if weak {
                unimplemented!("can't generate weak keys yet: unimplemented.");
            }

            if new.exists() {
                bail!("{new:?} already exists; won't overwrite!");
            }

            for key in with.split(",") {
                if let Some((_, key_group)) = KEY_GROUPS.iter().find(|(k, _)| *k == key) {
                    keys_to_copy.extend(key_group.into_iter());
                } else if INDIVIDUAL_KEYS.contains(&key) {
                    keys_to_copy.insert(key);
                } else {
                    bail!("Not a known key or key group: {key:?}");
                }
            }

            let keyring = match &from {
                Some(path) => pre_open_keyring_at_path(path).await?,
                None => pre_open_keyring(Path::new(".")).await?,
            };
            let keyring = open_keyring_interactive(keyring).await?;

            let new_keyring = keyring
                .partial_copy(&keys_to_copy)
                .context("failed to make partial copy of keyring")?;

            let new_keyring_password = if no_password {
                warn!("Not setting a new keyring password. The new keyring will be unprotected.");
                None
            } else {
                println!("enter new keyring password:");
                let stdin = stdin();
                let mut stdin_br = BufReader::new(stdin);
                let mut line = String::new();
                stdin_br.read_line(&mut line).await?;
                Some(line.trim().to_owned())
            };

            let new_keyring_packed = pack_keyring(
                new_keyring,
                new_keyring_password.as_ref().map(|s| s.as_ref()),
            )?;

            tokio::fs::write(&new, &new_keyring_packed.into_byte_vec())
                .await
                .context("couldn't write to new keyring file")?;

            info!("new keyring created!");
        }
    }

    Ok(())
}

async fn set_up_connection(
    sftp: bool,
    s3: bool,
    local_dir: &Path,
) -> eyre::Result<PileConnectionScheme> {
    let stdin = stdin();
    let mut stdin_br = BufReader::new(stdin);
    let mut line = String::new();

    match (sftp, s3) {
        (true, true) => {
            bail!("Can only choose one of --sftp or --s3 (or local)!");
        }
        (true, false) => {
            // SFTP
            println!("Enter user@host for SFTP:");
            stdin_br.read_line(&mut line).await?;
            let user_at_host = line.trim().to_owned();
            line.clear();

            println!("Enter remote path (can be relative to user home):");
            stdin_br.read_line(&mut line).await?;
            let remote_path = line.trim().to_owned();
            line.clear();

            Ok(PileConnectionScheme::Sftp {
                user_at_host,
                directory: remote_path,
            })
        }
        (false, true) => {
            // S3
            todo!()
        }
        (false, false) => {
            // Local filesystem
            Ok(PileConnectionScheme::Local {
                directory: local_dir.to_owned(),
            })
        }
    }
}
