/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/

use std::collections::BTreeMap;
use std::io::SeekFrom;
use std::path::{Path, PathBuf};

use clap::Parser;
use dust_style_filetree_display::display::{draw_it, InitialDisplayData};

use dust_style_filetree_display::filter::AggregateData;
use dust_style_filetree_display::node::Node;
use dust_style_filetree_display::{get_height_of_terminal, get_width_of_terminal, init_color};
use eyre::{bail, Context, ContextCompat};
use patricia_tree::PatriciaMap;
use tokio::fs::OpenOptions;
use tokio::io::{AsyncReadExt, AsyncSeekExt, AsyncWriteExt};

use yama::scan;

use tracing_subscriber::layer::SubscriberExt;
use tracing_subscriber::util::SubscriberInitExt;

use yama_pile::tree::unpopulated::ScanEntry;

#[derive(Parser, Clone, Debug)]
pub enum YamaScanCommand {
    /// Add an entry to an ignore file
    #[command(alias = "i")]
    Ignore {
        /// What to ignore
        path: String,

        /// Don't anchor the match to this directory.
        #[arg(short = 'a')]
        unanchored: bool,
    },

    /// Show dust-style usage graph of the current directory, excluding excluded files.
    #[command(alias = "du")]
    Usage {
        /// Specify an ignore rule. Can use multiple times.
        #[arg(short = 'I', long = "ignore")]
        ignore: Vec<String>,
    },
}

#[tokio::main]
async fn main() -> eyre::Result<()> {
    tracing_subscriber::registry()
        .with(
            tracing_subscriber::EnvFilter::try_from_default_env()
                .unwrap_or_else(|_| "sqlx=warn,yama=debug,info".into()),
        )
        .with(tracing_subscriber::fmt::layer())
        .init();

    match YamaScanCommand::parse() {
        YamaScanCommand::Usage { ignore } => {
            let idd = InitialDisplayData {
                short_paths: true,
                is_reversed: false,
                colors_on: !init_color(false),
                by_filecount: false,
                is_screen_reader: false,
                iso: false,
            };

            let scan = scan::scan(Path::new("."), &ignore).context("Couldn't scan")?;
            let top_nodes = assemble_display_tree_from_scan_entries(scan)?.children;

            let root_display_node = dust_style_filetree_display::filter::get_biggest(
                top_nodes,
                AggregateData {
                    min_size: None,
                    only_dir: false,
                    only_file: false,
                    number_of_lines: get_height_of_terminal(),
                    depth: usize::MAX,
                    using_a_filter: false,
                },
            )
            .expect("no root?");

            draw_it(
                idd,
                false,
                get_width_of_terminal(),
                &root_display_node,
                false,
            )
        }
        YamaScanCommand::Ignore { path, unanchored } => {
            let mut oo = OpenOptions::new()
                .read(true)
                .write(true)
                .create(true)
                .truncate(false)
                .open(".yamaignore")
                .await
                .context("failed to open .yamaignore for r/w")?;
            let pos = oo.seek(SeekFrom::End(0)).await?;
            if pos > 1 {
                oo.seek(SeekFrom::End(-1)).await?;
                let last_byte = oo.read_u8().await?;
                if last_byte != b'\n' {
                    oo.write_u8(b'\n').await?;
                }
            }

            if unanchored {
                oo.write_all(format!("{}\n", path).as_bytes()).await?;
            } else {
                oo.write_all(format!("/{}\n", path).as_bytes()).await?;
            }

            oo.flush().await?;
            drop(oo);
        }
        _other => todo!(),
    }

    Ok(())
}

pub fn assemble_display_tree_from_scan_entries(scan: PatriciaMap<ScanEntry>) -> eyre::Result<Node> {
    let mut dirs: BTreeMap<String, BTreeMap<String, Node>> = BTreeMap::new();
    // special-case the root ("")
    dirs.insert(String::new(), BTreeMap::new());

    for (key, entry) in scan.into_iter() {
        let key_string = String::from_utf8(key).context("bad UTF-8 in PMap")?;
        let (parent_dir_name, child_name) =
            key_string.rsplit_once('/').unwrap_or(("", &key_string));
        match entry {
            ScanEntry::NormalFile { size, .. } => {
                // note: for the root, this inserts the root file entry as a child called "" within a fake root 'directory'.
                // That's fine. We'll patch this up later.
                dirs.get_mut(parent_dir_name)
                    .context("bad PMap: parent not seen first")?
                    .insert(
                        child_name.to_owned(),
                        Node {
                            name: PathBuf::from(&key_string),
                            size,
                            children: vec![],
                            inode_device: None,
                            depth: 0,
                        },
                    );
            }
            ScanEntry::Directory {
                ownership: _,
                permissions: _,
            } => {
                dirs.insert(key_string.clone(), BTreeMap::new());
                // note: for the root, this inserts the root directory entry as a child called "" within the root.
                // That's fine. We'll patch this up later.
                dirs.get_mut(parent_dir_name)
                    .context("bad PMap: parent not seen first")?
                    .insert(
                        child_name.to_owned(),
                        Node {
                            name: PathBuf::from(&key_string),
                            size: 4096,
                            children: vec![],
                            inode_device: None,
                            depth: 0,
                        },
                    );
            }
            ScanEntry::SymbolicLink {
                ownership: _,
                target: _,
            } => {
                // note: for the root, this inserts the root symlink entry as a child called "" within a fake root 'directory'.
                // That's fine. We'll patch this up later.
                dirs.get_mut(parent_dir_name)
                    .context("bad PMap: parent not seen first")?
                    .insert(
                        child_name.to_owned(),
                        Node {
                            name: PathBuf::from(&key_string),
                            size: 4096,
                            children: vec![],
                            inode_device: None,
                            depth: 0,
                        },
                    );
            }
        }
    }

    // Now roll up the directories. In Rustc v1.66 it'd be nice to use pop_last()...
    while let Some(last_key) = dirs.keys().last().cloned() {
        let mut last_children = dirs.remove(&last_key).unwrap();
        if last_key.is_empty() {
            assert!(
                dirs.is_empty(),
                "when pulling out root pseudo-dir, dirs must be empty for roll-up."
            );

            let mut real_root = last_children.remove("").unwrap();
            real_root.children = last_children.into_values().collect();
            real_root.size += real_root.children.iter().map(|c| c.size).sum::<u64>();
            return Ok(real_root);
        }

        // We want to roll up the directory last/key -> {child -> ...}
        // so last -> {key -> {child -> ...}}
        let (parent_dir, child_name) = last_key.rsplit_once('/').unwrap_or(("", &last_key));
        let parent = dirs
            .get_mut(parent_dir)
            .context("bad PMap? no parent in rollup")?;
        let child_in_parent = parent
            .get_mut(child_name)
            .context("dir child not populated")?;
        child_in_parent.children = last_children.into_values().collect();
        child_in_parent.size += child_in_parent.children.iter().map(|c| c.size).sum::<u64>();
    }

    bail!("no root found; bad PMap or bad roll-up???");
}
