use crate::pile_with_cache::PileWithCache;
use crate::retriever::decompressor::PipelineDecompressor;
use crate::retriever::{create_fixed_retriever, FileId, JobChunkReq, JobId, RetrieverResp};
use crate::PROGRESS_BAR_STYLE;
use eyre::{bail, ensure, eyre, Context, ContextCompat};
use flume::Receiver;
use indicatif::ProgressStyle;
use patricia_tree::PatriciaMap;
use std::cmp::Reverse;
use std::collections::{BTreeMap, BTreeSet};
use std::fs::Permissions;
use std::io::Write;
use std::os::unix::fs::PermissionsExt;
use std::path::{Path, PathBuf};
use std::sync::Arc;
use tokio::fs::OpenOptions;
use tokio::io::AsyncWriteExt;
use tokio::task::JoinSet;
use tracing::{info_span, Instrument, Span};
use tracing_indicatif::span_ext::IndicatifSpanExt;
use yama_midlevel_crypto::chunk_id::ChunkId;
use yama_pile::definitions::{BloblogId, RecursiveChunkRef};
use yama_pile::tree::unpopulated::ScanEntry;
use yama_pile::tree::{FilesystemPermissions, TreeNode};
use yama_wormfile::boxed::BoxedWormFileProvider;

#[derive(Clone, Debug, Default)]
pub struct FlattenedTree {
    pub files: PatriciaMap<(ScanEntry, RecursiveChunkRef)>,
    pub nonfiles: PatriciaMap<ScanEntry>,
}

pub fn flatten_treenode(root_node: &TreeNode) -> eyre::Result<FlattenedTree> {
    let mut flat = FlattenedTree::default();

    root_node.visit(
        &mut |node, path| {
            match node {
                TreeNode::NormalFile {
                    mtime,
                    ownership,
                    permissions,
                    size,
                    content,
                } => {
                    flat.files.insert(
                        path,
                        (
                            ScanEntry::NormalFile {
                                mtime: *mtime,
                                ownership: *ownership,
                                permissions: *permissions,
                                size: *size,
                            },
                            *content,
                        ),
                    );
                }
                TreeNode::Directory {
                    ownership,
                    permissions,
                    children: _,
                } => {
                    flat.nonfiles.insert(
                        path,
                        ScanEntry::Directory {
                            ownership: *ownership,
                            permissions: *permissions,
                        },
                    );
                }
                TreeNode::SymbolicLink { ownership, target } => {
                    flat.nonfiles.insert(
                        path,
                        ScanEntry::SymbolicLink {
                            ownership: *ownership,
                            target: target.clone(),
                        },
                    );
                }
                TreeNode::Deleted => {
                    bail!("unexpected TreeNode::Deleted in flatten_treenode");
                }
            }

            Ok(())
        },
        String::new(),
    )?;

    Ok(flat)
}

/// Create directories and symbolic links.
pub async fn unpack_nonfiles(
    root: &Path,
    nonfiles: &PatriciaMap<ScanEntry>,
    restore_ownership: bool,
    restore_permissions: bool,
) -> eyre::Result<()> {
    if restore_ownership {
        bail!("restoring ownership is not yet supported...");
    }
    for (rel_path, scan_entry) in nonfiles.iter() {
        let path = root
            .join(String::from_utf8(rel_path).context("nonfiles map contains non-string keys?")?);

        match scan_entry {
            ScanEntry::NormalFile { .. } => {
                bail!("found NormalFile in unpack_nonfiles()");
            }
            ScanEntry::Directory {
                ownership: _,
                permissions,
            } => {
                tokio::fs::create_dir(&path).await?;
                if restore_permissions {
                    tokio::fs::set_permissions(&path, Permissions::from_mode(permissions.mode))
                        .await?;
                }
            }
            ScanEntry::SymbolicLink {
                ownership: _,
                target,
            } => {
                tokio::fs::symlink(target, &path).await?;
            }
        }
    }
    Ok(())
}

// TODO(perf): move out file writes into separate tasks...
pub async fn unpack_files(
    pwc: &Arc<PileWithCache<BoxedWormFileProvider>>,
    root: &Path,
    files: &PatriciaMap<(ScanEntry, RecursiveChunkRef)>,
    restore_ownership: bool,
    restore_permissions: bool,
) -> eyre::Result<()> {
    if restore_ownership {
        bail!("restoring ownership is not yet supported...");
    }
    let expanded_chunkrefs = expand_chunkrefs(
        pwc,
        files
            .iter()
            .map(|(path_bytes, (scan_entry, rcr))| ((path_bytes, scan_entry), *rcr)),
    )
    .await?;

    let total_chunks = expanded_chunkrefs
        .iter()
        .map(|(_, cs)| cs.len() as u64)
        .sum::<u64>();
    let unpack_span = info_span!("unpack_files");

    async move {
        let unpack_span = Span::current();
        unpack_span.pb_set_style(&ProgressStyle::default_bar().template(
            PROGRESS_BAR_STYLE,
        ).unwrap());
        unpack_span.pb_set_message("unpack");
        unpack_span.pb_set_length(total_chunks);

        let mut join_set = JoinSet::new();

        let (file_part_retriever, mut jobs) =
            lookup_chunkrefs_and_create_retriever(pwc, expanded_chunkrefs).await?;
        let mut open_files = BTreeMap::new();

        loop {
            tokio::select! {
                Ok(next_part) = file_part_retriever.recv_async() => {
                    match next_part {
                        RetrieverResp::Blob { job, subjob, blob } => {
                            if subjob == 0 {
                                // eprintln!("subjob 0 for job {job:?}");
                                let (path_bytes, scan_entry) = jobs
                                    .remove(&job)
                                    .with_context(|| format!("bad job {job:?} to extract"))?;

                                let (permissions, _ownership) = if let ScanEntry::NormalFile {
                                    permissions,
                                    ownership,
                                    ..
                                } = scan_entry
                                {
                                    (permissions, ownership)
                                } else {
                                    bail!("not a Normal File in unpack_files()");
                                };

                                let path = root.join(String::from_utf8(path_bytes).context("bad utf-8 in PM")?);

                                let (tx, rx) = flume::bounded(16);

                                join_set.spawn(file_unpacker_writer(path, *permissions, restore_permissions, rx));
                                open_files.insert(job, tx);
                            }
                            open_files
                                .get_mut(&job)
                                .context("bad job to write file")?
                                .send_async(Some(blob))
                                .await
                                .map_err(|_| eyre!("file tx shutdown"))?;

                            unpack_span.pb_inc(1);
                        }
                        RetrieverResp::JobComplete(job) => {
                            open_files
                                .remove(&job)
                                .context("bad job to finish file")?
                                .send_async(None)
                                .await
                                .map_err(|_| eyre!("file tx shutdown"))?;
                        }
                    }
                },
                Some(join_result) = join_set.join_next() => {
                    join_result
                        .context("failed file unpacker writer (a)")?
                        .context("failed file unpacker writer (b)")?;
                },
                else => {
                    break;
                }
            }
        }

        // we should have already drained the join set, but check...
        assert!(join_set.join_next().await.is_none());

        if !open_files.is_empty() || !jobs.is_empty() {
            bail!("There were errors extracting.");
        }

        Ok(())
    }.instrument(unpack_span).await
}

pub async fn unpack_sync_stream(
    pwc: &Arc<PileWithCache<BoxedWormFileProvider>>,
    chunkref: RecursiveChunkRef,
    mut stream: impl Write,
) -> eyre::Result<()> {
    let expanded_chunkrefs = expand_chunkrefs(pwc, vec![((), chunkref)].into_iter()).await?;

    let total_chunks = expanded_chunkrefs
        .iter()
        .map(|(_, cs)| cs.len() as u64)
        .sum::<u64>();
    let unpack_span = info_span!("unpack_files");

    async move {
        let unpack_span = Span::current();
        unpack_span.pb_set_style(
            &ProgressStyle::default_bar()
                .template(PROGRESS_BAR_STYLE)
                .unwrap(),
        );
        unpack_span.pb_set_message("unpack");
        unpack_span.pb_set_length(total_chunks);

        let (file_part_retriever, _) =
            lookup_chunkrefs_and_create_retriever(pwc, expanded_chunkrefs).await?;
        let mut done = false;

        while let Ok(next_part) = file_part_retriever.recv_async().await {
            match next_part {
                RetrieverResp::Blob { blob, .. } => {
                    tokio::task::block_in_place(|| stream.write_all(&blob))
                        .context("Failed to write to output stream on Blob")?;

                    unpack_span.pb_inc(1);
                }
                RetrieverResp::JobComplete(_) => {
                    tokio::task::block_in_place(|| stream.flush())
                        .context("Failed to flush output stream on JobComplete")?;
                    done = true;
                }
            }
        }

        if !done {
            bail!("There were errors extracting.");
        }

        Ok(())
    }
    .instrument(unpack_span)
    .await
}

async fn file_unpacker_writer(
    path: PathBuf,
    permissions: FilesystemPermissions,
    restore_permissions: bool,
    rx: Receiver<Option<Vec<u8>>>,
) -> eyre::Result<()> {
    let mut oo = OpenOptions::new();
    oo.write(true).create_new(true);
    if restore_permissions {
        oo.mode(permissions.mode);
    };
    let mut file = oo
        .open(&path)
        .await
        .with_context(|| format!("can't create {path:?}"))?;

    loop {
        match rx.recv_async().await {
            Ok(Some(next_block)) => {
                file.write_all(&next_block).await?;
            }
            Ok(None) => {
                file.flush().await.context("failed to flush")?;
                return Ok(());
            }
            Err(_) => {
                bail!("rx for file unpacking into {path:?} disconnected unexpectedly");
            }
        }
    }
}

pub(crate) async fn expand_chunkrefs<T>(
    pwc: &Arc<PileWithCache<BoxedWormFileProvider>>,
    chunkrefs: impl Iterator<Item = (T, RecursiveChunkRef)>,
) -> eyre::Result<Vec<(T, Vec<ChunkId>)>> {
    let mut by_depth = BTreeMap::<Reverse<u32>, Vec<(T, Vec<ChunkId>)>>::new();
    for (t, rec) in chunkrefs {
        by_depth
            .entry(Reverse(rec.depth))
            .or_default()
            .push((t, vec![rec.chunk_id]));
    }

    while let Some(Reverse(next_depth)) = by_depth.keys().next().cloned() {
        let ts_and_chunks = by_depth.remove(&Reverse(next_depth)).unwrap();

        if next_depth == 0 {
            return Ok(ts_and_chunks);
        }

        let ec_span = info_span!("expand_chunkrefs");
        ec_span.pb_set_style(
            &ProgressStyle::default_bar()
                .template(PROGRESS_BAR_STYLE)
                .unwrap(),
        );
        ec_span.pb_set_length(
            ts_and_chunks
                .iter()
                .map(|(_, cs)| cs.len() as u64)
                .sum::<u64>(),
        );
        ec_span.pb_set_message(&format!("resolve (d={next_depth})"));
        let expanded_ts_and_chunks = expand_chunkrefs_one_layer(pwc, ts_and_chunks)
            .instrument(ec_span)
            .await?;
        by_depth
            .entry(Reverse(next_depth - 1))
            .or_default()
            .extend(expanded_ts_and_chunks);
    }

    Ok(Vec::new())
}

async fn lookup_chunkrefs_and_create_retriever<T>(
    pwc: &Arc<PileWithCache<BoxedWormFileProvider>>,
    input: Vec<(T, Vec<ChunkId>)>,
) -> eyre::Result<(Receiver<RetrieverResp>, BTreeMap<JobId, T>)> {
    let mut next_job_id = JobId(0);

    let chunks_to_lookup: BTreeSet<ChunkId> = input
        .iter()
        .flat_map(|(_t, chunkids)| chunkids)
        .copied()
        .collect();

    let looked_up_chunks = pwc
        .localcache
        .read()
        .await?
        .locate_chunks(&chunks_to_lookup)
        .await?;
    ensure!(
        chunks_to_lookup.len() == looked_up_chunks.len(),
        "chunks are missing"
    );

    let bloblog_ids: BTreeSet<BloblogId> = looked_up_chunks.values().map(|(bi, _)| *bi).collect();
    let num_bloblogs = bloblog_ids.len();
    let bloblog_to_file_ids: BTreeMap<BloblogId, FileId> = bloblog_ids
        .into_iter()
        .zip((0..num_bloblogs as u32).map(FileId))
        .collect();
    let files: BTreeMap<FileId, BloblogId> =
        bloblog_to_file_ids.iter().map(|(&k, &v)| (v, k)).collect();

    let mut out_by_job = BTreeMap::<JobId, T>::new();
    let mut jobs = BTreeMap::<JobId, Vec<JobChunkReq>>::new();
    for (t, chunks) in input {
        let job_id = next_job_id;
        next_job_id.0 += 1;
        out_by_job.insert(job_id, t);
        jobs.insert(
            job_id,
            chunks
                .into_iter()
                .map(|c| {
                    let (bloblog_id, blob_locator) = &looked_up_chunks[&c];
                    JobChunkReq {
                        file: bloblog_to_file_ids[bloblog_id],
                        offset: blob_locator.offset,
                        length: blob_locator.length,
                    }
                })
                .collect(),
        );
    }

    let retriever = create_fixed_retriever(pwc.clone(), jobs, files, 8)?;
    let retriever =
        PipelineDecompressor::start(pwc.pile.pile_config.zstd_dict.clone(), 2, retriever)?;
    Ok((retriever, out_by_job))
}

pub(crate) async fn expand_chunkrefs_one_layer<T>(
    pwc: &Arc<PileWithCache<BoxedWormFileProvider>>,
    input: Vec<(T, Vec<ChunkId>)>,
) -> eyre::Result<Vec<(T, Vec<ChunkId>)>> {
    let (retriever, jobs_to_ts) = lookup_chunkrefs_and_create_retriever(pwc, input).await?;

    let mut out_by_job: BTreeMap<JobId, (T, Vec<u8>)> = jobs_to_ts
        .into_iter()
        .map(|(ji, t)| (ji, (t, Vec::new())))
        .collect();

    let mut num_jobs_left = out_by_job.len();

    while let Ok(result) = retriever.recv_async().await {
        match result {
            RetrieverResp::Blob {
                job,
                subjob: _,
                blob,
            } => {
                out_by_job
                    .get_mut(&job)
                    .context("bad job gm")?
                    .1
                    .extend_from_slice(&blob);
                Span::current().pb_inc(1);
            }
            RetrieverResp::JobComplete(_) => {
                num_jobs_left -= 1;
            }
        }
    }

    ensure!(num_jobs_left == 0, "jobs left over, recovery not complete");

    out_by_job
        .into_values()
        .map(|(t, bytes)| {
            let chunk_ids = bytes
                .chunks(32)
                .map(|b| {
                    if b.len() != 32 {
                        bail!("wrong number of bytes for chunk refs");
                    }
                    let mut b32 = [0u8; 32];
                    b32.copy_from_slice(b);
                    Ok(ChunkId::from(b32))
                })
                .collect::<eyre::Result<_>>()?;
            Ok((t, chunk_ids))
        })
        .collect()
}
