use eyre::{bail, Context, ContextCompat};
use std::path::Path;
use tokio::io::AsyncWriteExt;
use yama_midlevel_crypto::byte_layer::{ByteLayer, CborSerde};
use yama_midlevel_crypto::key_derivation::KeyDerivationParameters;
use yama_midlevel_crypto::sym_box::SymBox;
use yama_pile::definitions::{PackedKeyring, PackedPileConfig, UnlockedOrLockedKeyring};
use yama_pile::keyring::Keyring;
use yama_pile::{DIR_BLOBLOGS, DIR_INDICES, DIR_LOCKS, FILE_MASTER_KEYRING, FILE_YAMA_CONFIG};
use yama_wormfile::paths::WormPath;
use yama_wormfile::{WormFileProvider, WormFileWriter};

/// Perform checks before we init a pile in the given directory.
pub async fn pre_init_check(path: &Path) -> eyre::Result<()> {
    if path.exists() && !path.is_dir() {
        bail!("{path:?} is not a directory; cannot create pile or connector here.");
    }

    for important_path in [
        "yama.toml",
        DIR_BLOBLOGS,
        DIR_LOCKS,
        FILE_YAMA_CONFIG,
        DIR_INDICES,
    ] {
        let important_path = path.join(important_path);
        if important_path.exists() {
            bail!("{important_path:?} already exists: can't create pile or connector here.");
        }
    }

    Ok(())
}

/// Perform checks before we init a pile in the given WormFileProvider.
pub async fn pre_init_check_wfp(wfp: &impl WormFileProvider) -> eyre::Result<()> {
    for important_path in ["yama.toml", FILE_YAMA_CONFIG] {
        let important_path = WormPath::new(important_path).unwrap();
        if wfp.is_regular_file(&important_path).await? {
            bail!("{important_path:?} already exists: can't create pile.");
        }
    }
    Ok(())
}

/// Initialise a pile.
/// Should be run after `pre_init_check_wfp`.
pub async fn init_pile(
    wfp: &impl WormFileProvider,
    pile_config: PackedPileConfig,
    master_keyring_copy: Option<PackedKeyring>,
) -> eyre::Result<()> {
    let mut writer = wfp.write().await?;
    writer.write_all(&pile_config.into_byte_vec()).await?;
    writer.flush().await?;
    writer
        .finalise(WormPath::new(FILE_YAMA_CONFIG).unwrap(), false)
        .await?;

    if let Some(master_keyring_copy) = master_keyring_copy {
        let mut writer = wfp.write().await?;
        writer
            .write_all(&master_keyring_copy.into_byte_vec())
            .await?;
        writer.flush().await?;
        writer
            .finalise(WormPath::new(FILE_MASTER_KEYRING).unwrap(), false)
            .await?;
    }

    Ok(())
}

// todo move this
pub fn pack_keyring(unpacked: Keyring, password: Option<&str>) -> eyre::Result<PackedKeyring> {
    let packed = if let Some(password) = password {
        let deriver = KeyDerivationParameters::new_recommended();
        let key = deriver
            .derive(password)
            .context("Failed to derive key from password")?;
        let symkey = key.into_symkey();

        let lockbox = SymBox::new(CborSerde::serialise(&unpacked).unwrap(), &symkey)
            .context("Failed to encrypt keyring")?;
        UnlockedOrLockedKeyring::Locked { deriver, lockbox }
    } else {
        UnlockedOrLockedKeyring::Unlocked(unpacked)
    };
    Ok(PackedKeyring::serialise(&packed).unwrap())
}
