use crate::pile_connector::PileConnectionScheme;
use crate::pile_with_cache::PileWithCache;
use eyre::{bail, Context, ContextCompat};
use std::borrow::Cow;
use std::collections::BTreeSet;
use std::hash::{Hash, Hasher};
use std::path::{Path, PathBuf};
use std::sync::Arc;
use tokio::io::{AsyncBufReadExt, BufReader};
use tracing::debug;
use twox_hash::XxHash64;
use yama_midlevel_crypto::byte_layer::ByteLayer;
use yama_pile::definitions::{IndexId, PackedKeyring, UnlockedOrLockedKeyring};
use yama_pile::keyring::Keyring;
use yama_pile::locks::LockKind;
use yama_pile::{Pile, FILE_YAMA_CONFIG, FILE_YAMA_CONNECTOR};
use yama_wormfile::boxed::BoxedWormFileProvider;

pub const KEYRING_LOOKUP_SEQ: [&'static str; 2] = ["access.yamakeyring", "master.yamakeyring"];

pub async fn pre_open_keyring(connector_in_dir: &Path) -> eyre::Result<UnlockedOrLockedKeyring> {
    for lookup in KEYRING_LOOKUP_SEQ {
        let keyring_path = connector_in_dir.join(lookup);
        if keyring_path.exists() {
            return pre_open_keyring_at_path(&keyring_path).await;
        }
    }

    bail!(
        "No keyring found in {:?}. Expected to see one at one of: {:?}",
        connector_in_dir,
        KEYRING_LOOKUP_SEQ
    );
}

pub async fn pre_open_keyring_at_path(
    keyring_path: &Path,
) -> eyre::Result<UnlockedOrLockedKeyring> {
    let packed_keyring_bytes = tokio::fs::read(&keyring_path)
        .await
        .with_context(|| format!("failed to read keyring file at {:?}", keyring_path))?;
    let packed_keyring = PackedKeyring::from_byte_vec(packed_keyring_bytes)
        .deserialise()
        .with_context(|| format!("failed to deserialise keyring file at {:?}", keyring_path))?;
    return Ok(packed_keyring);
}

pub async fn open_keyring_interactive(input: UnlockedOrLockedKeyring) -> eyre::Result<Keyring> {
    match input {
        UnlockedOrLockedKeyring::Locked { deriver, lockbox } => {
            println!("enter keyring password to decrypt:");
            let stdin = tokio::io::stdin();
            let mut stdin_br = BufReader::new(stdin);
            let mut line = String::new();
            stdin_br.read_line(&mut line).await?;

            let derived = deriver
                .derive(line.trim())
                .context("failed to derive key from password")?;
            let keyring = lockbox
                .unlock(&derived.into_symkey())
                .context("failed to decrypt keyring")?
                .deserialise()
                .context("failed to deserialise keyring")?;

            Ok(keyring)
        }
        UnlockedOrLockedKeyring::Unlocked(keyring) => Ok(keyring),
    }
}

pub async fn open_pile(
    connector_in_dir: &Path,
    keyring: Keyring,
    lock_kind: LockKind,
    lock_holder: String,
) -> eyre::Result<PileWithCache<BoxedWormFileProvider>> {
    let connection_scheme = if connector_in_dir.join(FILE_YAMA_CONFIG).exists() {
        PileConnectionScheme::Local {
            directory: connector_in_dir
                .canonicalize()
                .context("can't canonicalise local pile path")?
                .to_owned(),
        }
    } else if connector_in_dir.join(FILE_YAMA_CONNECTOR).exists() {
        let connector_toml = tokio::fs::read_to_string(&connector_in_dir.join(FILE_YAMA_CONNECTOR))
            .await
            .context("failed to read connector")?;
        let connector: PileConnectionScheme =
            toml::from_str(&connector_toml).context("failed to deserialise connector")?;
        connector
    } else {
        bail!("Neither yama.cfg nor yama.toml exists; doesn't look like a Yama pile or pile connector.");
    };

    // Calculate a prefix for the cache name
    let canon_connector_in_dir = connector_in_dir
        .canonicalize()
        .unwrap_or(connector_in_dir.to_owned());
    let cache_base_name = canon_connector_in_dir
        .file_name()
        .map(|f| f.to_string_lossy())
        .unwrap_or(Cow::Borrowed("_"));

    open_pile_using_connector(
        &connection_scheme,
        cache_base_name.as_ref(),
        keyring,
        lock_kind,
        lock_holder,
    )
    .await
}

pub async fn open_pile_using_connector(
    connection_scheme: &PileConnectionScheme,
    cache_base_name: &str,
    keyring: Keyring,
    lock_kind: LockKind,
    lock_holder: String,
) -> eyre::Result<PileWithCache<BoxedWormFileProvider>> {
    let wormfileprovider = Arc::new(connection_scheme.connect_to_wormfileprovider().await?);
    let pile = Pile::open_manual(wormfileprovider, lock_kind, lock_holder, keyring).await?;

    let cache_dir = appdirs::user_cache_dir(Some("yama"), None).expect("can't obtain cache dir!");

    let mut hasher = XxHash64::default();
    connection_scheme.hash(&mut hasher);
    let u64_hash = hasher.finish();

    let cache_key = format!("{}-{:016x}.sqlite3", cache_base_name, u64_hash);

    tokio::fs::create_dir_all(&cache_dir).await?;
    let cache_file = cache_dir.join(&cache_key);
    let localcache = yama_localcache::Store::new(&cache_file)
        .await
        .with_context(|| format!("failed to open local cache"))?;

    Ok(PileWithCache { pile, localcache })
}

pub async fn update_cache(pwc: &PileWithCache<BoxedWormFileProvider>) -> eyre::Result<()> {
    debug!("updating cache");
    let available_indices = pwc
        .pile
        .list_indices()
        .await
        .context("can't list available indices")?;
    let present_indices = pwc
        .localcache
        .read()
        .await?
        .list_indices()
        .await
        .context("can't list cached indices")?;

    let missing_indices: BTreeSet<IndexId> = available_indices
        .difference(&present_indices)
        .cloned()
        .collect();
    let deleted_indices: BTreeSet<IndexId> = present_indices
        .difference(&available_indices)
        .cloned()
        .collect();

    let mut downloaded_indices = Vec::new();

    debug!(
        "{} new indices to cache, {} deleted indices to back out",
        missing_indices.len(),
        deleted_indices.len()
    );
    for missing_index in missing_indices {
        debug!("downloading index {missing_index:?}");
        downloaded_indices.push((missing_index, pwc.pile.read_index(missing_index).await?));
    }

    let mut txn = pwc.localcache.write().await?;

    for deleted_index in deleted_indices {
        debug!("backing out index {deleted_index:?}");
        txn.delete_index(deleted_index).await?;
    }

    for (index_id, index) in downloaded_indices {
        debug!("applying index {index_id:?}");
        txn.apply_index(index_id, Arc::new(index)).await?;
    }

    debug!("finished updating cache");

    Ok(())
}

pub async fn open_lock_and_update_cache(
    pile_connector_path: PathBuf,
    lock_name: String,
) -> eyre::Result<Arc<PileWithCache<BoxedWormFileProvider>>> {
    let keyring = pre_open_keyring(&pile_connector_path).await?;
    let keyring = open_keyring_interactive(keyring).await?;

    let pwc = open_pile(&pile_connector_path, keyring, LockKind::Shared, lock_name).await?;
    update_cache(&pwc).await?;

    Ok(Arc::new(pwc))
}

pub async fn open_lock_and_update_cache_with_connector(
    pile_connection_scheme: &PileConnectionScheme,
    cache_base_name: &str,
    keyring_path: &Path,
    lock_name: String,
) -> eyre::Result<Arc<PileWithCache<BoxedWormFileProvider>>> {
    let keyring = pre_open_keyring_at_path(keyring_path).await?;
    let keyring = open_keyring_interactive(keyring).await?;

    let pwc = open_pile_using_connector(
        pile_connection_scheme,
        cache_base_name,
        keyring,
        LockKind::Shared,
        lock_name,
    )
    .await?;
    update_cache(&pwc).await?;

    Ok(Arc::new(pwc))
}
