use eyre::{bail, Context};
use serde::{Deserialize, Serialize};
use std::path::PathBuf;
use yama_wormfile::boxed::BoxedWormFileProvider;
use yama_wormfile_fs::LocalWormFilesystem;
use yama_wormfile_sftp::SftpWormFilesystem;

#[derive(Clone, Serialize, Deserialize, Debug, Hash)]
#[serde(tag = "scheme")]
pub enum PileConnectionScheme {
    #[serde(rename = "local")]
    Local { directory: PathBuf },
    #[serde(rename = "sftp")]
    Sftp {
        user_at_host: String,
        directory: String,
    },
    #[serde(rename = "s3")]
    S3 {},
}

impl PileConnectionScheme {
    pub async fn connect_to_wormfileprovider(&self) -> eyre::Result<BoxedWormFileProvider> {
        match self {
            PileConnectionScheme::Local { directory } => {
                if directory.exists() {
                    if !directory.is_dir() {
                        bail!("Can't connect to local pile {directory:?}: not a directory.");
                    }
                } else {
                    tokio::fs::create_dir(directory)
                        .await
                        .context("Can't connect to local pile; can't create directory.")?;
                }
                Ok(BoxedWormFileProvider::new(LocalWormFilesystem::new(
                    directory,
                )?))
            }
            PileConnectionScheme::Sftp {
                user_at_host,
                directory,
            } => Ok(BoxedWormFileProvider::new(
                SftpWormFilesystem::new(user_at_host, directory)
                    .await
                    .context("Failed SFTP connection")?,
            )),
            PileConnectionScheme::S3 { .. } => {
                //S3WormFilesystem::new()
                todo!()
            }
        }
    }
}

#[derive(Clone, Serialize, Deserialize)]
pub struct PileConnectionDetails {
    #[serde(flatten)]
    pub scheme: PileConnectionScheme,

    pub keyring: PathBuf,
}

impl PileConnectionDetails {
    pub async fn connect(self) -> eyre::Result<()> {
        // TODO
        Ok(())
    }
}
