use yama_localcache::Store;
use yama_pile::Pile;
use yama_pile::{pointers::Pointer, tree::IntegrationStats};
use yama_wormfile::WormFileProvider;

use crate::scan::integrate_uid_or_gid_map;
use async_recursion::async_recursion;
use eyre::{Context, ContextCompat};
use yama_pile::tree::integrate_node_in_place;

pub struct PileWithCache<WFP: WormFileProvider> {
    pub pile: Pile<WFP>,
    pub localcache: Store,
}

#[derive(Clone, Debug, Default)]
pub struct PointerIntegrationStatistics {
    pub integration: IntegrationStats,
    /// Number of pointers that were integrated to get here.
    pub depth: u64,
}

impl<WFP: WormFileProvider + 'static> PileWithCache<WFP> {
    pub async fn fully_integrate_pointer_in_place(
        &self,
        pointer: &mut Pointer,
        stats: &mut PointerIntegrationStatistics,
    ) -> eyre::Result<()> {
        if let Some(parent_pointer_name) = pointer.parent.as_ref() {
            let parent_pointer = self
                .read_pointer_fully_integrated(parent_pointer_name, stats)
                .await
                .with_context(|| {
                    format!("failed to read pointer {parent_pointer_name} whilst integrating")
                })?
                .with_context(|| {
                    format!("whilst integrating, expected pointer {parent_pointer_name} to exist")
                })?;

            stats.integration +=
                integrate_node_in_place(&mut pointer.root.node, &parent_pointer.root.node);
            integrate_uid_or_gid_map(&mut pointer.uids, &parent_pointer.uids);
            integrate_uid_or_gid_map(&mut pointer.gids, &parent_pointer.gids);
            pointer.parent = None;

            stats.depth += 1;
        }
        Ok(())
    }

    #[async_recursion]
    pub async fn read_pointer_fully_integrated(
        &self,
        name: &str,
        stats: &mut PointerIntegrationStatistics,
    ) -> eyre::Result<Option<Pointer>> {
        match self.pile.read_pointer(name).await? {
            Some(mut pointer) => {
                self.fully_integrate_pointer_in_place(&mut pointer, stats)
                    .await?;
                Ok(Some(pointer))
            }
            None => Ok(None),
        }
    }

    /// Gracefully close this pile + local cache.
    pub async fn close(self) -> eyre::Result<()> {
        self.pile.close().await?;
        Ok(())
    }
}
