// TODO The retriever should possibly live somewhere else

use crate::pile_with_cache::PileWithCache;
use eyre::{bail, ensure, eyre, ContextCompat};
use flume::{Receiver, Sender};
use std::collections::{BTreeMap, BTreeSet};
use std::pin::Pin;
use std::sync::Arc;
use tracing::error;
use yama_pile::bloblogs::BloblogReader;
use yama_pile::definitions::BloblogId;
use yama_wormfile::boxed::BoxedWormFileProvider;
use yama_wormfile::WormFileReader;

pub mod decompressor;

#[derive(Copy, Clone, Debug, Ord, PartialOrd, Eq, PartialEq)]
#[repr(transparent)]
pub struct JobId(pub u32);

#[derive(Copy, Clone, Debug, Ord, PartialOrd, Eq, PartialEq)]
#[repr(transparent)]
pub struct FileId(pub u32);

#[derive(Copy, Clone, Debug, Ord, PartialOrd, Eq, PartialEq)]
pub struct JobChunkReq {
    pub file: FileId,
    pub offset: u64,
    pub length: u64,
}

#[derive(Clone, Debug)]
pub enum RetrieverResp {
    Blob {
        job: JobId,
        subjob: u32,
        blob: Vec<u8>,
    },
    JobComplete(JobId),
}

#[derive(Copy, Clone, Debug, Ord, PartialOrd, Eq, PartialEq)]
struct FileRegionMarker {
    pub file: FileId,
    pub offset: u64,
    pub length: u64,
    pub job: JobId,
    pub subjob: u32,
}

#[derive(Debug)]
struct OpenFileState {
    pub req_tx: Sender<OpenFileReq>,
    pub offset: u64,
}

#[derive(Debug)]
struct OpenFileReq {
    pub offset: u64,
    pub length: u64,
    pub job: JobId,
    pub subjob: u32,
}

#[derive(Debug)]
struct ActiveJobState {
    pub subjobs: Vec<JobChunkReq>,
    pub next_subjob: u32,
    pub inflight: u32,
}

struct RetrieverInternals {
    pwc: Arc<PileWithCache<BoxedWormFileProvider>>,
    jobs_queue: BTreeMap<JobId, Vec<JobChunkReq>>,
    file_regions: BTreeSet<FileRegionMarker>,
    files: BTreeMap<FileId, BloblogId>,
    open_files: BTreeMap<FileId, OpenFileState>,
    results_tx: Sender<RetrieverResp>,
    active_jobs: BTreeMap<JobId, ActiveJobState>,
    ack_rx: Receiver<JobId>,

    self_ack_tx: Sender<JobId>,

    rec_active_jobs: u16,
}

pub fn create_fixed_retriever(
    pwc: Arc<PileWithCache<BoxedWormFileProvider>>,
    jobs: BTreeMap<JobId, Vec<JobChunkReq>>,
    files: BTreeMap<FileId, BloblogId>,
    rec_active_jobs: u16,
) -> eyre::Result<Receiver<RetrieverResp>> {
    let (results_tx, results_rx) = flume::bounded(4);
    let (self_ack_tx, ack_rx) = flume::bounded(4);
    let mut rint = RetrieverInternals {
        pwc,
        jobs_queue: Default::default(),
        file_regions: Default::default(),
        files,
        open_files: Default::default(),
        results_tx,
        active_jobs: Default::default(),
        ack_rx,
        self_ack_tx,
        rec_active_jobs,
    };
    for (job_id, job) in jobs {
        rint.set_up_job(job_id, job);
    }

    tokio::spawn(async_backtrace::frame!(async move {
        if let Err(e) = rint.retrieval_task().await {
            error!("retriever failed: {e:?}");
        }
    }));

    Ok(results_rx)
}

impl RetrieverInternals {
    fn set_up_job(&mut self, job_id: JobId, job: Vec<JobChunkReq>) {
        for (subjob, chunk) in job.iter().enumerate() {
            self.file_regions.insert(FileRegionMarker {
                file: chunk.file,
                offset: chunk.offset,
                length: chunk.length,
                job: job_id,
                subjob: subjob as u32,
            });
        }
        self.jobs_queue.insert(job_id, job);
        // eprintln!("new job {job_id:?}");
    }

    async fn file_request(
        open_file: &mut OpenFileState,
        job: JobId,
        subjob: u32,
        offset: u64,
        length: u64,
    ) -> eyre::Result<()> {
        // debug!("sched {job:?}->{subjob:?}");
        open_file
            .req_tx
            .send_async(OpenFileReq {
                offset,
                length,
                job,
                subjob,
            })
            .await
            .map_err(|_| eyre!("open file shut down :/"))?;
        open_file.offset = offset + length;
        Ok(())
    }

    async fn open_file(&mut self, file_id: FileId) -> eyre::Result<()> {
        assert!(!self.open_files.contains_key(&file_id));

        let &bloblog_id = self.files.get(&file_id).context("no file by that ID")?;
        let bloblog_reader = self.pwc.pile.read_bloblog(bloblog_id).await?;

        let completion_tx = self.results_tx.clone();
        let ack_tx = self.self_ack_tx.clone();

        let (subjob_tx, subjob_rx) = flume::unbounded();

        tokio::spawn(async_backtrace::frame!(async move {
            if let Err(e) =
                Self::reader_task(bloblog_reader, subjob_rx, ack_tx, completion_tx).await
            {
                error!("error in reader for {bloblog_id:?}: {e:?}");
            }
        }));

        self.open_files.insert(
            file_id,
            OpenFileState {
                req_tx: subjob_tx,
                offset: 0,
            },
        );

        Ok(())
    }

    async fn reader_task(
        mut bloblog_reader: BloblogReader<Pin<Box<dyn WormFileReader>>>,
        subjob_rx: Receiver<OpenFileReq>,
        ack_tx: Sender<JobId>,
        completion_tx: Sender<RetrieverResp>,
    ) -> eyre::Result<()> {
        while let Ok(next_job) = subjob_rx.recv_async().await {
            let mut blob = Vec::with_capacity(next_job.length as usize);
            bloblog_reader
                .read_to_buf(&mut blob, next_job.offset, next_job.length)
                .await?;

            completion_tx
                .send_async(RetrieverResp::Blob {
                    job: next_job.job,
                    subjob: next_job.subjob,
                    blob,
                })
                .await
                .expect("completions shut");

            // debug!("read,acking! {:?}", next_job);
            ack_tx.send_async(next_job.job).await?;
        }

        Ok(())
    }

    async fn retrieval_task(&mut self) -> eyre::Result<()> {
        // let mut icount = 0u64;
        loop {
            // icount += 1;
            // debug!("[{icount}] active jobs {:#?}", self.active_jobs);
            // debug!("[{icount}] open files {:#?}", self.open_files);

            // 0. Try to progress open jobs if they are staring right at the bytes they need...
            let mut to_remove = Vec::new();
            for (active_job_id, active_job) in &mut self.active_jobs {
                if active_job.inflight > 0 {
                    // skip if it's busy, we don't want to send blobs out of order...
                    continue;
                }
                if active_job.next_subjob as usize >= active_job.subjobs.len() {
                    // this job is to be finished!
                    to_remove.push(*active_job_id);
                    continue;
                }
                // Which file we are 'staring at' and requesting a run of chunks from
                let mut stare_file = None;
                'single_job_staring: loop {
                    let desired_blob = &active_job.subjobs[active_job.next_subjob as usize];
                    if stare_file.is_some() && stare_file != Some(desired_blob.file) {
                        // We have changed which file we are looking at, we can't request any further
                        // because they might get retrieved out of order.
                        break 'single_job_staring;
                    }
                    if let Some(open_file) = self.open_files.get_mut(&desired_blob.file) {
                        stare_file = Some(desired_blob.file);
                        if open_file.offset == desired_blob.offset {
                            Self::file_request(
                                open_file,
                                *active_job_id,
                                active_job.next_subjob,
                                desired_blob.offset,
                                desired_blob.length,
                            )
                            .await?;
                            ensure!(
                                self.file_regions.remove(&FileRegionMarker {
                                    file: desired_blob.file,
                                    offset: desired_blob.offset,
                                    length: desired_blob.length,
                                    job: *active_job_id,
                                    subjob: active_job.next_subjob,
                                }),
                                "no FRM to remove (0)"
                            );
                            active_job.next_subjob += 1;
                            active_job.inflight += 1;

                            if active_job.next_subjob as usize >= active_job.subjobs.len() {
                                // this job is to be finished!
                                break 'single_job_staring;
                            }
                        } else {
                            break 'single_job_staring;
                        }
                    } else {
                        break 'single_job_staring;
                    }
                }
            }

            for remove in to_remove {
                self.active_jobs.remove(&remove);
                // eprintln!("job complete {remove:?}");
                self.results_tx
                    .send_async(RetrieverResp::JobComplete(remove))
                    .await
                    .map_err(|_| eyre!("results_tx shutdown"))?;
            }

            // 1. Try to make the most of open files by opening new jobs in convenient locations.
            // Basically: if we have slots for new active jobs, then look to see if we have any
            // jobs that begin at the offset in question...
            if self.active_jobs.len() < self.rec_active_jobs as usize {
                let mut allowed = self.rec_active_jobs as usize - self.active_jobs.len();
                let mut progress = false;
                for (open_file_id, open_file_state) in &self.open_files {
                    for region in self.file_regions.range(
                        FileRegionMarker {
                            file: *open_file_id,
                            offset: open_file_state.offset,
                            length: 0,
                            job: JobId(0),
                            subjob: 0,
                        }..FileRegionMarker {
                            file: *open_file_id,
                            offset: open_file_state.offset + 1,
                            length: 0,
                            job: JobId(0),
                            subjob: 0,
                        },
                    ) {
                        if region.subjob != 0 {
                            // only accept this region if it's the start of a job
                            continue;
                        }
                        if let Some(subjobs) = self.jobs_queue.remove(&region.job) {
                            self.active_jobs.insert(
                                region.job,
                                ActiveJobState {
                                    subjobs,
                                    next_subjob: 0,
                                    inflight: 0,
                                },
                            );
                            allowed -= 1;
                            progress = true;
                            break;
                        }
                    }
                    if allowed == 0 {
                        break;
                    }
                }
                if progress {
                    continue;
                }
            }

            // 2. Try to progress active jobs, even if we have to open new files or seek.
            let mut files_to_open = BTreeSet::new();
            for (active_job_id, active_job) in &mut self.active_jobs {
                if active_job.inflight > 0 {
                    // skip if it's busy, we don't want to send blobs out of order...
                    continue;
                }

                let desired_blob = &active_job.subjobs[active_job.next_subjob as usize];
                if let Some(open_file) = self.open_files.get_mut(&desired_blob.file) {
                    Self::file_request(
                        open_file,
                        *active_job_id,
                        active_job.next_subjob,
                        desired_blob.offset,
                        desired_blob.length,
                    )
                    .await?;
                    ensure!(
                        self.file_regions.remove(&FileRegionMarker {
                            file: desired_blob.file,
                            offset: desired_blob.offset,
                            length: desired_blob.length,
                            job: *active_job_id,
                            subjob: active_job.next_subjob,
                        }),
                        "no FRM to remove (0)"
                    );
                    active_job.next_subjob += 1;
                    active_job.inflight += 1;
                } else {
                    // can't open immediately here due to mut borrow.
                    files_to_open.insert(desired_blob.file);
                }
            }

            if !files_to_open.is_empty() {
                for file in files_to_open {
                    self.open_file(file).await?;
                }
                continue;
            }

            // 3. Start new jobs
            if self.active_jobs.len() < self.rec_active_jobs as usize {
                // spawn a new job...
                if let Some(activate_job_id) = self.jobs_queue.keys().next().cloned() {
                    let new_job = self.jobs_queue.remove(&activate_job_id).unwrap();
                    self.active_jobs.insert(
                        activate_job_id,
                        ActiveJobState {
                            subjobs: new_job,
                            next_subjob: 0,
                            inflight: 0,
                        },
                    );
                    continue;
                }
            }

            // 4. Block for acks, unless there are no jobs in which case we should just finish!
            if self.active_jobs.is_empty() {
                break;
            }
            if let Ok(ack) = self.ack_rx.recv_async().await {
                if let Some(job) = self.active_jobs.get_mut(&ack) {
                    ensure!(job.inflight > 0, "recv'd ack for job that has 0 inflight");
                    job.inflight -= 1;
                } else {
                    bail!("recv'd ack for bad job {ack:?}");
                }
            }
        }
        Ok(())
    }
}
