use crate::retriever::{JobId, RetrieverResp};
use eyre::{bail, ensure, eyre, Context, ContextCompat};
use flume::{Receiver, Sender};
use std::collections::BTreeMap;
use std::sync::Arc;
use tracing::error;
use zstd::bulk::Decompressor;

pub const DECOMPRESS_CAPACITY: usize = 32 * 1024 * 1024;

pub struct PipelineDecompressor {
    rx: Receiver<RetrieverResp>,
    tx: Sender<RetrieverResp>,
    job_pool_tx: Sender<(JobId, u32, Vec<u8>)>,
    complete_rx: Receiver<(JobId, u32, Vec<u8>)>,

    processing: BTreeMap<JobId, JobState>,
}

struct JobState {
    pub next_submit_subjob: u32,
    pub next_enqueue_subjob: u32,
    pub queued: BTreeMap<u32, Vec<u8>>,
    pub complete: bool,
}

impl PipelineDecompressor {
    pub fn start(
        decom_dict: Option<Arc<Vec<u8>>>,
        num_decom: u8,
        rx: Receiver<RetrieverResp>,
    ) -> eyre::Result<Receiver<RetrieverResp>> {
        let (out_tx, out_rx) = flume::bounded(4);

        let (job_pool_tx, job_pool_rx) = flume::bounded(0);
        let (complete_tx, complete_rx) = flume::unbounded();

        for num in 0..num_decom {
            let decom_dict = decom_dict.clone();
            let job_pool_rx = job_pool_rx.clone();
            let complete_tx = complete_tx.clone();
            std::thread::Builder::new()
                .name(format!("decomp {num}"))
                .spawn(move || {
                    if let Err(err) =
                        Self::decompressor_worker(decom_dict, job_pool_rx, complete_tx)
                    {
                        error!("error in decompressor worker: {err:?}");
                    }
                })?;
        }

        let mut pd = PipelineDecompressor {
            rx,
            tx: out_tx,
            job_pool_tx,
            complete_rx,
            processing: Default::default(),
        };

        tokio::spawn(async_backtrace::frame!(async move {
            if let Err(e) = pd.decompressor_manager().await {
                eprintln!("pipeline decompressor error: {e:?}");
            }
        }));

        Ok(out_rx)
    }

    fn decompressor_worker(
        decom_dict: Option<Arc<Vec<u8>>>,
        job_pool_rx: Receiver<(JobId, u32, Vec<u8>)>,
        complete_tx: Sender<(JobId, u32, Vec<u8>)>,
    ) -> eyre::Result<()> {
        let mut decompressor = match decom_dict {
            Some(dict) => Decompressor::with_dictionary(&dict)?,
            None => Decompressor::new()?,
        };
        while let Ok((job_id, subjob, compressed_bytes)) = job_pool_rx.recv() {
            let decompressed_bytes = decompressor
                .decompress(&compressed_bytes, DECOMPRESS_CAPACITY)
                .context("failed to decompress")?;
            complete_tx
                .send((job_id, subjob, decompressed_bytes))
                .map_err(|_| eyre!("complete_tx shutdown"))?;
        }
        Ok(())
    }

    async fn decompressor_manager(&mut self) -> eyre::Result<()> {
        let mut incoming_open = true;
        loop {
            // Always process completed jobs as top priority
            while let Ok(completion) = self.complete_rx.try_recv() {
                self.handle_completion(completion).await?;
            }

            // Then it doesn't matter so much what we process after that
            tokio::select! {
                Ok(completion) = self.complete_rx.recv_async(), if !self.processing.is_empty() => {
                    self.handle_completion(completion).await?;
                },
                incoming_res = self.rx.recv_async(), if incoming_open => {
                    if let Ok(incoming) = incoming_res {
                        self.handle_incoming(incoming).await?;
                    } else {
                        incoming_open = false;
                    }
                }
                else => {
                    if !self.processing.is_empty() {
                        bail!("decompressor still procesing but shutting down?");
                    }
                    // eprintln!("D shutdown");
                    break Ok(());
                }
            };
        }
    }

    async fn handle_completion(
        &mut self,
        (job_id, subjob, decompressed): (JobId, u32, Vec<u8>),
    ) -> eyre::Result<()> {
        let state = self
            .processing
            .get_mut(&job_id)
            .context("bad job when recv complete decomp")?;
        ensure!(
            state.queued.insert(subjob, decompressed).is_none(),
            "overwrote decompressed block??"
        );
        while let Some(send_off) = state.queued.remove(&state.next_submit_subjob) {
            // eprintln!("D send off {job_id:?} {subjob}");
            self.tx
                .send(RetrieverResp::Blob {
                    job: job_id,
                    subjob: state.next_submit_subjob,
                    blob: send_off,
                })
                .map_err(|_| eyre!("tx shutdown"))?;
            state.next_submit_subjob += 1;
        }
        if state.queued.is_empty()
            && state.complete
            && state.next_submit_subjob == state.next_enqueue_subjob
        {
            // This job is done now
            // eprintln!("D jc {job_id:?}");
            self.tx
                .send(RetrieverResp::JobComplete(job_id))
                .map_err(|_| eyre!("tx shutdown"))?;
            self.processing.remove(&job_id);
        }
        Ok(())
    }

    async fn handle_incoming(&mut self, incoming: RetrieverResp) -> eyre::Result<()> {
        match incoming {
            RetrieverResp::Blob { job, subjob, blob } => {
                if subjob == 0 {
                    ensure!(
                        self.processing
                            .insert(
                                job,
                                JobState {
                                    next_submit_subjob: 0,
                                    next_enqueue_subjob: 0,
                                    queued: Default::default(),
                                    complete: false,
                                }
                            )
                            .is_none(),
                        "job was overwritten"
                    );
                }

                // debug!("blob {job:?} {subjob:?}");
                let state = self.processing.get_mut(&job).with_context(|| {
                    format!("bad job/not starting at 0 for job {job:?} (subjob={subjob:?})")
                })?;
                ensure!(
                    state.next_enqueue_subjob == subjob,
                    "out of order Blob commands"
                );
                state.next_enqueue_subjob += 1;
                self.job_pool_tx
                    .send_async((job, subjob, blob))
                    .await
                    .map_err(|_| eyre!("job_pool_tx shutdown"))?;
            }
            RetrieverResp::JobComplete(job) => {
                let state = self
                    .processing
                    .get_mut(&job)
                    .context("bad job to complete")?;
                state.complete = true;

                // debug!("complete {job:?}");

                let can_remove = state.next_submit_subjob == state.next_enqueue_subjob;

                if can_remove {
                    self.tx
                        .send(RetrieverResp::JobComplete(job))
                        .map_err(|_| eyre!("tx shutdown"))?;
                    self.processing.remove(&job);
                }
            }
        }
        Ok(())
    }
}
