use eyre::{bail, eyre, Context, ContextCompat};
use ignore::overrides::OverrideBuilder;
use ignore::WalkBuilder;
use patricia_tree::PatriciaMap;
use std::cmp::max;
use std::collections::{BTreeMap, BTreeSet};
use std::fs::{read_link, Metadata};
use std::io::ErrorKind;
use std::os::unix::fs::MetadataExt;
use std::path::{Component, Path};
use tracing::warn;
use yama_pile::definitions::RecursiveChunkRef;
use yama_pile::tree::unpopulated::ScanEntry;
use yama_pile::tree::{mtime_msec, FilesystemOwnership, FilesystemPermissions, TreeNode};

/// Given a node, recursively constructs a UID and GID lookup table based on THIS system's
/// users and groups.
///
/// Returns UIDs then GIDs.
pub fn create_uidgid_lookup_tables(
    node: &TreeNode,
) -> eyre::Result<(BTreeMap<u16, String>, BTreeMap<u16, String>)> {
    let mut uids = BTreeMap::<u16, String>::new();
    let mut gids = BTreeMap::<u16, String>::new();
    let mut used_uids = BTreeSet::new();
    let mut used_gids = BTreeSet::new();

    find_used_uidsgids(&node, &mut used_uids, &mut used_gids);

    for uid in used_uids {
        if let Some(user) = users::get_user_by_uid(uid.into()) {
            uids.insert(
                uid,
                user.name()
                    .to_str()
                    .ok_or(eyre!("uid leads to non-String name"))?
                    .to_owned(),
            );
        }
    }

    for gid in used_gids {
        if let Some(group) = users::get_group_by_gid(gid.into()) {
            gids.insert(
                gid,
                group
                    .name()
                    .to_str()
                    .ok_or(eyre!("gid leads to non-String name"))?
                    .to_owned(),
            );
        }
    }

    Ok((uids, gids))
}

fn find_used_uidsgids(node: &TreeNode, uids: &mut BTreeSet<u16>, gids: &mut BTreeSet<u16>) {
    match &node {
        TreeNode::NormalFile { ownership, .. }
        | TreeNode::Directory { ownership, .. }
        | TreeNode::SymbolicLink { ownership, .. } => {
            uids.insert(ownership.uid);
            gids.insert(ownership.gid);
        }
        TreeNode::Deleted => { /* nop */ }
    };

    if let TreeNode::Directory { children, .. } = &node {
        for (_name, child) in children {
            find_used_uidsgids(child, uids, gids);
        }
    }
}

/// Calculates the relative path.
///
/// Returns empty string if the paths are the same, otherwise it's a /-separated string.
/// The returned string is not allowed to contain any . or .. components.
pub fn relative_path(base: &Path, leaf: &Path) -> Option<String> {
    assert_eq!(std::path::MAIN_SEPARATOR, '/');

    let relative = leaf.strip_prefix(base).ok()?;

    if relative
        .components()
        .any(|c| c == Component::CurDir || c == Component::ParentDir || c == Component::RootDir)
    {
        return None;
    }

    relative.to_str().map(|s| s.to_owned())
}

/// Scans a directory tree.
///
/// Aborts if any errors (permission, bad .yamaignore files, etc) are encountered.
/// In the future, we possibly want to consider allowing
pub fn scan(root: &Path, ignores: &Vec<String>) -> eyre::Result<PatriciaMap<ScanEntry>> {
    let mut entries: PatriciaMap<ScanEntry> = PatriciaMap::new();

    if !root.is_dir() {
        let metadata = std::fs::symlink_metadata(root).context("reading metadata of root")?;
        entries.insert(
            "",
            scan_one_no_recurse(root, metadata)
                .context("failed to generate scan entry for root")?
                .context("root probably doesn't exist, or is ignored?")?,
        );
        return Ok(entries);
    }

    let mut walker = WalkBuilder::new(root);
    walker
        .standard_filters(false)
        .add_custom_ignore_filename(".yamaignore")
        .parents(false)
        .follow_links(false)
        .same_file_system(true);

    let mut overrides = OverrideBuilder::new(root);
    for ign in ignores {
        overrides
            .add(&("!".to_owned() + ign))
            .with_context(|| format!("failed to add ignore rule: {ign:?}"))?;
    }
    walker.overrides(
        overrides
            .build()
            .context("failed to create overrides with ignore entries")?,
    );

    // TODO(performance): We could use `WalkParallel` if we restructure this code.
    let walker = walker.build();

    for entry in walker {
        let entry = entry?;

        if !entry.path().starts_with(root) {
            bail!(
                "Scanned entry {:?} does not start with search path {:?}",
                entry.path(),
                root
            );
        }
        let rel_path = if let Some(rel_path) = relative_path(root, entry.path()) {
            rel_path
        } else {
            continue;
        };

        if !rel_path.is_empty() {
            let parent_relpath = rel_path
                .rsplit_once('/')
                .map(|(parent, _child)| parent)
                .unwrap_or("");
            assert!(
                entries.contains_key(parent_relpath),
                "have not scanned parent for {}",
                rel_path
            );
        }

        if let Some(single_scan) = scan_one_no_recurse(
            entry.path(),
            entry
                .metadata()
                .with_context(|| format!("Failed to read metadata for {:?}", rel_path))?,
        )
        .with_context(|| format!("Failed to scan {:?}", rel_path))?
        {
            entries.insert(&rel_path, single_scan);
        }
    }

    Ok(entries)
}

fn scan_one_no_recurse(path: &Path, metadata: Metadata) -> eyre::Result<Option<ScanEntry>> {
    let filetype = metadata.file_type();

    let ownership = FilesystemOwnership {
        uid: metadata.uid() as u16,
        gid: metadata.gid() as u16,
    };

    let permissions = FilesystemPermissions {
        mode: metadata.mode(),
    };

    if filetype.is_file() {
        // Leave an unpopulated file node. It's not my responsibility to chunk it right now.
        Ok(Some(ScanEntry::NormalFile {
            mtime: mtime_msec(&metadata),
            ownership,
            permissions,
            size: metadata.size(),
        }))
    } else if filetype.is_dir() {
        // TODO(perforance): this call is very likely wasteful
        let dir_read = path.read_dir();

        if let Err(e) = &dir_read {
            match e.kind() {
                ErrorKind::NotFound => {
                    warn!("vanished/: {:?}", path);
                    return Ok(None);
                }
                ErrorKind::PermissionDenied => {
                    warn!("permission denied/: {:?}", path);
                    return Ok(None);
                }
                _ => { /* nop */ }
            }
        }

        Ok(Some(ScanEntry::Directory {
            ownership,
            permissions,
        }))
    } else if filetype.is_symlink() {
        let target = read_link(path)?
            .to_str()
            .ok_or(eyre!("target path cannot be to_str()d"))?
            .to_owned();

        Ok(Some(ScanEntry::SymbolicLink { ownership, target }))
    } else {
        Ok(None)
    }
}

/// Given the parent pointer's root TreeNode and a scan entry map of the current pointer,
/// return a chunkings map prepopulated with the reusable entries.
/// Also returns a pruned and prepopulated copy of the scan entry map.
pub fn prepopulate_unmodified(
    parent_tree: &TreeNode,
    scan_entry_map: &PatriciaMap<ScanEntry>,
) -> (
    PatriciaMap<Option<(RecursiveChunkRef, u64)>>,
    PatriciaMap<ScanEntry>,
    PatriciaMap<ScanEntry>,
) {
    let mut reusable_chunkings = PatriciaMap::new();
    let mut prepopulated_scan_entry_map = PatriciaMap::new();
    let mut pruned_scan_entry_map = scan_entry_map.clone();
    parent_tree
        .visit(
            &mut |tree_node, path| {
                // TODO We should consider prepopulating symlinks and empty dirs too, if they're
                //      included in the parent.
                if let TreeNode::NormalFile {
                    mtime: prev_mtime,
                    ownership: prev_ownership,
                    permissions: prev_permissions,
                    size: prev_size,
                    content: prev_content,
                } = tree_node
                {
                    if let Some(ScanEntry::NormalFile {
                        mtime,
                        ownership,
                        permissions,
                        size,
                    }) = scan_entry_map.get(path)
                    {
                        if mtime == prev_mtime
                            && size == prev_size
                            && ownership == prev_ownership
                            && prev_permissions == permissions
                        {
                            // Nothing seems to have changed about this file, let's just reuse the `content` from last time.
                            reusable_chunkings.insert(path, Some((*prev_content, *size)));
                            prepopulated_scan_entry_map.insert(
                                path,
                                pruned_scan_entry_map.remove(path).expect("checked removal"),
                            );

                            // Pull out parent directories so our subset always contains the parents for their children.
                            for path_fragment in iterate_dirs_upwards(path.as_bytes()) {
                                if let Some(directory) = pruned_scan_entry_map.remove(path_fragment)
                                {
                                    prepopulated_scan_entry_map.insert(path_fragment, directory);
                                }
                            }
                        }
                    }
                }

                Ok(())
            },
            String::new(),
        )
        .expect("no reason to fail");
    (
        reusable_chunkings,
        pruned_scan_entry_map,
        prepopulated_scan_entry_map,
    )
}

pub fn integrate_uid_or_gid_map(new: &mut BTreeMap<u16, String>, old: &BTreeMap<u16, String>) {
    for (old_uid, old_user) in old {
        new.entry(*old_uid).or_insert_with(|| old_user.clone());
    }
}

/// Given a scan entry map, creates another one whose size is limited to not containing more than
/// the given number of file bytes to be stored.
/// There is one exception: if an individual file exceeds the max size, the returned scan entry map
/// may contain just that one file.
///
/// Useful for gradually starting backups without having to do the first in one shot.
pub fn limit_scan_entry_map_to_size(
    scan_entry_map: PatriciaMap<ScanEntry>,
    soft_max_size: u64,
) -> PatriciaMap<ScanEntry> {
    let mut accum_size: u64 = 0;
    let mut have_file = false;
    let mut result = PatriciaMap::new();
    // let mut removeds = PatriciaSet::new();
    let mut unincluded_directories = PatriciaMap::new();

    for (path_bytes, entry) in scan_entry_map.into_iter() {
        if accum_size >= soft_max_size {
            // we're already full!
            break;
        }
        let size_of_entry = match entry {
            ScanEntry::NormalFile { size, .. } => {
                // even zero-byte files are not for free, so don't let them be.
                max(size, 4096)
            }
            ScanEntry::Directory { .. } => {
                // slightly made up number, but typical size of an inode?
                4096
            }
            ScanEntry::SymbolicLink { .. } => {
                // slightly made up number, but typical size of an inode?
                4096
            }
        };

        let size_limit_would_be_respected = accum_size + size_of_entry <= soft_max_size;
        let this_is_the_only_file = !have_file && matches!(&entry, ScanEntry::NormalFile { .. });
        if size_limit_would_be_respected || this_is_the_only_file {
            have_file |= matches!(&entry, ScanEntry::NormalFile { .. });
            result.insert(&path_bytes, entry);
            accum_size += size_of_entry;

            // Pull out parent directories so our subset always contains the parents for their children.
            for path_fragment in iterate_dirs_upwards(&path_bytes) {
                if let Some(directory) = unincluded_directories.remove(path_fragment) {
                    result.insert(path_fragment, directory);
                    accum_size += 4096;
                }
            }
        } else if matches!(&entry, &ScanEntry::Directory { .. }) {
            // put the directory to one side in case we need it...
            unincluded_directories.insert(path_bytes, entry);
        } else {
            // removeds.insert(path_bytes);
        }
    }

    // for (key, _) in unincluded_directories {
    //     removeds.insert(key);
    // }

    // (result, removeds)
    result
}

/// Returns a list of all the parent paths of the given path (in bytes),
/// including the root, in order from leaf to root.
pub fn iterate_dirs_upwards(path_bytes: &[u8]) -> Vec<&[u8]> {
    let mut result = Vec::new();
    let mut path_fragment = &path_bytes[..];
    while let Some((index, _)) = path_fragment
        .iter()
        .enumerate()
        .rev()
        .find(|(_idx, char_byte)| **char_byte == b'/')
    {
        path_fragment = &path_bytes[0..index];
        result.push(path_fragment);
    }
    result.push(&path_bytes[0..0]);
    result
}

#[cfg(test)]
mod tests {
    use crate::scan::limit_scan_entry_map_to_size;
    use maplit::btreeset;
    use patricia_tree::PatriciaMap;
    use std::collections::BTreeSet;
    use yama_pile::tree::unpopulated::ScanEntry;
    use yama_pile::tree::{FilesystemOwnership, FilesystemPermissions};

    #[test]
    fn test_limit_scan_entry_map_to_size() {
        let mut orig = PatriciaMap::new();

        orig.insert(
            "somedir".as_bytes(),
            ScanEntry::Directory {
                ownership: FilesystemOwnership { uid: 0, gid: 0 },
                permissions: FilesystemPermissions { mode: 0 },
            },
        );
        orig.insert(
            "somedir/a_small_file".as_bytes(),
            ScanEntry::NormalFile {
                mtime: 0,
                ownership: FilesystemOwnership { uid: 0, gid: 0 },
                permissions: FilesystemPermissions { mode: 0 },
                size: 4,
            },
        );
        orig.insert(
            "somedir/somefile".as_bytes(),
            ScanEntry::NormalFile {
                mtime: 0,
                ownership: FilesystemOwnership { uid: 0, gid: 0 },
                permissions: FilesystemPermissions { mode: 0 },
                size: 8192,
            },
        );

        // 16k = 4k (dir) + 8k (somefile) + 4k (small file; minimum)
        assert_eq!(
            limit_scan_entry_map_to_size(orig.clone(), 16384)
                .keys()
                .collect::<BTreeSet<_>>(),
            btreeset! {
                b"somedir".to_vec(),
                b"somedir/a_small_file".to_vec(),
                b"somedir/somefile".to_vec(),
            }
        );

        // now we don't have room for the big file.
        assert_eq!(
            limit_scan_entry_map_to_size(orig.clone(), 16383)
                .keys()
                .collect::<BTreeSet<_>>(),
            btreeset! {
                b"somedir".to_vec(),
                b"somedir/a_small_file".to_vec(),
            }
        );

        // because we must always include at least one file so we make forward progress, it doesn't
        // matter that this violates the size limit.
        assert_eq!(
            limit_scan_entry_map_to_size(orig.clone(), 1)
                .keys()
                .collect::<BTreeSet<_>>(),
            btreeset! {
                b"somedir".to_vec(),
                b"somedir/a_small_file".to_vec(),
            }
        );
    }
}
