use crate::pile_with_cache::PileWithCache;
use dashmap::DashSet;
use eyre::{bail, eyre, Context};
use fastcdc::v2020::{FastCDC, StreamCDC};
use flume::{Receiver, RecvError, SendError, Sender};
use std::cmp::Reverse;
use std::collections::{BTreeMap, BTreeSet};
use std::fmt::Debug;
use std::fs::File;
use std::io;
use std::io::Read;
use std::path::{Path, PathBuf};
use std::pin::Pin;
use std::sync::Arc;
use std::thread::JoinHandle;
use tokio::runtime::Handle;
use tracing::{debug, error, info, info_span, warn};
use yama_localcache::StoreConnection;
use yama_midlevel_crypto::chunk_id::{ChunkId, ChunkIdKey};
use yama_pile::bloblogs::BloblogWriter;
use yama_pile::definitions::{BlobLocator, BloblogId, Index, IndexBloblogEntry, RecursiveChunkRef};
use yama_wormfile::boxed::BoxedWormFileProvider;
use yama_wormfile::WormFileWriter;
use zstd::bulk::Compressor;

pub const DESIRED_INDEX_SIZE_ENTRIES: usize = 32768;

// 256 kiB
pub const FASTCDC_MIN: u32 = 256 * 1024;
// 1 MiB
pub const FASTCDC_AVG: u32 = 1024 * 1024;
// 8 MiB
pub const FASTCDC_MAX: u32 = 8 * 1024 * 1024;

pub struct StoringState {
    /// A connection to the local cache for checking whether
    pub cache_conn: StoreConnection<false>,
    /// Set of unflushed chunks, not present in any index, which we can assume have been created in this session.
    pub new_unflushed_chunks: Arc<DashSet<ChunkId>>,
    /// New bloblogs that we have created but not yet written out indices for.
    pub new_bloblogs: Vec<(BloblogId, BTreeMap<ChunkId, BlobLocator>)>,

    pub pwc: Arc<PileWithCache<BoxedWormFileProvider>>,

    pub chunk_id_key: ChunkIdKey,

    pub compressor: zstd::bulk::Compressor<'static>,

    pub tokio_handle: Handle,
}

impl StoringState {
    pub async fn new(
        pwc: Arc<PileWithCache<BoxedWormFileProvider>>,
        new_unflushed_chunks: Arc<DashSet<ChunkId>>,
        tokio_handle: Handle,
    ) -> eyre::Result<Self> {
        let compressor = match pwc.pile.pile_config.zstd_dict.as_ref() {
            None => {
                Compressor::new(get_zstd_level()).context("can't create dictless compressor")?
            }
            Some(dict_bytes) => Compressor::with_dictionary(get_zstd_level(), dict_bytes)
                .context("can't create dictful compressor")?,
        };

        let chunk_id_key = pwc.pile.pile_config.chunk_id_key;
        Ok(StoringState {
            cache_conn: pwc.localcache.read().await?,
            new_unflushed_chunks,
            new_bloblogs: vec![],
            pwc,
            chunk_id_key,
            compressor,
            tokio_handle,
        })
    }
}

struct StoringIntermediate {
    /// New bloblogs that we have created but not yet written out indices for.
    pub new_bloblogs: Vec<(BloblogId, BTreeMap<ChunkId, BlobLocator>)>,
}

impl From<StoringState> for StoringIntermediate {
    fn from(ss: StoringState) -> Self {
        StoringIntermediate {
            new_bloblogs: ss.new_bloblogs,
        }
    }
}

#[derive(Default)]
pub struct StoringBloblogWriters {
    /// Bloblog writer for actual file contents (we try to keep file contents sequential in the
    /// common case)
    pub file_contents: Option<BloblogWriter<Pin<Box<dyn WormFileWriter>>>>,
    /// Bloblog writer for chunks of chunks
    pub metachunks: Option<BloblogWriter<Pin<Box<dyn WormFileWriter>>>>,
}

impl StoringBloblogWriters {
    pub async fn finish_bloblogs(&mut self, ss: &mut StoringState) -> eyre::Result<()> {
        if let Some(writer_to_finish) = self.file_contents.take() {
            let (_bloblog_path, bloblog_id, chunkmap) = writer_to_finish.finish().await?;
            ss.new_bloblogs.push((bloblog_id, chunkmap));
        }

        if let Some(writer_to_finish) = self.metachunks.take() {
            let (_bloblog_path, bloblog_id, chunkmap) = writer_to_finish.finish().await?;
            ss.new_bloblogs.push((bloblog_id, chunkmap));
        }
        Ok(())
    }
}

impl StoringState {
    /// Acquire a bloblog writer handle, reusing the existing one in the slot if suitable.
    #[async_backtrace::framed]
    async fn obtain_bloblog_writer<'a>(
        &mut self,
        slot: &'a mut Option<BloblogWriter<Pin<Box<dyn WormFileWriter>>>>,
    ) -> eyre::Result<&'a mut BloblogWriter<Pin<Box<dyn WormFileWriter>>>> {
        // if let Some(ref mut writer) = slot {
        //     if !writer.should_finish() {
        //         return Ok(writer);
        //     }
        // }
        // awkward avoidance of strange borrow issues that I don't fully grok
        if slot.as_ref().map(|w| w.should_finish()) == Some(false) {
            return Ok(slot.as_mut().unwrap());
        }

        if let Some(writer_to_finish) = slot.take() {
            let (_bloblog_path, bloblog_id, chunkmap) = writer_to_finish.finish().await?;
            self.new_bloblogs.push((bloblog_id, chunkmap));
        }

        *slot = Some(self.pwc.pile.create_bloblog().await?);
        Ok(slot.as_mut().unwrap())
    }

    /// For internal use only.
    fn process_chunk(
        &mut self,
        chunk_bytes: &[u8],
        result: &mut Vec<ChunkId>,
        slot: &mut Option<BloblogWriter<Pin<Box<dyn WormFileWriter>>>>,
    ) -> eyre::Result<()> {
        let chunk_id = ChunkId::compute(chunk_bytes, &self.chunk_id_key);
        result.push(chunk_id);
        let tokio_handle = self.tokio_handle.clone();
        let is_new = tokio_handle.block_on(async_backtrace::frame!(async {
            Ok::<bool, eyre::Report>(
                self.cache_conn.is_chunk_new(chunk_id).await?
                    && self.new_unflushed_chunks.insert(chunk_id),
            )
        }))?;

        if is_new {
            let compressed_bytes = self.compressor.compress(&chunk_bytes)?;

            tokio_handle.block_on(async_backtrace::frame!(async {
                let writer = self.obtain_bloblog_writer(slot).await?;
                writer.write_chunk(chunk_id, &compressed_bytes).await?;
                Ok::<(), eyre::Report>(())
            }))?;
        }

        Ok(())
    }

    fn store_full_slice_returning_chunks(
        &mut self,
        store_slice: &[u8],
        slot: &mut Option<BloblogWriter<Pin<Box<dyn WormFileWriter>>>>,
    ) -> eyre::Result<Vec<ChunkId>> {
        let mut result = Vec::new();

        for chunk in FastCDC::new(store_slice, FASTCDC_MIN, FASTCDC_AVG, FASTCDC_MAX) {
            let chunk_bytes = &store_slice[chunk.offset..chunk.offset + chunk.length];
            self.process_chunk(chunk_bytes, &mut result, slot)?
        }

        if result.is_empty() {
            self.process_chunk(&[], &mut result, slot)?;
        }

        Ok(result)
    }

    fn store_full_stream_returning_chunks(
        &mut self,
        store_stream: impl Read,
        slot: &mut Option<BloblogWriter<Pin<Box<dyn WormFileWriter>>>>,
    ) -> eyre::Result<(Vec<ChunkId>, u64)> {
        let mut stream_length = 0u64;
        let mut result = Vec::new();
        for chunk in StreamCDC::new(store_stream, FASTCDC_MIN, FASTCDC_AVG, FASTCDC_MAX) {
            let chunk = chunk.context("failed to read in for StreamCDC")?;
            let chunk_bytes = chunk.data.as_slice();
            stream_length += chunk_bytes.len() as u64;
            self.process_chunk(chunk_bytes, &mut result, slot)?;
        }

        if result.is_empty() {
            self.process_chunk(&[], &mut result, slot)?;
        }

        Ok((result, stream_length))
    }

    pub fn store_full_slice(
        &mut self,
        store_slice: &[u8],
        sbw: &mut StoringBloblogWriters,
    ) -> eyre::Result<RecursiveChunkRef> {
        // First calculate all the chunk IDs needed to be written here.
        let mut chunk_ids =
            self.store_full_slice_returning_chunks(store_slice, &mut sbw.file_contents)?;
        let mut depth = 0;

        // If we have the wrong number of chunks, we should chunk the chunk list...
        while chunk_ids.len() != 1 {
            let mut metachunks_list_bytes: Vec<u8> = Vec::with_capacity(chunk_ids.len() * 32);
            for chunk_id in chunk_ids {
                metachunks_list_bytes.extend_from_slice(&chunk_id.to_bytes());
            }

            // TODO It might be nice to store these in opposite order, so a read is a true sequential
            //      scan.
            //      i.e. (depth=3) (depth=2) (depth=1) (depth=0) ...
            chunk_ids = self
                .store_full_slice_returning_chunks(&metachunks_list_bytes, &mut sbw.metachunks)?;
            depth += 1;
        }

        Ok(RecursiveChunkRef {
            chunk_id: chunk_ids[0],
            depth,
        })
    }

    /// Stores a full stream (`Read`) and returns the recursive chunk ref plus the length of the
    /// stream.
    #[async_backtrace::framed]
    pub fn store_full_stream(
        &mut self,
        store_stream: impl Read,
        sbw: &mut StoringBloblogWriters,
    ) -> eyre::Result<(RecursiveChunkRef, u64)> {
        // First calculate all the chunk IDs needed to be written here.
        let (mut chunk_ids, stream_length) =
            self.store_full_stream_returning_chunks(store_stream, &mut sbw.file_contents)?;
        let mut depth = 0;

        // If we have the wrong number of chunks, we should chunk the chunk list...
        while chunk_ids.len() != 1 {
            let mut metachunks_list_bytes: Vec<u8> = Vec::with_capacity(chunk_ids.len() * 32);
            for chunk_id in chunk_ids {
                metachunks_list_bytes.extend_from_slice(&chunk_id.to_bytes());
            }

            // TODO It might be nice to store these in opposite order, so a read is a true sequential
            //      scan.
            //      i.e. (depth=3) (depth=2) (depth=1) (depth=0) ...
            chunk_ids = self
                .store_full_slice_returning_chunks(&metachunks_list_bytes, &mut sbw.metachunks)?;
            depth += 1;
        }

        Ok((
            RecursiveChunkRef {
                chunk_id: chunk_ids[0],
                depth,
            },
            stream_length,
        ))
    }
}

/// Stores a file, returning Ok(Some(...)) if fine, Ok(None) if the file doesn't exist (vanished)
/// or Err(...) for any other error.
///
/// WARNING! This memory-maps the file and should NOT be used on files that are being written to
/// by other applications. If the underlying data changes during storage, it can cause issues with
/// Zstd (and presumably can also cause the chunk hashes to be invalid).
///
/// Further, I have had issues with this seeming to 'use' a lot of memory. Whilst it should only
/// be virtual memory, for some reason it seems to cause swap to be used and it makes diagnosis
/// of REAL memory issues much harder.
/// For that reason it is hard to recommend this approach for now.
#[allow(dead_code)]
fn store_file_mmap_blocking(
    file_path: &Path,
    storing_state: &mut StoringState,
    sbw: &mut StoringBloblogWriters,
) -> eyre::Result<Option<(RecursiveChunkRef, u64)>> {
    let file = match File::open(file_path) {
        Ok(file) => file,
        Err(err) if err.kind() == io::ErrorKind::NotFound => {
            warn!("file vanished: {file_path:?}");
            return Ok(None);
        }
        Err(other) => {
            bail!("error storing {file_path:?}: {other:?}");
        }
    };
    let mapped = unsafe { memmap2::Mmap::map(&file) }?;
    let size_of_file = mapped.as_ref().len();
    let chunkref = storing_state.store_full_slice(mapped.as_ref(), sbw)?;
    Ok(Some((chunkref, size_of_file as u64)))
}

fn store_file_non_mmap_blocking(
    file_path: &Path,
    storing_state: &mut StoringState,
    sbw: &mut StoringBloblogWriters,
) -> eyre::Result<Option<(RecursiveChunkRef, u64)>> {
    let file = match File::open(file_path) {
        Ok(file) => file,
        Err(err) if err.kind() == io::ErrorKind::NotFound => {
            warn!("file vanished: {file_path:?}");
            return Ok(None);
        }
        Err(other) => {
            bail!("error storing {file_path:?}: {other:?}");
        }
    };
    storing_state.store_full_stream(file, sbw).map(Some)
}

pub struct StoragePipeline<JobName> {
    result_rx: Receiver<(JobName, Option<(RecursiveChunkRef, u64)>)>,
    join_set: Vec<JoinHandle<eyre::Result<StoringIntermediate>>>,
}

fn storage_pipeline_worker_blocking<JobName: Debug>(
    job_rx: Receiver<(JobName, PathBuf)>,
    result_tx: Sender<(JobName, Option<(RecursiveChunkRef, u64)>)>,
    mut storing_state: StoringState,
    tokio_handle: Handle,
) -> eyre::Result<StoringIntermediate> {
    let mut bloblog_writers = StoringBloblogWriters::default();

    debug!("SPW startup");

    while let Ok((job_id, file_path)) = job_rx.recv() {
        // TODO(span): is this correctly a child of the parent span?
        let span = info_span!("store_file", file=?file_path);
        let _span_entered = span.enter();

        let file_store_opt =
            store_file_non_mmap_blocking(&file_path, &mut storing_state, &mut bloblog_writers)
                .with_context(|| format!("failed to store {file_path:?}"))?;
        if let Err(SendError(to_be_sent)) = result_tx.send((job_id, file_store_opt)) {
            bail!("Can't return result for {to_be_sent:?} — result_tx shut down.");
        }
    }

    debug!("finishing bloblogs");
    tokio_handle.block_on(bloblog_writers.finish_bloblogs(&mut storing_state))?;
    debug!("finished bloblogs!");

    Ok(StoringIntermediate::from(storing_state))
}

fn get_zstd_level() -> i32 {
    // TODO Do something more proper
    if let Ok(var) = std::env::var("YAMA_HACK_ZSTD_LEVEL") {
        if let Ok(level) = var.parse() {
            info!("YAMA_HACK_ZSTD_LEVEL: using {level}");
            return level;
        } else {
            error!("YAMA_HACK_ZSTD_LEVEL was not set to a valid i32: {var:?}")
        }
    }

    return 16;
}

impl<JobName: Debug + Send + 'static> StoragePipeline<JobName> {
    #[async_backtrace::framed]
    pub async fn launch_new(
        workers: u32,
        pwc: Arc<PileWithCache<BoxedWormFileProvider>>,
        new_unflushed_chunks: Arc<DashSet<ChunkId>>,
    ) -> eyre::Result<(StoragePipeline<JobName>, Sender<(JobName, PathBuf)>)> {
        let (job_tx, job_rx) = flume::bounded(16);
        let (result_tx, result_rx) = flume::bounded(4);

        let mut join_set = Vec::new();
        for spw_num in 0..workers {
            let job_rx = job_rx.clone();
            let result_tx = result_tx.clone();

            let tokio_handle = Handle::current();

            let storing_state = StoringState::new(
                pwc.clone(),
                new_unflushed_chunks.clone(),
                tokio_handle.clone(),
            )
            .await
            .context("failed to create storing state")?;
            // make a logging span for the Storage Pipeline Workers
            let spw_span = info_span!("spw", n = spw_num);

            let thread = std::thread::Builder::new()
                .name(format!("spw-{spw_num}"))
                .spawn(move || {
                    let _spw_span_entered = spw_span.enter();
                    let result = storage_pipeline_worker_blocking(
                        job_rx,
                        result_tx,
                        storing_state,
                        tokio_handle,
                    );
                    if let Err(ref err) = result {
                        error!("Error in SPW {err:?}");
                    }
                    result
                })
                .expect("failed to spawn SPW thread!");

            join_set.push(thread);
        }

        Ok((
            StoragePipeline {
                result_rx,
                join_set,
            },
            job_tx,
        ))
    }

    #[inline]
    pub async fn next_result(
        &self,
    ) -> Result<(JobName, Option<(RecursiveChunkRef, u64)>), RecvError> {
        self.result_rx.recv_async().await
    }

    /// Must be sure that all results have been collected first.
    #[async_backtrace::framed]
    pub async fn finish_into_chunkmaps(
        mut self,
    ) -> eyre::Result<BTreeMap<BloblogId, IndexBloblogEntry>> {
        if let Ok(msg) = self.result_rx.recv_async().await {
            bail!("Haven't processed all results yet! {msg:?}");
        }

        let mut chunkmap = BTreeMap::new();

        while let Some(thread) = self.join_set.pop() {
            // TODO(blocking on async thread): do this differently.
            let join_resres = thread.join().map_err(|panic_err| eyre!("{panic_err:?}"));
            chunkmap.extend(join_resres??.new_bloblogs.into_iter().map(|(k, nb)| {
                (
                    k,
                    IndexBloblogEntry {
                        chunks: nb,
                        forgotten_bytes: 0,
                    },
                )
            }));
        }

        Ok(chunkmap)
    }
}

fn assemble_indices(chunkmap: BTreeMap<BloblogId, IndexBloblogEntry>) -> Vec<Index> {
    let mut sorted_map = BTreeMap::new();
    for (idx, chunkmap) in chunkmap.into_iter().enumerate() {
        let size_of_chunkmap = chunkmap.1.chunks.len() + 1;
        sorted_map.insert(Reverse((size_of_chunkmap, idx)), chunkmap);
    }

    let mut indices = Vec::new();

    while let Some((Reverse((size, _)), (bloblog_id, bloblog_chunks))) = sorted_map.pop_first() {
        let mut new_index_contents = BTreeMap::new();
        new_index_contents.insert(bloblog_id, bloblog_chunks);
        let mut new_index_size_so_far = size;

        while new_index_size_so_far < DESIRED_INDEX_SIZE_ENTRIES && !sorted_map.is_empty() {
            if let Some((k, _)) = sorted_map
                .range(
                    Reverse((
                        DESIRED_INDEX_SIZE_ENTRIES - new_index_size_so_far,
                        usize::MAX,
                    ))..,
                )
                .next()
            {
                let k = k.clone();
                let (Reverse((add_size, _)), (bloblog_id, bloblog_chunks)) =
                    sorted_map.remove_entry(&k).unwrap();
                new_index_size_so_far += add_size;
                new_index_contents.insert(bloblog_id, bloblog_chunks);
            } else {
                break;
            }
        }

        indices.push(Index {
            supersedes: BTreeSet::new(),
            bloblogs: new_index_contents,
        });
    }

    indices
}

async fn write_indices(
    pwc: &PileWithCache<BoxedWormFileProvider>,
    indices: Vec<Index>,
) -> eyre::Result<()> {
    for index in indices {
        let index_id = pwc.pile.create_index(&index).await?;
        if !pwc
            .localcache
            .write()
            .await?
            .apply_index(index_id, Arc::new(index))
            .await?
        {
            error!("freshly-created index wasn't new. This is suspicious.");
        };
    }

    Ok(())
}

#[async_backtrace::framed]
pub async fn assemble_and_write_indices(
    pwc: &PileWithCache<BoxedWormFileProvider>,
    chunkmap: BTreeMap<BloblogId, IndexBloblogEntry>,
) -> eyre::Result<()> {
    let indices = assemble_indices(chunkmap);
    write_indices(pwc, indices).await
}
