use crate::extract::expand_chunkrefs_one_layer;
use crate::pile_with_cache::PileWithCache;
use eyre::{bail, ensure, Context, ContextCompat};
use std::collections::{BTreeMap, BTreeSet};
use std::sync::Arc;
use tracing::info;
use yama_midlevel_crypto::chunk_id::ChunkId;
use yama_pile::definitions::IndexId;
use yama_pile::tree::TreeNode;
use yama_wormfile::boxed::BoxedWormFileProvider;

pub async fn find_forgettable_chunks(
    pwc: &Arc<PileWithCache<BoxedWormFileProvider>>,
    indices: BTreeSet<IndexId>,
) -> eyre::Result<BTreeSet<ChunkId>> {
    let mut unseen_chunk_ids = BTreeSet::new();

    // Find all chunks in the given indices
    {
        let mut cache_conn = pwc.localcache.read().await?;
        for index_id in &indices {
            unseen_chunk_ids.extend(cache_conn.list_chunks_in_index(*index_id).await?);
        }
    };

    let chunks_to_scan = prepare_chunkrefs_to_scan(pwc).await?;
    scan_chunks(pwc, &mut unseen_chunk_ids, chunks_to_scan)
        .await
        .context("failed to do a sweep")?;

    Ok(unseen_chunk_ids)
}

async fn prepare_chunkrefs_to_scan(
    pwc: &Arc<PileWithCache<BoxedWormFileProvider>>,
) -> eyre::Result<BTreeMap<u32, BTreeSet<ChunkId>>> {
    let pointer_names = pwc
        .pile
        .list_pointers()
        .await
        .context("failed to list pointers")?;
    let mut chunks_to_scan_by_depth: BTreeMap<u32, BTreeSet<ChunkId>> = BTreeMap::new();

    for pointer_name in &pointer_names {
        let pointer = pwc
            .pile
            .read_pointer(pointer_name)
            .await?
            .context("pointer vanished")?;
        if let Some(parent_name) = pointer.parent {
            if !pointer_names.contains(pointer_name) {
                bail!("{parent_name:?}, the parent of {pointer_name:?}, does not exist");
            }
        }

        pointer
            .root
            .node
            .visit(
                &mut |node, _| {
                    if let TreeNode::NormalFile { content, .. } = node {
                        chunks_to_scan_by_depth
                            .entry(content.depth)
                            .or_default()
                            .insert(content.chunk_id);
                    }
                    Ok(())
                },
                String::new(),
            )
            .unwrap();
    }

    Ok(chunks_to_scan_by_depth)
}

/// Scans the recursive chunkrefs that are passed in, ticking off chunks from the `unseen` set as
/// we go.
async fn scan_chunks(
    pwc: &Arc<PileWithCache<BoxedWormFileProvider>>,
    unseen: &mut BTreeSet<ChunkId>,
    chunks_to_scan_by_depth: BTreeMap<u32, BTreeSet<ChunkId>>,
) -> eyre::Result<()> {
    let mut to_scan: Vec<(u32, Vec<ChunkId>)> = chunks_to_scan_by_depth
        .into_iter()
        .flat_map(|(depth, chunkset)| {
            chunkset
                .into_iter()
                .map(move |chunk_id| (depth, vec![chunk_id]))
        })
        .collect();

    while !to_scan.is_empty() {
        // Mark as seen.
        for (_, chunk_ids) in &to_scan {
            for chunk_id in chunk_ids {
                unseen.remove(chunk_id);
            }
        }

        // Don't descend further into zero-depth elements.
        to_scan = to_scan
            .into_iter()
            .filter(|(depth, _)| *depth > 0)
            .collect();

        // Decrement depth counters.
        to_scan = expand_chunkrefs_one_layer(pwc, to_scan)
            .await?
            .into_iter()
            .map(|(old_depth, chunkids)| (old_depth - 1, chunkids))
            .collect();
    }

    Ok(())
}

pub async fn forget_chunks(
    pwc: &Arc<PileWithCache<BoxedWormFileProvider>>,
    indices: BTreeSet<IndexId>,
    forgettable: BTreeSet<ChunkId>,
) -> eyre::Result<()> {
    let mut indices_to_rewrite = Vec::new();
    // First do a cache-only check to see which indices need rewriting.
    {
        let mut cache_conn = pwc.localcache.read().await?;
        for index_id in &indices {
            let chunks_in_this_index = cache_conn.list_chunks_in_index(*index_id).await?;
            if !chunks_in_this_index.is_disjoint(&forgettable) {
                indices_to_rewrite.push(index_id);
            }
        }
    }

    info!(
        "{} indices to rewrite in order to forget chunks",
        indices_to_rewrite.len()
    );

    // Go through each index and clean out whatever needs forgetting (then re-create the index and
    // remove the old one).
    for index_id in indices_to_rewrite {
        let mut index = pwc.pile.read_index(*index_id).await?;
        let mut changed = false;
        for bloblog_entry in index.bloblogs.values_mut() {
            let removable: Vec<ChunkId> = bloblog_entry
                .chunks
                .keys()
                .filter(|ci| forgettable.contains(ci))
                .cloned()
                .collect();
            changed |= !removable.is_empty();
            for chunk_id in removable {
                bloblog_entry.forgotten_bytes +=
                    bloblog_entry.chunks.remove(&chunk_id).unwrap().length;
            }
        }

        ensure!(changed, "no change to index {index_id:?}");

        index.supersedes.clear();
        index.supersedes.insert(*index_id);

        // TODO APPLY THE NEW INDEX DIRECTLY (how do we do that again?)
        let new_index_id = pwc.pile.create_index(&index).await?;
        ensure!(new_index_id != *index_id, "index ID bounce");
        pwc.pile.delete_index_dangerous_exclusive(*index_id).await?;
    }

    Ok(())
}
