use crate::pile_with_cache::PileWithCache;
use eyre::{bail, Context};
use std::collections::btree_map::Entry;
use std::collections::BTreeSet;
use std::sync::Arc;
use tracing::{debug, warn};
use yama_pile::definitions::{Index, IndexId};
use yama_wormfile::boxed::BoxedWormFileProvider;

pub const MERGE_THRESHOLD_SIZE: u32 = 2 * 1024 * 1024;
pub const MERGE_TARGET_SIZE: u32 = 16 * 1024 * 1024;

/// Selects indices for merge.
///
/// Criteria:
/// - size is less than the `threshold_size`
/// - (FUTURE; TODO) two indices that cover the same bloblog should be merged
pub async fn select_indices_for_merge(
    pwc: &Arc<PileWithCache<BoxedWormFileProvider>>,
    target_size: u32,
    threshold_size: u32,
) -> eyre::Result<Vec<BTreeSet<IndexId>>> {
    let mut result = Vec::new();
    let mut mergeable_indices: BTreeSet<(u64, IndexId)> = pwc
        .pile
        .list_indices_with_meta()
        .await?
        .into_iter()
        .filter(|(_, meta)| meta.file_size < threshold_size as u64)
        .map(|(index_id, meta)| (meta.file_size, index_id))
        .collect();

    while mergeable_indices.len() >= 2 {
        let mut merge_set = BTreeSet::new();
        let mut merge_size = 0u64;

        let (first_size, first_index) = mergeable_indices.pop_first().unwrap();
        merge_size += first_size;
        merge_set.insert(first_index);

        while let Some((size, index)) = mergeable_indices.first() {
            if merge_size + *size < target_size as u64 {
                merge_size += *size;
                merge_set.insert(*index);
                mergeable_indices.pop_first();
            } else {
                break;
            }
        }

        if merge_set.len() > 1 {
            result.push(merge_set);
        }
    }

    Ok(result)
}

/// Merges some indices, deleting them in the process.
/// Requires exclusive lock.
/// (Note: in the future we could only supersede the indices, which only needs a shared lock.
/// However you need an exclusive lock to eventually delete superseded indices...).
pub async fn merge_indices(
    pwc: &Arc<PileWithCache<BoxedWormFileProvider>>,
    merge_sets: Vec<BTreeSet<IndexId>>,
) -> eyre::Result<()> {
    for merge_set in merge_sets {
        let mut final_index = Index {
            supersedes: merge_set.clone(),
            bloblogs: Default::default(),
        };

        for index_id in &merge_set {
            let index_being_subsumed = pwc.pile.read_index(*index_id).await?;
            // TODO: do we need to worry about the 'supersedes' property on the index here?
            //       I think not, or at least not if the superseded indices don't exist,
            //       but worth thinking about in the future if we don't immediately delete
            //       superseded indices...
            for (bloblog_id, bloblog_entry) in index_being_subsumed.bloblogs {
                match final_index.bloblogs.entry(bloblog_id) {
                    Entry::Vacant(ve) => {
                        ve.insert(bloblog_entry);
                    }
                    Entry::Occupied(mut oe) => {
                        let new_entry = oe.get_mut();
                        let (existing_chunks, new_chunks): (Vec<_>, Vec<_>) = bloblog_entry
                            .chunks
                            .into_iter()
                            .partition(|(chunk_id, _)| new_entry.chunks.contains_key(chunk_id));
                        for (chunk_id, locator) in new_chunks {
                            // Subtract from the forgotten byte count, since this may be us re-remembering bytes out of safety...
                            new_entry.forgotten_bytes =
                                new_entry.forgotten_bytes.saturating_sub(locator.length);
                            let is_new = new_entry.chunks.insert(chunk_id, locator).is_none();
                            assert!(is_new);
                        }
                        for (chunk_id, locator) in existing_chunks {
                            if &new_entry.chunks[&chunk_id] != &locator {
                                bail!("Attempted to merge indices that disagree about {bloblog_id:?}/{chunk_id:?}");
                            }
                        }
                    }
                }
            }
        }

        let merged_index_id = pwc
            .pile
            .create_index(&final_index)
            .await
            .context("failed to create merged index")?;

        if merge_set.contains(&merged_index_id) {
            // I don't see how this could be possible, but let's avoid deleting the new index if it somehow is a merge of itself...
            warn!("strange: created index ID is one of its own merges...");
            continue;
        }

        debug!("merged indices {merge_set:?} into {merged_index_id:?}; deleting mergees");
        for index_to_delete in merge_set {
            pwc.pile
                .delete_index_dangerous_exclusive(index_to_delete)
                .await?;
        }
    }
    Ok(())
}
