use crate::pile_with_cache::PileWithCache;
use crate::storing::assemble_and_write_indices;
use eyre::ContextCompat;
use std::collections::btree_map::Entry;
use std::collections::{BTreeMap, BTreeSet};
use std::sync::Arc;
use yama_localcache::BloblogStats;
use yama_midlevel_crypto::chunk_id::ChunkId;
use yama_pile::definitions::{BloblogId, IndexBloblogEntry};
use yama_wormfile::boxed::BoxedWormFileProvider;

/// Repack bloblogs that have this much forgotten space in them.
pub const REPACK_BLOBLOGS_TO_RECLAIM_SPACE_BYTES: u64 = 32 * 1024 * 1024;

/// Defines what a 'small bloblog' is (one that is below a certain size, excluding forgotten bytes).
pub const SMALL_BLOBLOG_THRESHOLD: u64 = 64 * 1024 * 1024;

/// Clump together small bloblogs when together they would hit or exceed this size.
pub const REPACK_BLOBLOGS_TO_CLUMP_TOGETHER_SMALL_BLOBLOGS_BYTES: u64 = 2 * 1024 * 1024 * 1024;

/// The target size to reach when repacking, in terms of blob bytes.
pub const REPACK_TARGET_SIZE: u64 = 4 * 1024 * 1024;

/// The limit size to use when repacking, in terms of blob bytes.
pub const REPACK_TARGET_LIMIT: u64 = 5 * 1024 * 1024;

/// Gets bloblogs' stats. Only considers bloblogs referenced by exactly one index, so we don't
/// have to deal with unifying indices.
pub async fn get_bloblogs_stats(
    pwc: &Arc<PileWithCache<BoxedWormFileProvider>>,
) -> eyre::Result<BTreeMap<BloblogId, BloblogStats>> {
    let mut cache_conn = pwc.localcache.read().await?;
    let indices = cache_conn.list_indices().await?;
    let mut bloblogs: BTreeMap<BloblogId, Option<BloblogStats>> = BTreeMap::new();

    for index in indices {
        for (bloblog, stats) in cache_conn.index_bloblog_stats(index).await? {
            match bloblogs.entry(bloblog) {
                Entry::Vacant(ve) => {
                    ve.insert(Some(stats));
                }
                Entry::Occupied(mut oe) => {
                    // only allow one stats per bloblog, then replace with None.
                    oe.insert(None);
                }
            }
        }
    }

    Ok(bloblogs
        .into_iter()
        .flat_map(|(k, v)| v.map(|v| (k, v)))
        .collect())
}

/// Choose some bloblogs to repack. Assumes an updated local cache.
///
/// Only bloblogs referenced by exactly one index will be considered for repacking.
pub async fn select_bloblogs_for_repack(
    stats: BTreeMap<BloblogId, BloblogStats>,
) -> eyre::Result<Vec<BTreeMap<BloblogId, BloblogStats>>> {
    let mut repack_for_space: BTreeSet<BloblogId> = stats
        .iter()
        .filter(|(_, v)| v.forgotten_bytes >= REPACK_BLOBLOGS_TO_RECLAIM_SPACE_BYTES)
        .map(|(&k, _)| k)
        .collect();
    let maybe_repack_for_clumping: BTreeSet<BloblogId> = stats
        .iter()
        .filter(|(_, v)| v.blob_size <= SMALL_BLOBLOG_THRESHOLD)
        .map(|(&k, _)| k)
        .collect();

    let should_repack_for_clumping = maybe_repack_for_clumping.len() > 1
        && maybe_repack_for_clumping
            .iter()
            .map(|bi| stats[bi].blob_size)
            .sum::<u64>()
            > REPACK_BLOBLOGS_TO_CLUMP_TOGETHER_SMALL_BLOBLOGS_BYTES;

    let to_repack = repack_for_space.clone();
    if should_repack_for_clumping {
        repack_for_space.extend(maybe_repack_for_clumping);
    }

    let mut to_repack: BTreeSet<(u64, BloblogId)> = to_repack
        .into_iter()
        .map(|bi| (stats[&bi].blob_size, bi))
        .collect();

    let mut repack_sets = Vec::new();

    while !to_repack.is_empty() {
        let mut new_repack_group = BTreeMap::new();
        let mut new_repack_group_size = 0u64;

        let (first_sz, first_to_repack) = to_repack.pop_last().unwrap();
        new_repack_group_size += first_sz;
        new_repack_group.insert(first_to_repack, stats[&first_to_repack].clone());

        while new_repack_group_size < REPACK_TARGET_SIZE {
            let Some((first_size, _)) = to_repack.first() else { break; };
            if new_repack_group_size + *first_size > REPACK_TARGET_LIMIT {
                break;
            }
            let (extra_size, extra_bloblog_id) = to_repack.pop_first().unwrap();
            new_repack_group_size += extra_size;
            new_repack_group.insert(extra_bloblog_id, stats[&extra_bloblog_id].clone());
        }

        // now check the repack group is good
        if new_repack_group
            .keys()
            .any(|bi| repack_for_space.contains(bi))
            || new_repack_group_size > REPACK_BLOBLOGS_TO_CLUMP_TOGETHER_SMALL_BLOBLOGS_BYTES
        {
            repack_sets.push(new_repack_group);
        }
    }

    Ok(repack_sets)
}

pub async fn perform_repack(
    pwc: Arc<PileWithCache<BoxedWormFileProvider>>,
    repack_sets: Vec<BTreeMap<BloblogId, BloblogStats>>,
) -> eyre::Result<()> {
    // 1. Write new bloblogs
    let mut indices_buffer = BTreeMap::new();
    let mut index_parts: BTreeMap<BloblogId, IndexBloblogEntry> = BTreeMap::new();
    for repack_set in &repack_sets {
        let mut new_bloblog = pwc.pile.create_bloblog().await?;

        for (old_bloblog_id, old_bloblog_stats) in repack_set {
            let index_id = old_bloblog_stats.in_index;
            if !indices_buffer.contains_key(&index_id) {
                indices_buffer.insert(index_id, pwc.pile.read_index(index_id).await?);
            }
            let index_bloblog_entry = indices_buffer
                .get_mut(&index_id)
                .unwrap()
                .bloblogs
                .remove(&old_bloblog_id)
                .context("bug: no IBE despite rewrite from context of this index")?;
            let mut old_bloblog = pwc.pile.read_bloblog(*old_bloblog_id).await?;
            let locators: BTreeMap<u64, ChunkId> = index_bloblog_entry
                .chunks
                .into_iter()
                .map(|(blob, locator)| (locator.offset, blob))
                .collect();
            for chunk_id in locators.into_values() {
                let chunk = old_bloblog
                    .read_chunk(chunk_id)
                    .await?
                    .context("bug or corrupt bloblog: promised chunk missing")?;
                new_bloblog.write_chunk(chunk_id, &chunk).await?;
            }
        }

        let (_wormpath, new_bloblog_id, new_bloblog_index_info) = new_bloblog.finish().await?;
        index_parts.insert(
            new_bloblog_id,
            IndexBloblogEntry {
                chunks: new_bloblog_index_info,
                forgotten_bytes: 0,
            },
        );
    }

    // 2. Write new indices, but make sure to also write out index entries for unaffected bloblogs
    //    that appear in the indices we want to replace shortly.
    for (_, index) in indices_buffer.iter_mut() {
        index_parts.extend(std::mem::take(&mut index.bloblogs));
    }
    assemble_and_write_indices(&pwc, index_parts).await?;

    // 3. Delete old indices
    for index_id in indices_buffer.into_keys() {
        pwc.pile.delete_index_dangerous_exclusive(index_id).await?;
    }

    // 4. Delete old bloblogs
    for repack_group in repack_sets {
        for bloblog_id in repack_group.into_keys() {
            pwc.pile
                .delete_bloblog_dangerous_exclusive(bloblog_id)
                .await?;
        }
    }

    Ok(())
}
