use eyre::Context;
use itertools::Itertools;
use sqlx::pool::PoolConnection;
use sqlx::sqlite::{
    SqliteConnectOptions, SqliteJournalMode, SqlitePoolOptions, SqliteRow, SqliteSynchronous,
};
use sqlx::{query, Connection, Row, Sqlite, SqlitePool};
use std::collections::{BTreeMap, BTreeSet};
use std::path::Path;
use std::str::FromStr;
use std::sync::Arc;
use tokio::sync::{OwnedSemaphorePermit, Semaphore};
use yama_midlevel_crypto::chunk_id::ChunkId;
use yama_pile::definitions::{BlobLocator, BloblogId, Index, IndexId};

#[derive(Clone)]
pub struct Store {
    pool: Arc<SqlitePool>,
    writer_semaphore: Arc<Semaphore>,
}

pub struct StoreConnection<const RW: bool> {
    /// The underlying 'connection'.
    conn: PoolConnection<Sqlite>,

    /// Permit to write. Only here so that it is dropped at the correct time.
    #[allow(dead_code)]
    writer_permit: Option<OwnedSemaphorePermit>,
}

const MAX_SQLITE_CONNECTIONS: u32 = 16;

impl Store {
    pub async fn new(path: &Path) -> eyre::Result<Store> {
        let pool = SqlitePoolOptions::new()
            .max_connections(MAX_SQLITE_CONNECTIONS)
            .connect_with(
                SqliteConnectOptions::new()
                    .create_if_missing(true)
                    .journal_mode(SqliteJournalMode::Wal)
                    .synchronous(SqliteSynchronous::Normal)
                    .foreign_keys(true)
                    .filename(path),
            )
            .await?;

        let store = Store {
            pool: Arc::new(pool),
            writer_semaphore: Arc::new(Semaphore::new(1)),
        };

        let mut conn = store.pool.acquire().await?;

        // This will run the necessary migrations.
        sqlx::migrate!("./migrations").run(&mut conn).await?;

        Ok(store)
    }

    async fn txn<const RW: bool>(&self) -> eyre::Result<StoreConnection<RW>> {
        let writer_permit = if RW {
            Some(self.writer_semaphore.clone().acquire_owned().await?)
        } else {
            None
        };

        let conn = self.pool.acquire().await?;

        Ok(StoreConnection {
            conn,
            writer_permit,
        })
    }

    pub async fn read(&self) -> eyre::Result<StoreConnection<false>> {
        self.txn().await
    }

    pub async fn write(&self) -> eyre::Result<StoreConnection<true>> {
        self.txn().await
    }
}

impl StoreConnection<true> {
    /// Store an index into the local index cache.
    /// If the index supersedes any others, then those supersedings are stored and the blob entries
    /// for the superseded indices are removed.
    ///
    /// Returns true iff the index was new.
    pub async fn apply_index(
        &mut self,
        index_id: IndexId,
        index: Arc<Index>,
    ) -> eyre::Result<bool> {
        let index_id_txt = index_id.to_string();
        self.conn.transaction(|txn| Box::pin(async move {
            let needs_index = query!("
                SELECT 1 AS one FROM indices WHERE index_sha256 = ?
            ", index_id_txt).fetch_optional(&mut *txn).await?.is_none();

            if !needs_index {
                return Ok(false);
            }

            let index_short_id = query!("
                INSERT INTO indices (index_sha256)
                VALUES (?)
                RETURNING index_short_id
            ", index_id_txt).fetch_one(&mut *txn).await?.index_short_id;

            for supersede in &index.supersedes {
                let supersede_txt = supersede.to_string();
                query!("
                    INSERT INTO indices_supersede (superseded_sha256, successor_sha256)
                    VALUES (?, ?)
                ", supersede_txt, index_id_txt).execute(&mut *txn).await?;

                if let Some(row) = query!("
                    SELECT index_short_id FROM indices WHERE index_sha256 = ?
                ", supersede_txt).fetch_optional(&mut *txn).await? {
                    // Clear out any chunk entries for the superseded indices.
                    // This ensures we don't rely on them in the future and also clears up space.
                    query!("
                        DELETE FROM blobs WHERE index_short_id = ?
                    ", row.index_short_id).execute(&mut *txn).await?;
                }
            }

            // Check that the index hasn't already been superseded, before adding blobs
            let is_superseded = query!("
                SELECT 1 as _yes FROM indices_supersede WHERE superseded_sha256 = ?",
                index_id_txt
            ).fetch_optional(&mut *txn).await?.is_some();

            if !is_superseded {
                for (bloblog_sha256, index_bloblog_entry) in &index.bloblogs {
                    let bloblog_sha256_txt = bloblog_sha256.to_string();
                    let bloblog_short_id_opt = query!("
                        SELECT bloblog_short_id FROM bloblogs WHERE bloblog_sha256 = ?
                    ", bloblog_sha256_txt).fetch_optional(&mut *txn).await?;

                    let bloblog_short_id = match bloblog_short_id_opt {
                        None => {
                            query!("
                                INSERT INTO bloblogs (bloblog_sha256)
                                VALUES (?)
                                RETURNING bloblog_short_id
                            ", bloblog_sha256_txt).fetch_one(&mut *txn).await?.bloblog_short_id
                        },
                        Some(row) => row.bloblog_short_id,
                    };

                    let forgotten_bytes = index_bloblog_entry.forgotten_bytes as i64;
                    query!("
                        INSERT INTO indices_bloblogs (index_short_id, bloblog_short_id, forgotten_bytes)
                        VALUES (?, ?, ?)
                    ", index_short_id, bloblog_short_id, forgotten_bytes)
                        .execute(&mut *txn)
                        .await?;

                    for (chunk_id, chunk_locator) in index_bloblog_entry.chunks.iter() {
                        let chunk_id_txt = chunk_id.to_string();
                        let coffset = chunk_locator.offset as i64;
                        let clen = chunk_locator.length as i64;
                        query!("
                            INSERT INTO blobs (index_short_id, bloblog_short_id, chunk_id, offset, size)
                            VALUES (?, ?, ?, ?, ?)
                        ", index_short_id, bloblog_short_id, chunk_id_txt, coffset, clen).execute(&mut *txn).await?;
                    }
                }
            }

            Ok(true)
        })).await
    }

    /// Delete an index from the cache, if the cache was deleted from the pile.
    pub async fn delete_index(&mut self, index_id: IndexId) -> eyre::Result<()> {
        self.conn
            .transaction(|txn| {
                Box::pin(async move {
                    let index_id_txt = index_id.to_string();
                    query!(
                        "
                DELETE FROM indices_supersede WHERE successor_sha256 = ?
            ",
                        index_id_txt
                    )
                    .execute(&mut *txn)
                    .await?;

                    let index_short_id = query!(
                        "
                SELECT index_short_id FROM indices WHERE index_sha256 = ?
            ",
                        index_id_txt
                    )
                    .fetch_one(&mut *txn)
                    .await?
                    .index_short_id;

                    query!(
                        "
                DELETE FROM blobs WHERE index_short_id = ?
            ",
                        index_short_id
                    )
                    .execute(&mut *txn)
                    .await?;

                    query!(
                        "
                DELETE FROM indices_bloblogs WHERE index_short_id = ?
            ",
                        index_short_id
                    )
                    .execute(&mut *txn)
                    .await?;

                    query!(
                        "
                DELETE FROM indices WHERE index_short_id = ?
            ",
                        index_short_id
                    )
                    .execute(&mut *txn)
                    .await?;
                    Ok::<_, eyre::Report>(())
                })
            })
            .await?;

        Ok(())
    }
}

impl<const RW: bool> StoreConnection<RW> {
    pub async fn locate_chunk(
        &mut self,
        chunk_id: ChunkId,
    ) -> eyre::Result<Option<(BloblogId, BlobLocator)>> {
        let chunk_id_text = chunk_id.to_string();
        let row_opt = query!(
            "
            SELECT bl.bloblog_sha256, b.offset, b.size
            FROM blobs b
            JOIN bloblogs bl USING (bloblog_short_id)
            WHERE b.chunk_id = ?
            LIMIT 1
        ",
            chunk_id_text
        )
        .fetch_optional(&mut *self.conn)
        .await?;

        match row_opt {
            None => Ok(None),
            Some(row) => {
                let bloblog_id =
                    BloblogId::try_from(row.bloblog_sha256.as_str()).with_context(|| {
                        format!("failed to decode bloblog ID: {:?}", row.bloblog_sha256)
                    })?;
                Ok(Some((
                    bloblog_id,
                    BlobLocator {
                        offset: row.offset.try_into().context("offset too big")?,
                        length: row.size.try_into().context("size too big")?,
                    },
                )))
            }
        }
    }

    /// Returns all chunk locations.
    /// If a chunk does not exist, it is just not returned in the output map.
    pub async fn locate_chunks(
        &mut self,
        chunk_ids: &BTreeSet<ChunkId>,
    ) -> eyre::Result<BTreeMap<ChunkId, (BloblogId, BlobLocator)>> {
        let mut out = BTreeMap::new();
        for batch in &chunk_ids.iter().chunks(64) {
            let chunk_id_texts: Vec<String> = batch.map(|ci| ci.to_string()).collect();
            let query_param_str = format!("({})", &",?".repeat(chunk_id_texts.len())[1..]);
            let sql = format!(
                "
                SELECT b.chunk_id, bl.bloblog_sha256, b.offset, b.size
                FROM blobs b
                JOIN bloblogs bl USING (bloblog_short_id)
                WHERE b.chunk_id IN {query_param_str}
            "
            );

            let mut q = query(&sql);
            for chunk_id in &chunk_id_texts {
                q = q.bind(chunk_id);
            }
            let rows = q
                .map(|row: SqliteRow| {
                    Ok::<_, eyre::Report>((
                        ChunkId::from_str(row.get::<&str, _>(0))?,
                        BloblogId::try_from(row.get::<&str, _>(1))?,
                        row.get::<i64, _>(2),
                        row.get::<i64, _>(3),
                    ))
                })
                .fetch_all(&mut *self.conn)
                .await?;
            for row in rows {
                let (chunk_id, bloblog_id, offset, size) = row?;
                out.insert(
                    chunk_id,
                    (
                        bloblog_id,
                        BlobLocator {
                            offset: offset as u64,
                            length: size as u64,
                        },
                    ),
                );
            }
        }
        Ok(out)
    }

    pub async fn list_indices(&mut self) -> eyre::Result<BTreeSet<IndexId>> {
        let row_results = query!(
            "
            SELECT index_sha256 FROM indices
        "
        )
        .map(|row| {
            IndexId::try_from(row.index_sha256.as_ref())
                .context("failed to decode IndexId in local cache")
        })
        .fetch_all(&mut *self.conn)
        .await?;
        row_results.into_iter().collect()
    }

    pub async fn list_bloblogs(&mut self) -> eyre::Result<BTreeSet<BloblogId>> {
        let row_results = query!(
            "
            SELECT bloblog_sha256 FROM bloblogs
        "
        )
        .map(|row| {
            BloblogId::try_from(row.bloblog_sha256.as_ref())
                .context("failed to decode BloblogId in local cache")
        })
        .fetch_all(&mut *self.conn)
        .await?;
        row_results.into_iter().collect()
    }

    pub async fn is_chunk_new(&mut self, chunk_id: ChunkId) -> eyre::Result<bool> {
        let chunk_id_text = chunk_id.to_string();
        let is_new = query!(
            "
            SELECT 1 AS _yes FROM blobs WHERE chunk_id = ?
        ",
            chunk_id_text
        )
        .fetch_optional(&mut *self.conn)
        .await?
        .is_none();
        Ok(is_new)
    }

    pub async fn list_chunks_in_index(
        &mut self,
        index_id: IndexId,
    ) -> eyre::Result<BTreeSet<ChunkId>> {
        let index_id_text = index_id.to_string();
        let row_results = query!(
            "
            SELECT chunk_id AS \"chunk_id!\" FROM indices i
            JOIN blobs b USING (index_short_id)
            WHERE index_sha256 = ?
        ",
            index_id_text
        )
        .map(|row| {
            ChunkId::from_str(&row.chunk_id).context("failed to decode ChunkId in local cache")
        })
        .fetch_all(&mut *self.conn)
        .await?;
        row_results.into_iter().collect()
    }

    pub async fn index_bloblog_stats(
        &mut self,
        index_id: IndexId,
    ) -> eyre::Result<BTreeMap<BloblogId, BloblogStats>> {
        let index_id_text = index_id.to_string();
        let row_results = query!("
            SELECT bloblog_sha256 AS bloblog_id, ib.forgotten_bytes AS forgotten_bytes, COUNT(size) AS \"num_chunks!: i64\", SUM(size) AS \"num_bytes!: i64\" FROM indices i
            LEFT JOIN indices_bloblogs ib USING (index_short_id)
            LEFT JOIN bloblogs b USING (bloblog_short_id)
            LEFT JOIN blobs USING (index_short_id, bloblog_short_id)
            WHERE index_sha256 = ?
            GROUP BY bloblog_sha256
        ", index_id_text)
            .map(|row| {
                Ok((BloblogId::try_from(row.bloblog_id.as_ref())?, BloblogStats {
                    in_index: index_id,
                    blob_size: row.num_bytes as u64,
                    forgotten_bytes: row.forgotten_bytes as u64,
                    num_chunks: row.num_chunks as u32,
                }))
            })
            .fetch_all(&mut *self.conn)
            .await?;
        row_results.into_iter().collect()
    }
}

#[derive(Clone, Debug)]
pub struct BloblogStats {
    pub in_index: IndexId,
    pub blob_size: u64,
    pub forgotten_bytes: u64,
    pub num_chunks: u32,
}
