use ed25519_dalek::SIGNATURE_LENGTH;
use serde::{Deserialize, Serialize};
use std::marker::PhantomData;

use crate::asym_keyx::{AsymKeyExchange, DecryptingKey, EncryptingKey, KEY_EXCHANGE_LENGTH};
use crate::asym_signed::{SignedBytes, SigningKey, VerifyingKey};
use crate::byte_layer::ByteLayer;
use crate::sym_box::{SymBox, SymKey};

/// A locked box storing something using asymmetric cryptography.
///
/// For key encapsulation: x25519 and kyber (quantum-resistant)
/// For signing: ed25519 (not quantum-resistant)
///
#[derive(Clone, Serialize, Deserialize)]
#[serde(transparent)]
pub struct AsymBox<T> {
    payload: SignedBytes,
    #[serde(skip, default)]
    _phantom: PhantomData<T>,
}

/// A public key, needed to lock an AsymBox or verify the signature when unlocking an AsymBox.
#[derive(Clone, Serialize, Deserialize)]
pub struct AsymPublicKey {
    verify: VerifyingKey,
    encrypt: EncryptingKey,
}

/// A private key, needed to open an AsymBox or to sign an AsymBox that is being locked.
#[derive(Clone, Serialize, Deserialize)]
pub struct AsymPrivateKey {
    sign: SigningKey,
    decrypt: DecryptingKey,
}

impl<T> ByteLayer for AsymBox<T> {
    fn from_byte_vec(bytes: Vec<u8>) -> Self {
        Self {
            payload: SignedBytes::from_bytes_vec_assumed(bytes),
            _phantom: Default::default(),
        }
    }

    fn into_byte_vec(self) -> Vec<u8> {
        self.payload.into_bytes_vec()
    }
}

impl<T: ByteLayer> AsymBox<T> {
    // TODO error
    pub fn unlock(
        self,
        receiver_decrypt_key: &DecryptingKey,
        sender_verify_key: &VerifyingKey,
    ) -> Option<T> {
        let verified_payload = self.payload.into_verified(sender_verify_key)?;

        let cutoff = verified_payload.len() - KEY_EXCHANGE_LENGTH;

        let key_exchanger_bytes = &verified_payload[cutoff..];
        let key_exchanger = AsymKeyExchange::load_from_bytes(key_exchanger_bytes)
            .expect("can't load AsymKeyExchange");
        let exchanged = key_exchanger.unlock(receiver_decrypt_key)?;
        let symkey = SymKey::from(exchanged);

        let symbox: SymBox<T> = SymBox::new_from_raw(&verified_payload[0..cutoff]);
        symbox.unlock(&symkey)
    }
}

impl<T: ByteLayer> AsymBox<T> {
    // TODO error
    pub fn new(
        contents: T,
        sender_signing_key: &SigningKey,
        receiver_encrypt_key: &EncryptingKey,
    ) -> Option<Self> {
        let (key_exchanger, exchanged) = AsymKeyExchange::lock(receiver_encrypt_key);
        let symkey = SymKey::from(exchanged);

        let mut signed_bytes = SymBox::new(contents, &symkey)?.into_vec();
        signed_bytes.reserve(KEY_EXCHANGE_LENGTH + SIGNATURE_LENGTH);
        signed_bytes.extend_from_slice(key_exchanger.as_bytes());

        let signed = SignedBytes::new(signed_bytes, sender_signing_key);
        Some(Self {
            payload: signed,
            _phantom: Default::default(),
        })
    }
}
