use crate::keys_kyber::{kyber_keypair, KyberPrivateKey, KyberPublicKey};
use crate::keys_x25519::{x25519_keypair, X25519PrivateKey, X25519PublicKey};
use serde::{Deserialize, Serialize};
use std::borrow::Cow;
use std::io::Read;

// x25519 ephemeral pubkey (32) + kyber (1568)
pub const KEY_EXCHANGE_LENGTH: usize = 32 + 1568;

#[derive(Clone, Serialize, Deserialize)]
#[serde(transparent)]
pub struct AsymKeyExchange<'bytes> {
    inner: Cow<'bytes, [u8]>,
}

#[derive(Clone, Serialize, Deserialize, PartialEq)]
pub struct EncryptingKey {
    x25519: X25519PublicKey,
    kyber: KyberPublicKey,
}

#[derive(Clone, Serialize, Deserialize, PartialEq)]
pub struct DecryptingKey {
    x25519: X25519PrivateKey,
    kyber: KyberPrivateKey,

    x25519_pub: X25519PublicKey,
    kyber_pub: KyberPublicKey,
}

pub fn generate_asym_keypair() -> (EncryptingKey, DecryptingKey) {
    let (x25519_pub, x25519_priv) = x25519_keypair();
    let (kyber_pub, kyber_priv) = kyber_keypair();

    (
        EncryptingKey {
            x25519: x25519_pub.clone(),
            kyber: kyber_pub.clone(),
        },
        DecryptingKey {
            x25519: x25519_priv,
            kyber: kyber_priv,
            x25519_pub,
            kyber_pub,
        },
    )
}

pub struct Exchanged(pub(crate) [u8; 64]);

impl<'bytes> AsymKeyExchange<'bytes> {
    pub fn load_from_bytes(bytes: &'bytes [u8]) -> Option<Self> {
        if bytes.len() != KEY_EXCHANGE_LENGTH {
            return None;
        }
        Some(Self {
            inner: Cow::Borrowed(&bytes),
        })
    }

    pub fn as_bytes(&self) -> &[u8] {
        self.inner.as_ref()
    }

    pub fn lock(ek: &EncryptingKey) -> (AsymKeyExchange, Exchanged) {
        let mut public_bytes = Vec::with_capacity(KEY_EXCHANGE_LENGTH);

        let mut rand = rand::thread_rng();

        // X25519
        let ephemeral_privkey = x25519_dalek::EphemeralSecret::random_from_rng(&mut rand);
        let ephemeral_pubkey = x25519_dalek::PublicKey::from(&ephemeral_privkey);
        public_bytes.extend_from_slice(ephemeral_pubkey.as_bytes());
        let shared_secret_x25519 = ephemeral_privkey.diffie_hellman(&ek.x25519.inner);

        // Kyber
        let kyber = ek.kyber.encapsulate(&mut rand);
        public_bytes.extend_from_slice(&kyber.public_bytes);

        assert_eq!(public_bytes.len(), KEY_EXCHANGE_LENGTH);

        let exchanged = Self::perform_exchange(
            &public_bytes,
            &ek.x25519,
            &ek.kyber,
            shared_secret_x25519.as_bytes(),
            &kyber.shared_secret,
        );

        (
            AsymKeyExchange {
                inner: Cow::Owned(public_bytes),
            },
            exchanged,
        )
    }

    fn perform_exchange(
        public_bytes: &[u8],
        rx_x25519_pub: &X25519PublicKey,
        rx_kyber_pub: &KyberPublicKey,
        x25519_ss: &[u8; 32],
        kyber_ss: &[u8; 32],
    ) -> Exchanged {
        assert_eq!(public_bytes.len(), KEY_EXCHANGE_LENGTH);

        let mut hasher = blake3::Hasher::new_derive_key("yama AsymKeyExchange");
        // Includes the pubkeys of the writer
        hasher.update(public_bytes);

        // Include the pubkeys of the receiver
        hasher.update(rx_x25519_pub.as_bytes());
        hasher.update(rx_kyber_pub.as_bytes());

        // Include what was exchanged
        hasher.update(x25519_ss);
        hasher.update(kyber_ss);

        let mut exchanged_bytes = [0u8; 64];
        let mut out = hasher.finalize_xof();
        out.read_exact(&mut exchanged_bytes)
            .expect("failed to read 64b from blake3");

        Exchanged(exchanged_bytes)
    }

    pub fn unlock(&self, dk: &DecryptingKey) -> Option<Exchanged> {
        if self.inner.len() != KEY_EXCHANGE_LENGTH {
            return None;
        }

        // X25519
        let ephemeral_x25519_pubkey_bytes: &[u8; 32] = &self.inner[0..32].try_into().unwrap();
        let ephemeral_x25519_pubkey = x25519_dalek::PublicKey::from(*ephemeral_x25519_pubkey_bytes);
        let shared_secret_x25519 = dk.x25519.inner.diffie_hellman(&ephemeral_x25519_pubkey);

        // Kyber
        let kyber_ciphertext_bytes: &[u8; 1568] = &self.inner[32..].try_into().unwrap();
        let kyber = dk.kyber.decapsulate(&kyber_ciphertext_bytes);

        Some(Self::perform_exchange(
            &self.inner,
            &dk.x25519_pub,
            &dk.kyber_pub,
            shared_secret_x25519.as_bytes(),
            &kyber,
        ))
    }
}
