use serde::de::Error as DeError;
use serde::{Deserialize, Deserializer, Serialize, Serializer};

use ed25519_dalek::{
    Signature, Signer, SigningKey as Ed25519PrivateKey, Verifier, VerifyingKey as Ed25519PublicKey,
};

pub use ed25519_dalek::SIGNATURE_LENGTH;
use rand::thread_rng;

#[derive(Clone, Serialize, Deserialize)]
#[serde(transparent)]
pub struct SignedBytes {
    inner: Vec<u8>,
}

#[derive(Clone)]
pub struct SigningKey {
    ed25519: Ed25519PrivateKey,
}

impl PartialEq for SigningKey {
    fn eq(&self, other: &Self) -> bool {
        self.ed25519.to_keypair_bytes() == other.ed25519.to_keypair_bytes()
    }
}

#[derive(Clone)]
pub struct VerifyingKey {
    ed25519: Ed25519PublicKey,
}

impl PartialEq for VerifyingKey {
    fn eq(&self, other: &Self) -> bool {
        self.ed25519.as_bytes() == other.ed25519.as_bytes()
    }
}

pub fn asym_signing_keypair() -> (SigningKey, VerifyingKey) {
    let mut rng = thread_rng();
    let keypair = ed25519_dalek::SigningKey::generate(&mut rng);

    (
        SigningKey {
            ed25519: keypair.clone(),
        },
        VerifyingKey {
            ed25519: keypair.verifying_key(),
        },
    )
}

impl Serialize for SigningKey {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let ed25519 = self.ed25519.to_bytes();
        <[u8]>::serialize(&ed25519, serializer)
    }
}

impl<'d> Deserialize<'d> for SigningKey {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'d>,
    {
        let bytes = Vec::<u8>::deserialize(deserializer)?;

        let mut ed25519 = [0u8; 32];
        if bytes.len() != ed25519.len() {
            return Err(D::Error::custom("wrong length of ed25519 key"));
        }
        ed25519.copy_from_slice(&bytes);

        Ok(SigningKey {
            ed25519: Ed25519PrivateKey::from_bytes(&ed25519),
        })
    }
}

impl Serialize for VerifyingKey {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        (self.ed25519.as_bytes() as &[u8]).serialize(serializer)
    }
}

impl<'d> Deserialize<'d> for VerifyingKey {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'d>,
    {
        let bytes = Vec::<u8>::deserialize(deserializer)?;

        let mut ed25519 = [0u8; 32];
        if bytes.len() != ed25519.len() {
            return Err(D::Error::custom("wrong length of ed25519 key"));
        }
        ed25519.copy_from_slice(&bytes);

        Ok(VerifyingKey {
            ed25519: Ed25519PublicKey::from_bytes(&ed25519).map_err(D::Error::custom)?,
        })
    }
}

impl SignedBytes {
    pub fn new(mut bytes: Vec<u8>, sign_with: &SigningKey) -> SignedBytes {
        let signature = sign_with.ed25519.sign(&bytes);
        let sig = signature.to_bytes();
        assert_eq!(sig.len(), SIGNATURE_LENGTH);

        bytes.extend(sig);
        SignedBytes { inner: bytes }
    }

    pub fn into_verified(mut self, verify_with: &VerifyingKey) -> Option<Vec<u8>> {
        if self.inner.len() < SIGNATURE_LENGTH {
            return None;
        }
        let (payload, sig) = self.inner.split_at(self.inner.len() - SIGNATURE_LENGTH);
        let sig = Signature::from_bytes(sig.try_into().expect("wrong split"));
        verify_with.ed25519.verify(&payload, &sig).ok()?;
        self.inner.drain(self.inner.len() - SIGNATURE_LENGTH..);
        Some(self.inner)
    }

    pub fn into_bytes_vec(self) -> Vec<u8> {
        self.inner
    }

    pub fn from_bytes_vec_assumed(inner: Vec<u8>) -> Self {
        Self { inner }
    }
}
