use serde::de::DeserializeOwned;
use serde::Serialize;
use std::marker::PhantomData;

/// Trait to help layering byte transformers together.
pub trait ByteLayer {
    fn from_byte_vec(bytes: Vec<u8>) -> Self;
    fn into_byte_vec(self) -> Vec<u8>;
}

#[derive(Clone)]
pub struct CborSerde<T> {
    bytes: Vec<u8>,
    marker: PhantomData<T>,
}

impl<T: Serialize + DeserializeOwned> ByteLayer for CborSerde<T> {
    fn from_byte_vec(bytes: Vec<u8>) -> Self {
        CborSerde {
            bytes,
            marker: PhantomData::default(),
        }
    }

    fn into_byte_vec(self) -> Vec<u8> {
        self.bytes
    }
}

impl<T: Serialize> CborSerde<T> {
    pub fn serialise(from: &T) -> eyre::Result<Self> {
        let mut bytes = Vec::new();
        ciborium::ser::into_writer(from, &mut bytes)?;
        Ok(CborSerde {
            bytes,
            marker: Default::default(),
        })
    }
}

impl<T: DeserializeOwned> CborSerde<T> {
    pub fn deserialise(&self) -> eyre::Result<T> {
        Ok(ciborium::de::from_reader(&self.bytes[..])?)
    }
}

impl ByteLayer for Vec<u8> {
    fn from_byte_vec(bytes: Vec<u8>) -> Self {
        bytes
    }

    fn into_byte_vec(self) -> Vec<u8> {
        self
    }
}
