use eyre::bail;
use rand::{thread_rng, Rng};
use serde::{Deserialize, Serialize};
use std::fmt::{Debug, Formatter};
use std::str::FromStr;

/// The ID of a chunk. It's a 256-bit BLAKE3 hash.
#[derive(Copy, Clone, Serialize, Deserialize, Ord, PartialOrd, Eq, PartialEq, Hash)]
#[repr(transparent)]
#[serde(transparent)]
pub struct ChunkId {
    blake3: [u8; 32],
}

impl Debug for ChunkId {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        for &byte in self.blake3.iter() {
            write!(f, "{:02x}", byte)?;
        }
        Ok(())
    }
}

impl ToString for ChunkId {
    fn to_string(&self) -> String {
        hex::encode(&self.blake3)
    }
}

impl From<[u8; 32]> for ChunkId {
    fn from(bytes: [u8; 32]) -> Self {
        ChunkId { blake3: bytes }
    }
}

impl FromStr for ChunkId {
    type Err = eyre::Report;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        if s.len() != 64 {
            bail!("chunk ID of wrong length");
        }
        let decoded = hex::decode(s)?;
        let mut new = ChunkId {
            blake3: Default::default(),
        };
        new.blake3.copy_from_slice(&decoded);
        Ok(new)
    }
}

impl ChunkId {
    pub fn to_bytes(self) -> [u8; 32] {
        self.blake3
    }
}

/// Key needed to create and verify chunk IDs. It's a 256-bit key for the BLAKE3 keyed hash function.
#[derive(Copy, Clone, Serialize, Deserialize)]
pub struct ChunkIdKey {
    key: [u8; 32],
}

impl ChunkIdKey {
    pub fn new_rand() -> ChunkIdKey {
        let mut rng = thread_rng();
        ChunkIdKey { key: rng.gen() }
    }
}

impl Debug for ChunkIdKey {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        // for privacy, don't leak the contents of the key
        write!(f, "ChunkIdKey(...)")
    }
}

impl ChunkId {
    /// Computes a chunk ID given the input and key.
    pub fn compute(input: &[u8], key: &ChunkIdKey) -> ChunkId {
        ChunkId {
            blake3: blake3::keyed_hash(&key.key, input).into(),
        }
    }

    /// Returns true iff this Chunk ID is correct for the given input and key.
    pub fn verify(&self, input: &[u8], key: &ChunkIdKey) -> bool {
        let comparison = Self::compute(input, key);
        self == &comparison
    }
}
