use crate::sym_box::SymKey;
use argon2::Algorithm::Argon2id;
use argon2::{Argon2, Version};
use eyre::{bail, Context};
use rand::{thread_rng, Rng};
use serde::{Deserialize, Serialize};

/// 1 GiB. Intended to prevent maliciously large memory costs; not sure if that's a real risk.
pub const MAX_MEMORY_COST_KIBIBYTES: u32 = 1048576;

/// 512 MiB
pub const DEFAULT_MEMORY_COST_KIBIBYTES: u32 = 524288;

pub const DEFAULT_LANES: u32 = 1;
pub const DEFAULT_ITERATIONS: u32 = 256;

#[derive(Clone, Serialize, Deserialize)]
pub struct KeyDerivationParameters {
    salt: [u8; 32],
    iterations: u32,
    memory_kibibytes: u32,
    lanes: u32,
}

impl KeyDerivationParameters {
    pub fn new_recommended() -> KeyDerivationParameters {
        let mut rng = thread_rng();
        KeyDerivationParameters {
            salt: rng.gen(),
            iterations: DEFAULT_ITERATIONS,
            memory_kibibytes: DEFAULT_MEMORY_COST_KIBIBYTES,
            lanes: DEFAULT_LANES,
        }
    }

    pub fn derive(&self, password: &str) -> eyre::Result<DerivedKey> {
        if self.memory_kibibytes > MAX_MEMORY_COST_KIBIBYTES {
            bail!(
                "Too much memory needed for key derivation! {} > {}",
                self.memory_kibibytes,
                MAX_MEMORY_COST_KIBIBYTES
            );
        }

        let mut params = argon2::ParamsBuilder::new();
        params
            .m_cost(self.memory_kibibytes)
            .unwrap()
            .p_cost(self.lanes)
            .unwrap()
            .t_cost(self.iterations)
            .unwrap()
            .output_len(64)
            .unwrap();
        let params = params.params().unwrap();
        let argon = Argon2::new(Argon2id, Version::V0x13, params.clone());
        let mut derived_key = DerivedKey([0; 64]);
        argon
            .hash_password_into(password.as_bytes(), &self.salt, &mut derived_key.0)
            .context("failed to hash password")?;

        Ok(derived_key)
    }
}

pub struct DerivedKey(pub(crate) [u8; 64]);

impl DerivedKey {
    pub fn into_symkey(self) -> SymKey {
        SymKey::from(self)
    }
}
