use pqc_kyber::CryptoRng;
use rand::{thread_rng, RngCore};
use serde::de::Error;
use serde::{Deserialize, Deserializer, Serialize, Serializer};
use std::borrow::Cow;

#[derive(Clone, Eq, PartialEq)]
pub struct KyberPublicKey {
    inner: pqc_kyber::PublicKey,
}

#[derive(Clone, Eq, PartialEq)]
pub struct KyberPrivateKey {
    inner: pqc_kyber::SecretKey,
}

pub fn kyber_keypair() -> (KyberPublicKey, KyberPrivateKey) {
    let mut rng = thread_rng();
    let kyber_pair = pqc_kyber::keypair(&mut rng);

    (
        KyberPublicKey {
            inner: kyber_pair.public,
        },
        KyberPrivateKey {
            inner: kyber_pair.secret,
        },
    )
}

impl Serialize for KyberPublicKey {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        <[u8]>::serialize(&self.inner, serializer)
    }
}

impl<'d> Deserialize<'d> for KyberPublicKey {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'d>,
    {
        let bytes = Cow::<'d, [u8]>::deserialize(deserializer)?;

        let mut kyber: pqc_kyber::PublicKey = [0; 1568];
        if kyber.len() != bytes.len() {
            return Err(D::Error::custom("wrong length of kyber key"));
        }
        kyber.copy_from_slice(&bytes);

        Ok(KyberPublicKey { inner: kyber })
    }
}

impl KyberPublicKey {
    pub fn as_bytes(&self) -> &[u8; 1568] {
        &self.inner
    }
}

impl Serialize for KyberPrivateKey {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        <[u8]>::serialize(&self.inner, serializer)
    }
}

impl<'d> Deserialize<'d> for KyberPrivateKey {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'d>,
    {
        let bytes = Cow::<'d, [u8]>::deserialize(deserializer)?;

        let mut kyber: pqc_kyber::SecretKey = [0; 3168];
        if kyber.len() != bytes.len() {
            return Err(D::Error::custom("wrong length of kyber key"));
        }
        kyber.copy_from_slice(&bytes);

        Ok(KyberPrivateKey { inner: kyber })
    }
}

pub struct KyberEncapsulation {
    pub shared_secret: pqc_kyber::SharedSecret,
    pub public_bytes: [u8; pqc_kyber::KYBER_CIPHERTEXTBYTES],
}

impl KyberPublicKey {
    pub fn encapsulate<T: RngCore + CryptoRng>(&self, rng: &mut T) -> KyberEncapsulation {
        let (public_bytes, shared_secret) =
            pqc_kyber::encapsulate(&self.inner, rng).expect("bad kyber encapsulation");
        KyberEncapsulation {
            shared_secret,
            public_bytes,
        }
    }
}

impl KyberPrivateKey {
    pub fn decapsulate(&self, ciphertext: &[u8; 1568]) -> pqc_kyber::SharedSecret {
        pqc_kyber::decapsulate(ciphertext, &self.inner).expect("bad kyber decapsulation")
    }
}
