use rand::thread_rng;
use serde::de::Error as _DeError;
use serde::{Deserialize, Deserializer, Serialize, Serializer};
use std::borrow::Cow;

#[derive(Clone)]
#[repr(transparent)]
pub struct X25519PrivateKey {
    pub(crate) inner: x25519_dalek::StaticSecret,
}

impl PartialEq for X25519PrivateKey {
    fn eq(&self, other: &Self) -> bool {
        self.inner.as_bytes() == other.inner.as_bytes()
    }
}

#[derive(Clone, PartialEq)]
#[repr(transparent)]
pub struct X25519PublicKey {
    pub(crate) inner: x25519_dalek::PublicKey,
}

pub fn x25519_keypair() -> (X25519PublicKey, X25519PrivateKey) {
    let mut rng = thread_rng();
    let x25519_priv = x25519_dalek::StaticSecret::random_from_rng(&mut rng);
    let x25519_pub = x25519_dalek::PublicKey::from(&x25519_priv);
    (
        X25519PublicKey { inner: x25519_pub },
        X25519PrivateKey { inner: x25519_priv },
    )
}

impl Serialize for X25519PrivateKey {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        <[u8]>::serialize(self.inner.as_bytes(), serializer)
    }
}

impl<'d> Deserialize<'d> for X25519PrivateKey {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'d>,
    {
        let bytes = Cow::<'d, [u8]>::deserialize(deserializer)?;
        let counted_bytes: [u8; 32] = bytes.as_ref().try_into().map_err(D::Error::custom)?;
        Ok(X25519PrivateKey {
            inner: x25519_dalek::StaticSecret::from(counted_bytes),
        })
    }
}

impl X25519PublicKey {
    pub fn as_bytes(&self) -> &[u8; 32] {
        self.inner.as_bytes()
    }
}

impl Serialize for X25519PublicKey {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        <[u8]>::serialize(self.inner.as_bytes(), serializer)
    }
}

impl<'d> Deserialize<'d> for X25519PublicKey {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'d>,
    {
        let bytes = Cow::<'d, [u8]>::deserialize(deserializer)?;
        let counted_bytes: [u8; 32] = bytes.as_ref().try_into().map_err(D::Error::custom)?;
        Ok(X25519PublicKey {
            inner: x25519_dalek::PublicKey::from(counted_bytes),
        })
    }
}
