use chacha20::cipher::{KeyIvInit, StreamCipher, StreamCipherSeek};
use chacha20::XChaCha20;
use rand::{CryptoRng, Rng, RngCore};
use serde::{Deserialize, Deserializer, Serialize, Serializer};
use std::fmt::{Debug, Formatter};

/// A symmetric key needed to encrypt or decrypt a symmetric stream.
pub struct SymStreamKey {
    /// 256-bit key
    key: [u8; 32],

    /// 192-bit nonce for XChaCha20
    nonce: [u8; 24],

    /// Cipher from chacha20 crate.
    cipher: XChaCha20,
}

impl Clone for SymStreamKey {
    fn clone(&self) -> Self {
        SymStreamKey {
            key: self.key,
            nonce: self.nonce,
            cipher: XChaCha20::new(&self.key.into(), &self.nonce.into()),
        }
    }
}

#[derive(Serialize, Deserialize)]
struct SymKeySerialisable {
    key: [u8; 32],
    nonce: [u8; 24],
}

impl Serialize for SymStreamKey {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        SymKeySerialisable {
            key: self.key,
            nonce: self.nonce,
        }
        .serialize(serializer)
    }
}

impl<'de> Deserialize<'de> for SymStreamKey {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        let SymKeySerialisable { key, nonce } = SymKeySerialisable::deserialize(deserializer)?;
        Ok(SymStreamKey {
            key,
            nonce,
            cipher: XChaCha20::new(&key.into(), &nonce.into()),
        })
    }
}

impl Debug for SymStreamKey {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "SymKey(...)")
    }
}

impl SymStreamKey {
    pub fn new_random<R: RngCore + CryptoRng>(rng: &mut R) -> Self {
        let key: [u8; 32] = rng.gen();
        let nonce: [u8; 24] = rng.gen();
        let cipher = XChaCha20::new(&key.into(), &nonce.into());
        Self { key, nonce, cipher }
    }

    #[inline]
    pub fn apply_xor(&mut self, offset: u64, content: &mut [u8]) {
        self.cipher.seek(offset);
        self.cipher.apply_keystream(content);
    }
}
