use crate::definitions::{BlobLocator, BloblogFooter, BloblogId, PackedBloblogFooter};
use crate::keyring::{Keyring, WriterKey};
use crate::locks::{LockHandle, LockKind};
use crate::utils::{HashedWormWriter, SymStreamReader, SymStreamWriter};
use eyre::{bail, Context, ContextCompat};
use rand::thread_rng;
use std::collections::BTreeMap;
use std::io::SeekFrom;
use std::sync::Arc;
use tokio::io::{AsyncReadExt, AsyncSeekExt, AsyncWriteExt};
use tracing::warn;
use yama_midlevel_crypto::asym_box::AsymBox;
use yama_midlevel_crypto::byte_layer::{ByteLayer, CborSerde};
use yama_midlevel_crypto::chunk_id::ChunkId;
use yama_midlevel_crypto::sym_stream::SymStreamKey;
use yama_midlevel_crypto::zstd_box::Zstd;
use yama_wormfile::paths::WormPathBuf;
use yama_wormfile::{WormFileReader, WormFileWriter};

/// The recommended bloblog size is 2 GiB.
/// This is below the 4 GiB file size limit on FAT32.
pub const RECOMMENDED_BLOBLOG_SIZE: u64 = 2 * 1024 * 1024 * 1024;

/// A writer for a bloblog.
pub struct BloblogWriter<W: WormFileWriter + Unpin> {
    writer: SymStreamWriter<HashedWormWriter<W>>,
    footer_write_key: WriterKey,
    footer: BloblogFooter,
    lock: Arc<LockHandle>,
    written: u64,
}

impl<WFW: WormFileWriter + Unpin> BloblogWriter<WFW> {
    /// Creates a bloblog writer.
    #[async_backtrace::framed]
    pub async fn new(
        worm_writer: WFW,
        keyring: &Keyring,
        lock: Arc<LockHandle>,
    ) -> eyre::Result<Self> {
        let hashed_writer = HashedWormWriter::new(worm_writer);
        let mut rng = thread_rng();
        let secret_content_stream_key = SymStreamKey::new_random(&mut rng);

        let content_stream_key = keyring
            .w_bloblog_contents
            .as_ref()
            .context("No `w_bloblog_contents` key on keyring")?
            .make_locked_asymbox(CborSerde::serialise(&secret_content_stream_key)?);

        let footer_write_key = keyring
            .w_bloblog_footer
            .clone()
            .context("No `w_bloblog_footer` key on keyring")?;

        let writer = SymStreamWriter::new(hashed_writer, secret_content_stream_key);

        Ok(Self {
            writer,
            footer_write_key,
            footer: BloblogFooter {
                content_stream_key,
                chunks: Default::default(),
            },
            lock,
            written: 0,
        })
    }

    /// Adds a chunk to the bloblog.
    #[async_backtrace::framed]
    pub async fn write_chunk(&mut self, chunk_id: ChunkId, chunk: &[u8]) -> eyre::Result<()> {
        let locator = BlobLocator {
            offset: self.writer.offset(),
            length: chunk.len() as u64,
        };
        self.writer.write_all(&chunk).await?;
        if self.footer.chunks.insert(chunk_id, locator).is_some() {
            warn!("Duplicate chunk ID inserted into bloblog: {:?}", chunk_id);
        }
        self.written += chunk.len() as u64;
        Ok(())
    }

    /// Returns true iff this bloblog writer should be finished to be close to the recommended
    /// maximum size.
    pub fn should_finish(&self) -> bool {
        self.written >= RECOMMENDED_BLOBLOG_SIZE
    }

    /// Finishes this bloblog.
    ///
    /// This:
    /// - writes the header
    /// - flushes
    /// - finishes computing the hash of the file
    /// - moves the bloblog to the correct place
    #[async_backtrace::framed]
    pub async fn finish(
        mut self,
    ) -> eyre::Result<(WormPathBuf, BloblogId, BTreeMap<ChunkId, BlobLocator>)> {
        self.writer.flush().await?;
        let mut hashed_writer = self.writer.finish();

        // Write the footer, then the length of the footer.
        let packed_footer: PackedBloblogFooter = self
            .footer_write_key
            .make_locked_asymbox(Zstd::compress(CborSerde::serialise(&self.footer)?));
        let footer_encoded = packed_footer.into_byte_vec();
        hashed_writer.write_all(&footer_encoded).await?;
        hashed_writer.write_u32(footer_encoded.len() as u32).await?;

        hashed_writer.flush().await?;
        let (mut worm_writer, file_hash) = hashed_writer.finalise();

        let target_path = WormPathBuf::new(format!(
            "bloblogs/{}/{}",
            hex::encode(&file_hash.0[0..1]),
            file_hash.to_string()
        ))
        .unwrap();

        if !self.lock.is_active_now(LockKind::Shared) {
            bail!(
                "Can't complete finish() on bloblog {:?} because lock expired",
                target_path
            );
        }

        worm_writer.finalise(target_path.as_ref(), false).await?;

        Ok((target_path, BloblogId(file_hash), self.footer.chunks))
    }
}

pub struct BloblogReader<R: WormFileReader> {
    reader: SymStreamReader<R>,
    footer: BloblogFooter,
}

impl<R: WormFileReader + Unpin> BloblogReader<R> {
    /// Read the bloblog footer from the file only.
    /// This only requires the `r_bloblog_footer` key.
    pub async fn read_footer_only(
        worm_reader: &mut R,
        keyring: &Keyring,
    ) -> eyre::Result<BloblogFooter> {
        let r_bloblog_footer = keyring
            .r_bloblog_footer
            .as_ref()
            .context("No key `r_bloblog_footer` on keyring")?;
        worm_reader.seek(SeekFrom::End(-4)).await?;
        let footer_length = worm_reader.read_u32().await?;

        worm_reader
            .seek(SeekFrom::End(-4 - footer_length as i64))
            .await?;
        let mut footer_enc = vec![0u8; footer_length as usize];
        worm_reader.read_exact(&mut footer_enc).await?;

        let footer_wrapped: PackedBloblogFooter = AsymBox::from_byte_vec(footer_enc);
        let footer: BloblogFooter = r_bloblog_footer
            .unlock_asymbox(footer_wrapped)
            .context("failed to decrypt bloblog footer")?
            .decompress()?
            .deserialise()?;
        Ok(footer)
    }

    pub async fn new(mut worm_reader: R, keyring: &Keyring) -> eyre::Result<Self> {
        let r_bloblog_contents = keyring
            .r_bloblog_contents
            .clone()
            .context("No key `r_bloblog_contents` on keyring")?;

        let footer = Self::read_footer_only(&mut worm_reader, keyring)
            .await
            .context("failed to read footer")?;

        let stream_key = r_bloblog_contents
            .unlock_asymbox(footer.content_stream_key.clone())
            .context("failed to decrypt stream key")?
            .deserialise()?;

        let reader = SymStreamReader::new(worm_reader, stream_key);

        Ok(Self { reader, footer })
    }

    pub async fn read_to_buf(
        &mut self,
        buf: &mut Vec<u8>,
        offset: u64,
        read_length: u64,
    ) -> eyre::Result<()> {
        // eprintln!("RTB @ {offset} r{read_length}");
        self.reader.seek(SeekFrom::Start(offset)).await?;
        buf.resize(read_length as usize, 0);
        self.reader.read_exact(buf).await?;
        Ok(())
    }

    pub async fn read_chunk(&mut self, chunk_id: ChunkId) -> eyre::Result<Option<Vec<u8>>> {
        match self.footer.chunks.get(&chunk_id) {
            Some(chunk_locator) => {
                let mut buf = Vec::with_capacity(chunk_locator.length as usize);
                self.read_to_buf(&mut buf, chunk_locator.offset, chunk_locator.length)
                    .await?;
                Ok(Some(buf))
            }
            None => Ok(None),
        }
    }

    pub fn footer(&self) -> &BloblogFooter {
        &self.footer
    }
}
