use crate::keyring::Keyring;
use crate::utils::Sha256;
use serde::{Deserialize, Serialize};
use std::collections::{BTreeMap, BTreeSet};
use std::fmt::{Debug, Formatter};
use std::sync::Arc;
use yama_midlevel_crypto::asym_box::AsymBox;
use yama_midlevel_crypto::byte_layer::CborSerde;
use yama_midlevel_crypto::chunk_id::{ChunkId, ChunkIdKey};
use yama_midlevel_crypto::key_derivation::KeyDerivationParameters;
use yama_midlevel_crypto::sym_box::SymBox;
use yama_midlevel_crypto::sym_stream::SymStreamKey;
use yama_midlevel_crypto::zstd_box::Zstd;

/// The footer at the end of a bloblog.
/// This footer should be encrypted and signed.
#[derive(Serialize, Deserialize)]
pub struct BloblogFooter {
    /// The key needed to read the content stream.
    pub content_stream_key: AsymBox<CborSerde<SymStreamKey>>,

    /// IDs of chunks and whereabouts they are in the bloblog.
    pub chunks: BTreeMap<ChunkId, BlobLocator>,
}

pub type PackedBloblogFooter = AsymBox<Zstd<CborSerde<BloblogFooter>>>;

/// Locator for a blob within a bloblog.
#[derive(Serialize, Deserialize, Eq, PartialEq)]
pub struct BlobLocator {
    pub offset: u64,
    pub length: u64,
}

#[derive(Copy, Clone, Serialize, Deserialize, Ord, PartialOrd, Eq, PartialEq, Hash)]
#[serde(transparent)]
pub struct BloblogId(pub Sha256);

impl ToString for BloblogId {
    fn to_string(&self) -> String {
        self.0.to_string()
    }
}

impl TryFrom<&str> for BloblogId {
    type Error = eyre::Error;

    fn try_from(value: &str) -> Result<Self, Self::Error> {
        Sha256::try_from(value).map(BloblogId)
    }
}

impl Debug for BloblogId {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "BloblogId({})", &self.0)
    }
}

#[derive(Copy, Clone, Serialize, Deserialize, Ord, PartialOrd, Eq, PartialEq, Hash)]
#[serde(transparent)]
pub struct IndexId(pub Sha256);

impl ToString for IndexId {
    fn to_string(&self) -> String {
        self.0.to_string()
    }
}

impl TryFrom<&str> for IndexId {
    type Error = eyre::Error;

    fn try_from(value: &str) -> Result<Self, Self::Error> {
        Sha256::try_from(value).map(IndexId)
    }
}

impl Debug for IndexId {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "IndexId({})", &self.0)
    }
}

/// An index.
#[derive(Serialize, Deserialize)]
pub struct Index {
    /// What indices this index supersedes.
    /// In other words: applying this index means you can ignore the stated indices below.
    pub supersedes: BTreeSet<IndexId>,

    /// Index
    pub bloblogs: BTreeMap<BloblogId, IndexBloblogEntry>,
}

/// One bloblog's entry within an index. Must contain an exhaustive list of chunks for that bloblog.
#[derive(Serialize, Deserialize)]
pub struct IndexBloblogEntry {
    /// Chunk locators
    pub chunks: BTreeMap<ChunkId, BlobLocator>,

    /// How much space, in bytes, has been deallocated / forgotten about
    /// (the cumulative size of chunks that have been removed from the chunk map)
    pub forgotten_bytes: u64,
}

pub type PackedIndex = AsymBox<Zstd<CborSerde<Index>>>;

#[derive(Copy, Clone, Serialize, Deserialize, PartialEq, Eq, Hash, Ord, PartialOrd)]
pub struct RecursiveChunkRef {
    /// The root Chunk ID.
    pub chunk_id: ChunkId,
    /// The depth of the data bytes.
    /// 0 means that the chunk addressed by `chunk_id` contains data bytes.
    /// 1 means that the chunk addressed by `chunk_id` contains references to chunk that contain
    /// data bytes.
    /// (and so on)
    pub depth: u32,
}

impl Debug for RecursiveChunkRef {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "{:?}<{}>", &self.chunk_id, self.depth)
    }
}

#[derive(Clone, Serialize, Deserialize)]
#[serde(tag = "_kind")]
pub enum UnlockedOrLockedKeyring {
    Locked {
        deriver: KeyDerivationParameters,
        lockbox: SymBox<'static, CborSerde<Keyring>>,
    },
    Unlocked(Keyring),
}

pub type PackedKeyring = CborSerde<UnlockedOrLockedKeyring>;

pub const SUPPORTED_YAMA_PILE_VERSION: &'static str = "yama v0.7.0 pile format";

#[derive(Clone, Serialize, Deserialize)]
pub struct PileConfig {
    pub yama_pile_version: String,
    pub chunk_id_key: ChunkIdKey,
    pub zstd_dict: Option<Arc<Vec<u8>>>,
}

pub type PackedPileConfig = AsymBox<CborSerde<PileConfig>>;
