use eyre::ContextCompat;
use serde::{Deserialize, Serialize};
use std::collections::BTreeSet;
use yama_midlevel_crypto::asym_box::AsymBox;
use yama_midlevel_crypto::asym_keyx::{generate_asym_keypair, DecryptingKey, EncryptingKey};
use yama_midlevel_crypto::asym_signed::{asym_signing_keypair, SigningKey, VerifyingKey};
use yama_midlevel_crypto::byte_layer::ByteLayer;

/// Exhaustive list of all keys
pub const INDIVIDUAL_KEYS: &[&str] = &[
    "r_config",
    "w_config",
    "r_bloblog_footer",
    "w_bloblog_footer",
    "r_bloblog_contents",
    "w_bloblog_contents",
    "r_locks",
    "w_locks",
    "r_pointer",
    "w_pointer",
];

/// Opinionated named groups of keys intended for specific purposes.
/// ALL: all keys
/// BACKUP:
pub const KEY_GROUPS: &[(&str, &[&str])] = &[
    ("ALL", INDIVIDUAL_KEYS),
    (
        "BACKUP",
        &[
            "r_config",
            "r_bloblog_footer",
            "w_bloblog_footer",
            "w_bloblog_contents",
            "r_locks",
            "w_locks",
            "r_pointer",
            "w_pointer",
        ],
    ),
    // TODO add more key groups... ("RESTORE", &[])
];

#[derive(Clone, Serialize, Deserialize, PartialEq)]
pub struct Keyring {
    pub r_config: Option<ReaderKey>,
    pub w_config: Option<WriterKey>,

    pub r_bloblog_footer: Option<ReaderKey>,
    pub w_bloblog_footer: Option<WriterKey>,

    pub r_bloblog_contents: Option<ReaderKey>,
    pub w_bloblog_contents: Option<WriterKey>,

    pub r_locks: Option<ReaderKey>,
    pub w_locks: Option<WriterKey>,

    pub r_pointer: Option<ReaderKey>,
    pub w_pointer: Option<WriterKey>,
}

impl Keyring {
    pub fn generate_new_master() -> Self {
        let (r_config, w_config) = generate_r_w_keys();
        let (r_bloblog_footer, w_bloblog_footer) = generate_r_w_keys();
        let (r_bloblog_contents, w_bloblog_contents) = generate_r_w_keys();
        let (r_locks, w_locks) = generate_r_w_keys();
        let (r_pointer, w_pointer) = generate_r_w_keys();
        Keyring {
            r_config: Some(r_config),
            w_config: Some(w_config),
            r_bloblog_footer: Some(r_bloblog_footer),
            w_bloblog_footer: Some(w_bloblog_footer),
            r_bloblog_contents: Some(r_bloblog_contents),
            w_bloblog_contents: Some(w_bloblog_contents),
            r_locks: Some(r_locks),
            w_locks: Some(w_locks),
            r_pointer: Some(r_pointer),
            w_pointer: Some(w_pointer),
        }
    }

    pub fn partial_copy(&self, keys_to_copy: &BTreeSet<&str>) -> eyre::Result<Self> {
        Ok(Keyring {
            r_config: if keys_to_copy.contains("r_config") {
                Some(
                    self.r_config
                        .as_ref()
                        .context("no r_config in original keyring")?
                        .clone(),
                )
            } else {
                None
            },
            w_config: if keys_to_copy.contains("w_config") {
                Some(
                    self.w_config
                        .as_ref()
                        .context("no w_config in original keyring")?
                        .clone(),
                )
            } else {
                None
            },
            r_bloblog_footer: if keys_to_copy.contains("r_bloblog_footer") {
                Some(
                    self.r_bloblog_footer
                        .as_ref()
                        .context("no r_bloblog_footer in original keyring")?
                        .clone(),
                )
            } else {
                None
            },
            w_bloblog_footer: if keys_to_copy.contains("w_bloblog_footer") {
                Some(
                    self.w_bloblog_footer
                        .as_ref()
                        .context("no w_bloblog_footer in original keyring")?
                        .clone(),
                )
            } else {
                None
            },
            r_bloblog_contents: if keys_to_copy.contains("r_bloblog_contents") {
                Some(
                    self.r_bloblog_contents
                        .as_ref()
                        .context("no r_bloblog_contents in original keyring")?
                        .clone(),
                )
            } else {
                None
            },
            w_bloblog_contents: if keys_to_copy.contains("w_bloblog_contents") {
                Some(
                    self.w_bloblog_contents
                        .as_ref()
                        .context("no w_bloblog_contents in original keyring")?
                        .clone(),
                )
            } else {
                None
            },
            r_locks: if keys_to_copy.contains("r_locks") {
                Some(
                    self.r_locks
                        .as_ref()
                        .context("no r_locks in original keyring")?
                        .clone(),
                )
            } else {
                None
            },
            w_locks: if keys_to_copy.contains("w_locks") {
                Some(
                    self.w_locks
                        .as_ref()
                        .context("no w_locks in original keyring")?
                        .clone(),
                )
            } else {
                None
            },
            r_pointer: if keys_to_copy.contains("r_pointer") {
                Some(
                    self.r_pointer
                        .as_ref()
                        .context("no r_pointer in original keyring")?
                        .clone(),
                )
            } else {
                None
            },
            w_pointer: if keys_to_copy.contains("w_pointer") {
                Some(
                    self.w_pointer
                        .as_ref()
                        .context("no w_pointer in original keyring")?
                        .clone(),
                )
            } else {
                None
            },
        })
    }
}

pub fn generate_r_w_keys() -> (ReaderKey, WriterKey) {
    let (encrypt, decrypt) = generate_asym_keypair();
    let (sign, verify) = asym_signing_keypair();
    (
        ReaderKey::new(decrypt, verify),
        WriterKey::new(encrypt, sign),
    )
}

#[derive(Clone, Serialize, Deserialize, PartialEq)]
pub struct WriterKey {
    // boxed because these take up a lot of stack space otherwise!
    #[serde(flatten)]
    inner: Box<WriterKeyInner>,
}

#[derive(Clone, Serialize, Deserialize, PartialEq)]
struct WriterKeyInner {
    encrypt: EncryptingKey,
    sign: SigningKey,
}

impl WriterKey {
    pub fn new(encrypt: EncryptingKey, sign: SigningKey) -> Self {
        Self {
            inner: Box::new(WriterKeyInner { encrypt, sign }),
        }
    }

    pub fn make_locked_asymbox<T: ByteLayer>(&self, contents: T) -> AsymBox<T> {
        AsymBox::new(contents, &self.inner.sign, &self.inner.encrypt).unwrap()
    }
}

#[derive(Clone, Serialize, Deserialize, PartialEq)]
pub struct ReaderKey {
    // boxed because these take up a lot of stack space otherwise!
    #[serde(flatten)]
    inner: Box<ReaderKeyInner>,
}

#[derive(Clone, Serialize, Deserialize, PartialEq)]
pub struct ReaderKeyInner {
    decrypt: DecryptingKey,
    verify: VerifyingKey,
}

impl ReaderKey {
    pub fn new(decrypt: DecryptingKey, verify: VerifyingKey) -> Self {
        Self {
            inner: Box::new(ReaderKeyInner { decrypt, verify }),
        }
    }

    pub fn unlock_asymbox<T: ByteLayer>(&self, asymbox: AsymBox<T>) -> Option<T> {
        asymbox.unlock(&self.inner.decrypt, &self.inner.verify)
    }
}

#[cfg(test)]
mod tests {
    use crate::keyring::{Keyring, INDIVIDUAL_KEYS};
    use std::collections::BTreeSet;

    #[test]
    fn test_keyring_partial_copy_full() {
        let kr = Keyring::generate_new_master();

        let keys_to_copy: BTreeSet<&str> = INDIVIDUAL_KEYS.into_iter().cloned().collect();
        assert!(
            kr.partial_copy(&keys_to_copy).unwrap() == kr,
            "keyrings not identical!"
        );
    }
}
