use crate::bloblogs::{BloblogReader, BloblogWriter};
use crate::definitions::{
    BloblogId, Index, IndexId, PackedIndex, PackedPileConfig, PileConfig,
    SUPPORTED_YAMA_PILE_VERSION,
};
use crate::keyring::Keyring;
use crate::locks::{LockHandle, LockKind};
use crate::pointers::{PackedPointer, Pointer};
use crate::utils::HashedWormWriter;
use eyre::{bail, Context, ContextCompat};
use std::collections::{BTreeMap, BTreeSet};
use std::sync::Arc;
use tokio::io::{AsyncReadExt, AsyncWriteExt};
use yama_midlevel_crypto::byte_layer::{ByteLayer, CborSerde};
use yama_midlevel_crypto::zstd_box::Zstd;
use yama_wormfile::paths::{WormPath, WormPathBuf};
use yama_wormfile::{WormFileMeta, WormFileProvider, WormFileWriter};

pub mod definitions;

pub mod bloblogs;
pub mod keyring;
pub mod locks;
pub mod pointers;
pub mod tree;
pub mod utils;

pub const DIR_LOCKS: &'static str = "locks";
pub const DIR_BLOBLOGS: &'static str = "bloblogs";
pub const DIR_INDICES: &'static str = "indices";
pub const FILE_YAMA_CONFIG: &'static str = "yama.cfg";
pub const FILE_YAMA_CONNECTOR: &'static str = "yama.toml";
pub const FILE_MASTER_KEYRING: &'static str = "master.yamakeyring";

pub struct Pile<WFP: WormFileProvider> {
    provider: Arc<WFP>,
    lock: Arc<LockHandle>,
    keyring: Arc<Keyring>,
    pub pile_config: Arc<PileConfig>,
}

impl<WFP: WormFileProvider + 'static> Pile<WFP> {
    pub async fn open_manual(
        provider: Arc<WFP>,
        lock_kind: LockKind,
        holder: String,
        keyring: Keyring,
    ) -> eyre::Result<Self> {
        let lock = Arc::new(LockHandle::new(provider.clone(), lock_kind, holder, &keyring).await?);
        let keyring = Arc::new(keyring);

        let r_config = keyring.r_config.as_ref().context("No r_config key")?;

        let pile_config: PileConfig = {
            let mut file = provider.read(WormPath::new("yama.cfg").unwrap()).await?;
            let mut buf = Vec::new();
            file.read_to_end(&mut buf).await?;
            let packed = PackedPileConfig::from_byte_vec(buf);
            r_config
                .unlock_asymbox(packed)
                .context("Failed to decrypt pile config")?
                .deserialise()
                .context("Failed to deserialise pile config")?
        };

        if &pile_config.yama_pile_version != SUPPORTED_YAMA_PILE_VERSION {
            bail!(
                "Unsupported pile version: {:?} (expected {:?})",
                pile_config.yama_pile_version,
                SUPPORTED_YAMA_PILE_VERSION
            );
        }

        Ok(Pile {
            provider,
            lock,
            keyring,
            pile_config: Arc::new(pile_config),
        })
    }

    /// Creates a new bloblog writer.
    ///
    /// Requires key: w_bloblog_footer, w_bloblog_contents
    pub async fn create_bloblog(&self) -> eyre::Result<BloblogWriter<WFP::WormFileWriter>> {
        if !self.lock.is_active_now(LockKind::Shared) {
            bail!("can't create bloblog: lock not active");
        }
        let writer = BloblogWriter::new(
            self.provider.write().await?,
            &self.keyring,
            self.lock.clone(),
        )
        .await?;
        Ok(writer)
    }

    pub async fn read_bloblog(
        &self,
        bloblog_id: BloblogId,
    ) -> eyre::Result<BloblogReader<WFP::WormFileReader>> {
        let bloblog_path = WormPathBuf::new(format!(
            "bloblogs/{}/{}",
            hex::encode(&bloblog_id.0 .0[0..1]),
            bloblog_id.0.to_string()
        ))
        .unwrap();
        let worm_reader = self.provider.read(&bloblog_path).await?;
        Ok(BloblogReader::new(worm_reader, &self.keyring).await?)
    }

    /// Delete a bloblog from the pile.
    /// This is dangerous: should only really be done in the vacuum operation.
    /// Requires an exclusive lock!
    ///
    ///
    pub async fn delete_bloblog_dangerous_exclusive(
        &self,
        bloblog_id: BloblogId,
    ) -> eyre::Result<()> {
        if !self.lock.is_active_now(LockKind::Exclusive) {
            bail!("can't delete bloblog: exclusive lock not active");
        }
        let bloblog_path = WormPathBuf::new(format!(
            "bloblogs/{}/{}",
            hex::encode(&bloblog_id.0 .0[0..1]),
            bloblog_id.0.to_string()
        ))
        .unwrap();
        self.provider.delete(bloblog_path.as_ref()).await?;
        Ok(())
    }

    /// Create a new index, returning the index ID.
    ///
    /// Requires key: w_bloblog_footer
    pub async fn create_index(&self, index: &Index) -> eyre::Result<IndexId> {
        if !self.lock.is_active_now(LockKind::Shared) {
            bail!("can't create index: lock not active");
        }
        let worm_writer = self.provider.write().await?;
        let mut writer = HashedWormWriter::new(worm_writer);
        let packed_index: PackedIndex = self
            .keyring
            .w_bloblog_footer
            .as_ref()
            .context("No w_bloblog_footer key")?
            .make_locked_asymbox(Zstd::compress(CborSerde::serialise(index)?));
        writer.write_all(&packed_index.into_byte_vec()).await?;
        let (mut worm_writer, hash) = writer.finalise();

        let target = WormPathBuf::new(format!("indices/{}", hash)).unwrap();
        worm_writer.flush().await?;
        worm_writer.finalise(target.as_ref(), false).await?;

        Ok(IndexId(hash))
    }

    /// List all indices present in the pile.
    pub async fn list_indices(&self) -> eyre::Result<BTreeSet<IndexId>> {
        if !self.lock.is_active_now(LockKind::Shared) {
            bail!("can't list indices: lock not active");
        }
        let files = self
            .provider
            .list(WormPath::new("indices").unwrap())
            .await
            .context("failed to list indices")?;
        let mut result = BTreeSet::new();
        for file in files {
            let (_, filename) = file
                .as_ref()
                .as_str()
                .rsplit_once('/')
                .context("index listing entry should split at /")?;
            let index_id = IndexId::try_from(filename)
                .with_context(|| format!("not a valid index ID: {filename:?}"))?;
            result.insert(index_id);
        }
        Ok(result)
    }

    /// List all indices present in the pile, with their metadata.
    pub async fn list_indices_with_meta(&self) -> eyre::Result<BTreeMap<IndexId, WormFileMeta>> {
        if !self.lock.is_active_now(LockKind::Shared) {
            bail!("can't list indices: lock not active");
        }
        let files = self
            .provider
            .list_meta(WormPath::new("indices").unwrap())
            .await
            .context("failed to list indices")?;
        let mut result = BTreeMap::new();
        for (file, meta) in files {
            let (_, filename) = file
                .as_ref()
                .as_str()
                .rsplit_once('/')
                .context("index listing entry should split at /")?;
            let index_id = IndexId::try_from(filename)
                .with_context(|| format!("not a valid index ID: {filename:?}"))?;
            result.insert(index_id, meta);
        }
        Ok(result)
    }

    /// Read an index from the pile.
    ///
    /// Requires key: r_bloblog_footer
    pub async fn read_index(&self, index_id: IndexId) -> eyre::Result<Index> {
        if !self.lock.is_active_now(LockKind::Shared) {
            bail!("can't read index: lock not active");
        }
        let r_bloblog_footer = self
            .keyring
            .r_bloblog_footer
            .as_ref()
            .context("No r_bloblog_footer key")?;
        let target = WormPathBuf::new(format!("indices/{}", index_id.0)).unwrap();
        let mut reader = self.provider.read(target.as_ref()).await?;
        let mut buf = Vec::new();
        reader.read_to_end(&mut buf).await?;
        let packed_index = PackedIndex::from_byte_vec(buf);
        let index = r_bloblog_footer
            .unlock_asymbox(packed_index)
            .context("can't unlock packed index")?
            .decompress()?
            .deserialise()?;
        Ok(index)
    }

    /// Delete an index from the pile.
    /// This is dangerous: should only really be done in the vacuum operation.
    /// Requires an exclusive lock!
    ///
    ///
    pub async fn delete_index_dangerous_exclusive(&self, index_id: IndexId) -> eyre::Result<()> {
        if !self.lock.is_active_now(LockKind::Exclusive) {
            bail!("can't delete index: exclusive lock not active");
        }
        let target = WormPathBuf::new(format!("indices/{}", index_id.0)).unwrap();
        self.provider.delete(target.as_ref()).await?;
        Ok(())
    }

    pub async fn read_pointer(&self, name: &str) -> eyre::Result<Option<Pointer>> {
        let r_pointer = self
            .keyring
            .r_pointer
            .as_ref()
            .context("No r_pointer key")?;

        let pointer_path = WormPath::new("pointers")
            .unwrap()
            .join(name)
            .with_context(|| format!("bad pointer name {name:?}"))?;

        if !self.provider.is_regular_file(pointer_path.as_ref()).await? {
            return Ok(None);
        }

        let mut reader = self.provider.read(pointer_path.as_ref()).await?;
        let mut buf = Vec::new();
        reader.read_to_end(&mut buf).await?;

        let packed_pointer = PackedPointer::from_byte_vec(buf);

        Ok(Some(
            r_pointer
                .unlock_asymbox(packed_pointer)
                .context("failed to decrypt packed pointer")?
                .decompress()
                .context("failed to decompress packed pointer")?
                .deserialise()
                .context("failed to deserialise packed pointer")?,
        ))
    }

    pub async fn write_pointer(
        &self,
        name: &str,
        replace: bool,
        data: &Pointer,
    ) -> eyre::Result<()> {
        let w_pointer = self
            .keyring
            .w_pointer
            .as_ref()
            .context("No w_pointer key")?;

        let packed_pointer: PackedPointer = w_pointer.make_locked_asymbox(Zstd::compress(
            CborSerde::serialise(data).context("can't serialise pointer")?,
        ));

        let pointer_path = WormPath::new("pointers")
            .unwrap()
            .join(name)
            .with_context(|| format!("bad pointer name {name:?}"))?;

        let mut writer = self.provider.write().await?;
        writer.write_all(&packed_pointer.into_byte_vec()).await?;
        writer.finalise(pointer_path.as_ref(), replace).await?;
        Ok(())
    }

    pub async fn list_pointers(&self) -> eyre::Result<Vec<String>> {
        let files = self
            .provider
            .list(WormPath::new("pointers").unwrap())
            .await?;
        Ok(files
            .into_iter()
            .map(|file| {
                let (_dir, pointer) = file.as_ref().as_str().rsplit_once('/').unwrap();
                pointer.to_owned()
            })
            .collect())
    }

    pub async fn close(self) -> eyre::Result<()> {
        match Arc::try_unwrap(self.lock) {
            Ok(lock) => {
                lock.close()
                    .await
                    .context("failed to release lock gracefully")?;
            }
            Err(arc) => {
                bail!(
                    "could not close pile gracefully; lock arc has {} strong refs and {} weak refs",
                    Arc::strong_count(&arc),
                    Arc::weak_count(&arc)
                );
            }
        }
        Ok(())
    }
}
