use crate::keyring::{Keyring, ReaderKey, WriterKey};
use chrono::{DateTime, Duration, Utc};
use eyre::{bail, eyre, Context, ContextCompat};
use serde::{Deserialize, Serialize};
use std::collections::BTreeMap;
use std::sync::{Arc, RwLock};
use tokio::io::{AsyncReadExt, AsyncWriteExt};
use tokio::sync::oneshot;
use tokio::task::JoinHandle;
use tokio::time::Instant;
use tracing::{error, info, warn};
use uuid::Uuid;
use yama_midlevel_crypto::asym_box::AsymBox;
use yama_midlevel_crypto::byte_layer::{ByteLayer, CborSerde};
use yama_wormfile::paths::{WormPath, WormPathBuf};
use yama_wormfile::{WormFileProvider, WormFileWriter};

#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct LockDesc {
    /// Expiry time of the lock. Should be renewed occasionally.
    /// Should not exceed more than 10 minutes in the future.
    expires_at: DateTime<Utc>,

    /// Human-friendly description of who is holding the lock.
    holder: String,

    /// What kind of lock this is.
    kind: LockKind,
}

pub type PackedLock = AsymBox<CborSerde<LockDesc>>;

#[derive(Copy, Clone, Debug, Eq, PartialEq, Serialize, Deserialize)]
pub enum LockKind {
    /// Lock can coexist with other Shared and PendingExclusive locks.
    /// Lock may not be created when there are non-Shared locks.
    Shared,

    /// Lock can coexist with Shared locks.
    PendingExclusive,

    /// Lock can not coexist with other locks at all.
    Exclusive,
}

impl LockKind {
    pub fn conflicts_with_lock(self, newer_lock: LockKind) -> bool {
        match (self, newer_lock) {
            (LockKind::Shared, LockKind::Shared) => false,
            // We can add PendingExclusives after a Shared, but not the other way around.
            (LockKind::Shared, LockKind::PendingExclusive) => false,
            _ => true,
        }
    }
}

#[derive(Debug)]
pub struct LockHandle {
    /// The latest lock descriptor
    lock: Arc<RwLock<LockDesc>>,

    /// Our lock ID
    lock_id: String,

    /// Path to the lock
    lock_path: WormPathBuf,

    /// A signal for relinquishing the lock.
    lock_release_tx: Option<oneshot::Sender<()>>,

    /// Handle for waiting for a graceful shutdown of the lock.
    lock_task_join_handle: Option<JoinHandle<()>>,
}

impl Drop for LockHandle {
    fn drop(&mut self) {
        if let Some(lock_release_tx) = self.lock_release_tx.take() {
            lock_release_tx.send(()).expect("can't drop lock");
        }
    }
}

impl LockHandle {
    pub async fn close(mut self) -> eyre::Result<()> {
        self.lock_release_tx
            .take()
            .unwrap()
            .send(())
            .map_err(|_| eyre!("can't drop lock"))?;
        self.lock_task_join_handle
            .take()
            .unwrap()
            .await
            .context("lock task fail")?;
        Ok(())
    }

    pub fn is_active_at(&self, kind: LockKind, now: DateTime<Utc>) -> bool {
        let lock = self.lock.read().unwrap();
        lock.kind == kind || kind == LockKind::Shared && lock.expires_at > now
    }

    /// Returns true iff the lock is active now.
    /// This actually looks 1 minute into the future to afford some divergence for clock drift.
    pub fn is_active_now(&self, kind: LockKind) -> bool {
        let now = Utc::now() + Duration::minutes(1);
        self.is_active_at(kind, now)
    }

    pub async fn new(
        provider: Arc<impl WormFileProvider + 'static>,
        kind: LockKind,
        holder: String,
        keyring: &Keyring,
    ) -> eyre::Result<LockHandle> {
        let duration = Duration::minutes(10);
        let until = Utc::now() + duration;
        Self::new_until(provider, kind, holder, keyring, until, duration).await
    }

    pub async fn new_until(
        provider: Arc<impl WormFileProvider + 'static>,
        kind: LockKind,
        holder: String,
        keyring: &Keyring,
        until: DateTime<Utc>,
        duration: Duration,
    ) -> eyre::Result<LockHandle> {
        let lock_desc = LockDesc {
            expires_at: until,
            holder,
            kind,
        };
        let w_locks = keyring.w_locks.clone().context("No w_locks key")?;
        let r_locks = keyring.r_locks.clone().context("No r_locks key")?;

        'retry: loop {
            let lock_id = Uuid::new_v4().to_string();
            let lock_path = WormPathBuf::new(format!("locks/{lock_id}")).unwrap();
            let (lock_release_tx, lock_release_rx) = oneshot::channel();

            let now = Utc::now();

            // Stage 1: create lock
            let stage1_locks = scan_locks(provider.as_ref(), &r_locks, now).await?;
            if let Some(blocker) = find_lock_blocker(&stage1_locks, &lock_id, kind) {
                let lock = &stage1_locks[blocker];
                warn!("{:?} lock {} held by {:?} currently expiring at {} is blocking our potential lock.", lock.kind, lock_id, lock.holder, lock.expires_at);

                tokio::time::sleep(tokio::time::Duration::from_secs(
                    (lock.expires_at - now).num_seconds().max(0) as u64 + 10,
                ))
                .await;
                continue 'retry;
            }

            let mut writer = provider.write().await?;
            let packed_lock: PackedLock =
                w_locks.make_locked_asymbox(CborSerde::serialise(&lock_desc)?);
            writer.write_all(&packed_lock.into_byte_vec()).await?;
            writer.flush().await?;
            writer.finalise(lock_path.as_ref(), false).await?;

            // Stage 2: confirm lock
            let stage2_locks = scan_locks(provider.as_ref(), &r_locks, now).await?;
            if let Some(blocker) = find_lock_blocker(&stage2_locks, &lock_id, kind) {
                let lock = &stage2_locks[blocker];
                warn!("{:?} lock {} held by {:?} currently expiring at {} blocked our lock; backing out.", lock.kind, lock_id, lock.holder, lock.expires_at);

                // Back out our lock.
                provider.delete(lock_path.as_ref()).await?;
                tokio::time::sleep(tokio::time::Duration::from_secs(10)).await;
                continue 'retry;
            }

            info!("Acquired {:?} lock {}", kind, lock_id);

            let lock = Arc::new(RwLock::new(lock_desc));
            let lock2 = lock.clone();

            // Good. Now start a background task for refreshing it as necessary.
            // TODO spawn this onto a joinset and then make sure we release locks at end of program...
            let lock_path2 = lock_path.clone();
            let lock_task_join_handle = Some(tokio::spawn(async_backtrace::frame!(async move {
                if let Err(err) = lock_renewal(
                    provider,
                    lock_path2,
                    lock2,
                    lock_release_rx,
                    w_locks,
                    duration,
                )
                .await
                {
                    error!("Lock renewal task failed: {err:?}");
                }
            })));

            break Ok(LockHandle {
                lock,
                lock_path,
                lock_id,
                lock_release_tx: Some(lock_release_tx),
                lock_task_join_handle,
            });
        }
    }

    pub async fn upgrade_pending_exclusive(
        &self,
        provider: Arc<impl WormFileProvider + 'static>,
        r_locks: &ReaderKey,
        w_locks: &WriterKey,
    ) -> eyre::Result<()> {
        'retry: loop {
            if !self.is_active_now(LockKind::PendingExclusive) {
                bail!("PendingExclusive not active: can't upgrade to Exclusive");
            }

            let now = Utc::now();
            let locks = scan_locks(provider.as_ref(), r_locks, Utc::now()).await?;
            if let Some((conflicting_lock_id, conflicting_lock)) =
                locks.iter().find(|(lock_id, _)| lock_id != &&self.lock_id)
            {
                warn!("Conflicting {:?} lock {} held by {:?} expiring at {} is blocking us from upgrading to Exclusive", conflicting_lock.kind, conflicting_lock_id, conflicting_lock.holder, conflicting_lock.expires_at);
                tokio::time::sleep(tokio::time::Duration::from_secs(
                    (conflicting_lock.expires_at - now).num_seconds().max(0) as u64 + 10,
                ))
                .await;
                continue 'retry;
            }

            let mut lock = self.lock.write().unwrap();
            lock.kind = LockKind::Exclusive;
            let mut writer = provider.write().await?;
            let packed_lock: PackedLock =
                w_locks.make_locked_asymbox(CborSerde::serialise(&*lock)?);
            writer.write_all(&packed_lock.into_byte_vec()).await?;
            writer.flush().await?;
            writer.finalise(self.lock_path.as_ref(), true).await?;

            break Ok(());
        }
    }
}

async fn lock_renewal(
    provider: Arc<impl WormFileProvider>,
    lock_path: WormPathBuf,
    lock: Arc<RwLock<LockDesc>>,
    mut lock_release_rx: oneshot::Receiver<()>,
    w_locks: WriterKey,
    duration: Duration,
) -> eyre::Result<()> {
    loop {
        let next_renewal = {
            let lock = lock.read().unwrap();
            let secs_until_renewal = ((lock.expires_at - Utc::now()).num_seconds() - 60).max(0);
            Instant::now() + tokio::time::Duration::from_secs(secs_until_renewal as u64)
        };

        tokio::select! {
            _ = &mut lock_release_rx => {
                provider.delete(lock_path.as_ref()).await
                    .context("failed to delete lock upon release")?;
                break Ok(());
            },
            _ = tokio::time::sleep_until(next_renewal) => {
                // nop.
            }
        }

        let old_lock_desc: LockDesc = {
            let lock = lock.read().unwrap();
            lock.clone()
        };

        let now = Utc::now();
        if now > old_lock_desc.expires_at {
            error!("Missed lock renewal; {now} > {}", old_lock_desc.expires_at);
        }

        let new_lock_desc = LockDesc {
            expires_at: now + duration,
            ..old_lock_desc.clone()
        };

        let mut writer = provider
            .write()
            .await
            .context("failed to acquire writing handle for write of packed lock")?;
        let packed_lock: PackedLock = w_locks.make_locked_asymbox(
            CborSerde::serialise(&new_lock_desc)
                .context("failed to serialise new lock description")?,
        );
        writer
            .write_all(&packed_lock.into_byte_vec())
            .await
            .context("failed to write packed lock")?;
        writer
            .flush()
            .await
            .context("failed to flush write of packed lock")?;
        writer
            .finalise(lock_path.as_ref(), true)
            .await
            .context("failed to finalise write of packed lock")?;

        *(lock.write().unwrap()) = new_lock_desc;
    }
}

async fn scan_locks(
    provider: &impl WormFileProvider,
    r_locks: &ReaderKey,
    now: DateTime<Utc>,
) -> eyre::Result<BTreeMap<String, LockDesc>> {
    let mut result = BTreeMap::new();

    let files = provider
        .list(WormPath::new("locks").unwrap())
        .await
        .context("failed to list locks")?;
    for file in files {
        let (_, lock_id) = file
            .as_ref()
            .as_str()
            .rsplit_once('/')
            .context("bad lock split")?;

        let mut reader = provider.read(&file).await?;
        let mut buf = Vec::new();
        reader.read_to_end(&mut buf).await?;
        let packed_lock = PackedLock::from_byte_vec(buf);
        let lock_desc = r_locks
            .unlock_asymbox(packed_lock)
            .context("failed to decrypt lock")?
            .deserialise()?;

        if lock_desc.expires_at > now {
            result.insert(lock_id.to_owned(), lock_desc);
        }
    }

    Ok(result)
}

fn find_lock_blocker<'a>(
    locks: &'a BTreeMap<String, LockDesc>,
    our_lock_id: &'a str,
    our_lock_kind: LockKind,
) -> Option<&'a str> {
    for (lock_id, lock_desc) in locks {
        if lock_id == our_lock_id {
            continue;
        }
        if lock_desc.kind.conflicts_with_lock(our_lock_kind) {
            return Some(lock_id);
        }
    }
    None
}
