/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/

use serde::{Deserialize, Serialize};
use std::collections::btree_map::Entry;
use std::collections::BTreeMap;
use std::fmt::{Debug, Formatter};
use std::fs::Metadata;
use std::ops::{Add, AddAssign};
use std::os::unix::fs::MetadataExt;

use crate::definitions::RecursiveChunkRef;
use crate::tree::unpopulated::ScanEntry;
use eyre::{bail, Context, ContextCompat};
use patricia_tree::PatriciaMap;

pub mod unpopulated;

/// Given a file's metadata, returns the mtime in milliseconds.
pub fn mtime_msec(metadata: &Metadata) -> u64 {
    (metadata.mtime() * 1000 + metadata.mtime_nsec() / 1_000_000) as u64
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct RootTreeNode {
    pub name: String,
    pub node: TreeNode,
}

#[derive(Debug, Clone, Serialize, Deserialize, Eq, PartialEq)]
#[serde(tag = "t")]
pub enum TreeNode {
    #[serde(rename = "F")]
    NormalFile {
        /// modification time in ms
        mtime: u64,
        #[serde(flatten)]
        ownership: FilesystemOwnership,
        #[serde(flatten)]
        permissions: FilesystemPermissions,
        size: u64,
        #[serde(flatten)]
        content: RecursiveChunkRef,
    },
    #[serde(rename = "D")]
    Directory {
        #[serde(flatten)]
        ownership: FilesystemOwnership,
        #[serde(flatten)]
        permissions: FilesystemPermissions,
        children: BTreeMap<String, TreeNode>,
    },
    #[serde(rename = "L")]
    SymbolicLink {
        #[serde(flatten)]
        ownership: FilesystemOwnership,
        target: String,
    },
    // TODO is there any other kind of file we need to store?
    #[serde(rename = "X")]
    Deleted,
}

impl TreeNode {
    pub fn metadata_invalidates(&self, other: &TreeNode) -> eyre::Result<bool> {
        Ok(match self {
            TreeNode::NormalFile {
                mtime,
                ownership,
                permissions,
                size,
                ..
            } => {
                if let TreeNode::NormalFile {
                    mtime: other_mtime,
                    ownership: other_ownership,
                    permissions: other_permissions,
                    size: other_size,
                    ..
                } = other
                {
                    mtime != other_mtime
                        || size != other_size
                        || ownership != other_ownership
                        || permissions != other_permissions
                } else {
                    true
                }
            }
            TreeNode::Directory {
                ownership,
                permissions,
                children,
            } => {
                if let TreeNode::Directory {
                    ownership: other_ownership,
                    permissions: other_permissions,
                    children: other_children,
                } = other
                {
                    if ownership != other_ownership || permissions != other_permissions {
                        return Ok(true);
                    }
                    if children.len() != other_children.len() {
                        return Ok(true);
                    }
                    for ((left_name, left_node), (right_name, right_node)) in
                        children.iter().zip(other_children.iter())
                    {
                        if left_name != right_name || left_node.metadata_invalidates(right_node)? {
                            return Ok(true);
                        }
                    }
                    false
                } else {
                    true
                }
            }
            TreeNode::SymbolicLink { ownership, target } => {
                if let TreeNode::SymbolicLink {
                    ownership: other_ownership,
                    target: other_target,
                } = other
                {
                    ownership != other_ownership || target != other_target
                } else {
                    true
                }
            }
            TreeNode::Deleted => {
                // unreachable
                bail!("Why is Deleted here?");
            }
        })
    }

    pub fn count_normal_files(&self) -> u32 {
        match self {
            TreeNode::NormalFile { .. } => 1,
            TreeNode::Directory { children, .. } => children
                .values()
                .map(|child| child.count_normal_files())
                .sum(),
            _ => 0,
        }
    }

    pub fn visit<F: FnMut(&TreeNode, &str) -> eyre::Result<()>>(
        &self,
        func: &mut F,
        prefix: String,
    ) -> eyre::Result<()> {
        func(self, &prefix)?;
        if let TreeNode::Directory { children, .. } = &self {
            for (name, child) in children.iter() {
                if prefix.is_empty() {
                    // don't want a slash prefix
                    child.visit(func, name.clone())?;
                } else {
                    child.visit(func, format!("{}/{}", prefix, name))?;
                }
            }
        }
        Ok(())
    }

    pub fn visit_mut<F: FnMut(&mut TreeNode, &str) -> eyre::Result<()>>(
        &mut self,
        func: &mut F,
        prefix: String,
    ) -> eyre::Result<()> {
        func(self, &prefix)?;
        if let TreeNode::Directory { children, .. } = self {
            for (name, child) in children.iter_mut() {
                if prefix.is_empty() {
                    // don't want a slash prefix
                    child.visit_mut(func, name.clone())?;
                } else {
                    child.visit_mut(func, format!("{}/{}", prefix, name))?;
                }
            }
        }
        Ok(())
    }

    pub fn exists(&self, pieces: &[&str]) -> bool {
        match pieces.first() {
            None => true,
            Some(subpath) => {
                if let TreeNode::Directory { children, .. } = self {
                    match children.get(*subpath) {
                        None => false,
                        Some(child) => child.exists(&pieces[1..]),
                    }
                } else {
                    false
                }
            }
        }
    }

    /// Recurses into a child by name, or returns Err with a reason.
    pub fn child(&mut self, name: &str) -> Result<&mut TreeNode, &'static str> {
        match self {
            TreeNode::NormalFile { .. } => Err("not a directory: normal file"),
            TreeNode::Directory { children, .. } => match children.get_mut(name) {
                None => Err("child not in directory"),
                Some(node) => Ok(node),
            },
            TreeNode::SymbolicLink { .. } => Err("not a directory: symlink"),
            TreeNode::Deleted => Err("not a directory: deleted"),
        }
    }

    /// Flattens this treenode to a PatriciaMap of ScanEntries.
    pub fn flatten(&self) -> eyre::Result<PatriciaMap<ScanEntry>> {
        let mut out = PatriciaMap::new();
        Self::flatten_impl(self, "", &mut out)?;
        Ok(out)
    }

    fn flatten_impl(
        tree_node: &TreeNode,
        prefix: &str,
        out: &mut PatriciaMap<ScanEntry>,
    ) -> eyre::Result<()> {
        match tree_node {
            TreeNode::NormalFile {
                mtime,
                ownership,
                permissions,
                size,
                content: _,
            } => {
                out.insert(
                    prefix,
                    ScanEntry::NormalFile {
                        mtime: *mtime,
                        ownership: ownership.clone(),
                        permissions: permissions.clone(),
                        size: *size,
                    },
                );
            }
            TreeNode::Directory {
                ownership,
                permissions,
                children,
            } => {
                out.insert(
                    prefix,
                    ScanEntry::Directory {
                        ownership: ownership.clone(),
                        permissions: permissions.clone(),
                    },
                );

                for (child_name, node) in children {
                    let new_prefix = if prefix.is_empty() {
                        child_name.clone()
                    } else {
                        format!("{}/{}", prefix, child_name)
                    };
                    Self::flatten_impl(node, &new_prefix, out)?;
                }
            }
            TreeNode::SymbolicLink { ownership, target } => {
                out.insert(
                    prefix,
                    ScanEntry::SymbolicLink {
                        ownership: ownership.clone(),
                        target: target.clone(),
                    },
                );
            }
            TreeNode::Deleted => {
                bail!("found Deleted at {prefix:?} when flattening");
            }
        }
        Ok(())
    }
}

#[derive(Copy, Clone, Serialize, Deserialize, PartialEq, Eq)]
pub struct FilesystemOwnership {
    pub uid: u16,
    pub gid: u16,
}

impl Debug for FilesystemOwnership {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}:{}", self.uid, self.gid)
    }
}

#[derive(Copy, Clone, Serialize, Deserialize, PartialEq, Eq)]
pub struct FilesystemPermissions {
    pub mode: u32,
}

impl Debug for FilesystemPermissions {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", unix_mode::to_string(self.mode))
    }
}

/// Differentiates a node in place.
/// This makes `old` the parent of `new` (though it is up to the caller to properly update the
/// `PointerData` to reflect this!).
/// Loosely speaking, `new` is modified to contain the differences that, when applied to `old`, will
/// result in the original value of `new`.
/// See `integrate_node_in_place` for the inverse of this operation.
pub fn differentiate_node_in_place(new: &mut TreeNode, old: &TreeNode) -> eyre::Result<()> {
    if let TreeNode::Directory { children, .. } = new {
        if let TreeNode::Directory {
            children: old_children,
            ..
        } = old
        {
            for (name, old_node) in old_children.iter() {
                match children.entry(name.clone()) {
                    Entry::Vacant(vacant_entry) => {
                        vacant_entry.insert(TreeNode::Deleted);
                    }
                    Entry::Occupied(occupied_entry) => {
                        if !occupied_entry.get().metadata_invalidates(old_node)? {
                            // The entry is the same, so we don't need to store it!
                            occupied_entry.remove_entry();
                        } else {
                            differentiate_node_in_place(occupied_entry.into_mut(), old_node)?;
                        }
                    }
                }
            }
        }
    }
    Ok(())
}

#[derive(Clone, Debug, Default)]
pub struct IntegrationStats {
    pub descends: u64,
    pub inserts: u64,
    pub deletes: u64,
    pub nops: u64,
}

impl AddAssign for IntegrationStats {
    fn add_assign(&mut self, rhs: Self) {
        self.descends += rhs.descends;
        self.inserts += rhs.inserts;
        self.deletes += rhs.deletes;
        self.nops += rhs.nops;
    }
}

/// Integrates a node in place.
/// This makes `new` no longer have a parent (remember, the caller is responsible for updating
/// `PointerData` appropriately if needed to reflect this).
///
/// Loosely speaking, `new` is treated as a set of differences that are applied to `old`, though the
/// result is in-place.
///
/// Preconditions:
/// - `old` must be an integrated pointer. (Otherwise this algorithm is not correct.)
/// - `old` is the parent of `new`
pub fn integrate_node_in_place(new: &mut TreeNode, old: &TreeNode) -> IntegrationStats {
    let mut stats = IntegrationStats::default();
    if let TreeNode::Directory { children, .. } = new {
        if let TreeNode::Directory {
            children: old_children,
            ..
        } = old
        {
            for (name, node) in old_children.iter() {
                match children.entry(name.clone()) {
                    Entry::Vacant(entry) => {
                        entry.insert(node.clone());
                        stats.inserts += 1;
                    }
                    Entry::Occupied(entry) => {
                        if entry.get() == &TreeNode::Deleted {
                            // We don't insert the old node but we do remove the 'deleted' marker
                            // node!
                            entry.remove();
                            stats.deletes += 1;
                        } else {
                            stats += integrate_node_in_place(entry.into_mut(), node);
                        }
                    }
                }
            }
        }
    } else {
        // the node stays the same...
        // intentional NOP!
        stats.nops += 1;
    }

    stats
}

/// `new_scan` and `reused_scan` should be disjoint, but `reused_scan` must contain all directories
/// needed for itself; `new_scan` is allowed to reuse those directories.
pub fn assemble_tree_from_scan_entries(
    new_scan: PatriciaMap<ScanEntry>,
    reused_scan: PatriciaMap<ScanEntry>,
    mut chunkings: PatriciaMap<Option<(RecursiveChunkRef, u64)>>,
) -> eyre::Result<TreeNode> {
    let mut dirs: BTreeMap<String, BTreeMap<String, TreeNode>> = BTreeMap::new();

    for (key, entry) in reused_scan.into_iter().chain(new_scan.into_iter()) {
        let key_string = String::from_utf8(key).context("bad UTF-8 in PMap")?;
        let (parent_dir_name, child_name) =
            key_string.rsplit_once('/').unwrap_or(("", &key_string));
        match entry {
            ScanEntry::NormalFile {
                mtime,
                ownership,
                permissions,
                size: _unverified_size_ignore,
            } => {
                if let Some((content, size)) = chunkings
                    .remove(&key_string)
                    .with_context(|| format!("bad chunkings PMap: missing entry: {key_string:?}"))?
                {
                    // note: for the root, this inserts the root file entry as a child called "" within a fake root 'directory'.
                    // That's fine. We'll patch this up later.
                    dirs.get_mut(parent_dir_name)
                        .with_context(|| {
                            format!("bad PMap: parent {parent_dir_name} not seen first")
                        })?
                        .insert(
                            child_name.to_owned(),
                            TreeNode::NormalFile {
                                mtime,
                                ownership,
                                permissions,
                                size,
                                content,
                            },
                        );
                }
            }
            ScanEntry::Directory {
                ownership,
                permissions,
            } => {
                if !dirs.insert(key_string.clone(), BTreeMap::new()).is_none() {
                    bail!("bug: {key_string:?} already in dirs...");
                }
                // note: for the root, this inserts the root directory entry as a child called "" within the root.
                // That's fine. We'll patch this up later.
                dirs.get_mut(parent_dir_name)
                    .context("bad PMap: parent not seen first")?
                    .insert(
                        child_name.to_owned(),
                        TreeNode::Directory {
                            ownership,
                            permissions,
                            children: BTreeMap::new(),
                        },
                    );
            }
            ScanEntry::SymbolicLink { ownership, target } => {
                // note: for the root, this inserts the root symlink entry as a child called "" within a fake root 'directory'.
                // That's fine. We'll patch this up later.
                dirs.get_mut(parent_dir_name)
                    .context("bad PMap: parent not seen first")?
                    .insert(
                        child_name.to_owned(),
                        TreeNode::SymbolicLink { ownership, target },
                    );
            }
        }
    }

    // Now roll up the directories. In Rustc v1.66 it'd be nice to use pop_last()...
    while let Some(last_key) = dirs.keys().last().cloned() {
        let mut last_children = dirs.remove(&last_key).unwrap();
        if last_key.is_empty() {
            assert!(
                dirs.is_empty(),
                "when pulling out root pseudo-dir, dirs must be empty for roll-up."
            );

            let mut real_root = last_children.remove("").unwrap();
            if let TreeNode::Directory { children, .. } = &mut real_root {
                *children = last_children;
            } else if !last_children.is_empty() {
                bail!("root is not a directory but it contains children...");
            }

            return Ok(real_root);
        }

        // We want to roll up the directory last/key -> {child -> ...}
        // so last -> {key -> {child -> ...}}
        let (parent_dir, child_name) = last_key.rsplit_once('/').unwrap_or(("", &last_key));
        let parent = dirs
            .get_mut(parent_dir)
            .context("bad PMap? no parent in rollup")?;
        let child_in_parent = parent
            .get_mut(child_name)
            .context("dir child not populated")?;
        if let TreeNode::Directory { children, .. } = child_in_parent {
            *children = last_children;
        } else {
            bail!("child in parent not a directory...");
        }
    }

    bail!("no root found; bad PMap or bad roll-up???");
}

#[cfg(test)]
mod tests {
    use crate::definitions::RecursiveChunkRef;
    use crate::tree::{
        differentiate_node_in_place, integrate_node_in_place, FilesystemOwnership,
        FilesystemPermissions, TreeNode,
    };
    use std::collections::BTreeMap;
    use yama_midlevel_crypto::chunk_id::ChunkId;

    #[test]
    pub fn test_differentiate_in_place_primitive() {
        let mut new = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 42, gid: 47 },
            permissions: FilesystemPermissions { mode: 1337 },
            children: BTreeMap::new(),
        };
        let old = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 42, gid: 47 },
            permissions: FilesystemPermissions { mode: 1337 },
            children: BTreeMap::new(),
        };
        assert!(differentiate_node_in_place(&mut new, &old).is_ok());

        assert_eq!(
            new,
            TreeNode::Directory {
                ownership: FilesystemOwnership { uid: 42, gid: 47 },
                permissions: FilesystemPermissions { mode: 1337 },
                children: BTreeMap::new()
            }
        );
    }

    /// Tests the creation of a file in the tree.
    #[test]
    pub fn test_differentiate_in_place_create_only() {
        let alice = TreeNode::NormalFile {
            mtime: 98347523,
            ownership: FilesystemOwnership { uid: 43, gid: 48 },
            permissions: FilesystemPermissions { mode: 1338 },
            size: 42,
            content: RecursiveChunkRef {
                chunk_id: ChunkId::from([36; 32]),
                depth: 22,
            },
        };
        let bob_new = TreeNode::SymbolicLink {
            ownership: FilesystemOwnership { uid: 43, gid: 48 },
            target: "alice".to_string(),
        };

        let mut children_new = BTreeMap::new();
        children_new.insert("bob".to_owned(), bob_new.clone());
        children_new.insert("alice".to_owned(), alice.clone());

        let mut children_old = BTreeMap::new();
        children_old.insert("alice".to_owned(), alice.clone());

        let mut new = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 42, gid: 47 },
            permissions: FilesystemPermissions { mode: 1337 },
            children: children_new.clone(),
        };
        let old = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 41, gid: 46 },
            permissions: FilesystemPermissions { mode: 1336 },
            children: children_old.clone(),
        };

        let mut children_result = BTreeMap::new();
        children_result.insert("bob".to_owned(), bob_new);

        assert!(differentiate_node_in_place(&mut new, &old).is_ok());
        assert_eq!(
            new,
            TreeNode::Directory {
                ownership: FilesystemOwnership { uid: 42, gid: 47 },
                permissions: FilesystemPermissions { mode: 1337 },
                children: children_result
            }
        );
    }

    /// Tests only a change in metadata in the tree.
    #[test]
    pub fn test_differentiate_in_place_meta_only() {
        let alice = TreeNode::NormalFile {
            mtime: 98347523,
            ownership: FilesystemOwnership { uid: 43, gid: 48 },
            permissions: FilesystemPermissions { mode: 1338 },
            size: 42,
            content: RecursiveChunkRef {
                chunk_id: ChunkId::from([36; 32]),
                depth: 22,
            },
        };

        let mut children = BTreeMap::new();
        children.insert("alice".to_owned(), alice);

        let mut new = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 42, gid: 47 },
            permissions: FilesystemPermissions { mode: 1337 },
            children: children.clone(),
        };
        let old = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 41, gid: 46 },
            permissions: FilesystemPermissions { mode: 1336 },
            children: children.clone(),
        };

        assert!(differentiate_node_in_place(&mut new, &old).is_ok());
        assert_eq!(
            new,
            TreeNode::Directory {
                ownership: FilesystemOwnership { uid: 42, gid: 47 },
                permissions: FilesystemPermissions { mode: 1337 },
                children: BTreeMap::new()
            }
        );
    }

    /// Tests that nodes that vanish lead to creation of a Deleted node.
    #[test]
    pub fn test_differences_in_place_deleted() {
        let alice = TreeNode::NormalFile {
            mtime: 98347523,
            ownership: FilesystemOwnership { uid: 43, gid: 48 },
            permissions: FilesystemPermissions { mode: 1338 },
            size: 42,
            content: RecursiveChunkRef {
                chunk_id: ChunkId::from([36; 32]),
                depth: 22,
            },
        };
        let bob_old = TreeNode::SymbolicLink {
            ownership: FilesystemOwnership { uid: 43, gid: 48 },
            target: "alice".to_string(),
        };

        let mut children_old = BTreeMap::new();
        children_old.insert("bob".to_owned(), bob_old.clone());
        children_old.insert("alice".to_owned(), alice.clone());

        let mut children_new = BTreeMap::new();
        children_new.insert("alice".to_owned(), alice.clone());

        let old = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 42, gid: 47 },
            permissions: FilesystemPermissions { mode: 1337 },
            children: children_old.clone(),
        };
        let mut new = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 41, gid: 46 },
            permissions: FilesystemPermissions { mode: 1336 },
            children: children_new.clone(),
        };

        let mut children_result = BTreeMap::new();
        children_result.insert("bob".to_owned(), TreeNode::Deleted);

        assert!(differentiate_node_in_place(&mut new, &old).is_ok());
        assert_eq!(
            new,
            TreeNode::Directory {
                ownership: FilesystemOwnership { uid: 41, gid: 46 },
                permissions: FilesystemPermissions { mode: 1336 },
                children: children_result
            }
        );
    }

    #[test]
    pub fn test_differentiate_node_in_place_mega_example() {
        // TODO extend this example
        let parent = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 47, gid: 49 },
            permissions: FilesystemPermissions { mode: 0660 },
            children: vec![(
                "dir1".to_string(),
                TreeNode::Directory {
                    ownership: FilesystemOwnership { uid: 46, gid: 50 },
                    permissions: FilesystemPermissions { mode: 0550 },
                    children: vec![
                        (
                            "file1".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1996,
                                ownership: FilesystemOwnership { uid: 54, gid: 59 },
                                permissions: FilesystemPermissions { mode: 0311 },
                                size: 42,
                                content: RecursiveChunkRef {
                                    chunk_id: ChunkId::from([37; 32]),
                                    depth: 2,
                                },
                            },
                        ),
                        (
                            "file2".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1970,
                                ownership: FilesystemOwnership { uid: 55, gid: 60 },
                                permissions: FilesystemPermissions { mode: 0321 },
                                size: 42,
                                content: RecursiveChunkRef {
                                    chunk_id: ChunkId::from([42; 32]),
                                    depth: 29,
                                },
                            },
                        ),
                    ]
                    .into_iter()
                    .collect(),
                },
            )]
            .into_iter()
            .collect(),
        };

        let child_full = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 47, gid: 49 },
            permissions: FilesystemPermissions { mode: 0660 },
            children: vec![(
                "dir1".to_string(),
                TreeNode::Directory {
                    ownership: FilesystemOwnership { uid: 46, gid: 50 },
                    permissions: FilesystemPermissions { mode: 0560 },
                    children: vec![
                        (
                            "file1".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1996,
                                ownership: FilesystemOwnership { uid: 54, gid: 59 },
                                permissions: FilesystemPermissions { mode: 0311 },
                                size: 42,
                                content: RecursiveChunkRef {
                                    chunk_id: ChunkId::from([37; 32]),
                                    depth: 2,
                                },
                            },
                        ),
                        (
                            "file42".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1970,
                                ownership: FilesystemOwnership { uid: 55, gid: 60 },
                                permissions: FilesystemPermissions { mode: 0321 },
                                size: 42,
                                content: RecursiveChunkRef {
                                    chunk_id: ChunkId::from([42; 32]),
                                    depth: 29,
                                },
                            },
                        ),
                    ]
                    .into_iter()
                    .collect(),
                },
            )]
            .into_iter()
            .collect(),
        };

        let mut child_diff = child_full.clone();
        differentiate_node_in_place(&mut child_diff, &parent).unwrap();

        let expected_child_diff = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 47, gid: 49 },
            permissions: FilesystemPermissions { mode: 0660 },
            children: vec![(
                "dir1".to_string(),
                TreeNode::Directory {
                    ownership: FilesystemOwnership { uid: 46, gid: 50 },
                    permissions: FilesystemPermissions { mode: 0560 },
                    children: vec![
                        ("file2".to_string(), TreeNode::Deleted),
                        (
                            "file42".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1970,
                                ownership: FilesystemOwnership { uid: 55, gid: 60 },
                                permissions: FilesystemPermissions { mode: 0321 },
                                size: 42,
                                content: RecursiveChunkRef {
                                    chunk_id: ChunkId::from([42; 32]),
                                    depth: 29,
                                },
                            },
                        ),
                    ]
                    .into_iter()
                    .collect(),
                },
            )]
            .into_iter()
            .collect(),
        };

        assert_eq!(child_diff, expected_child_diff);
    }

    #[test]
    pub fn test_integrate_node_in_place_mega_example() {
        // TODO extend this example
        let parent = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 47, gid: 49 },
            permissions: FilesystemPermissions { mode: 0660 },
            children: vec![(
                "dir1".to_string(),
                TreeNode::Directory {
                    ownership: FilesystemOwnership { uid: 46, gid: 50 },
                    permissions: FilesystemPermissions { mode: 0550 },
                    children: vec![
                        (
                            "file1".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1996,
                                ownership: FilesystemOwnership { uid: 54, gid: 59 },
                                permissions: FilesystemPermissions { mode: 0311 },
                                size: 42,
                                content: RecursiveChunkRef {
                                    chunk_id: ChunkId::from([37; 32]),
                                    depth: 2,
                                },
                            },
                        ),
                        (
                            "file2".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1970,
                                ownership: FilesystemOwnership { uid: 55, gid: 60 },
                                permissions: FilesystemPermissions { mode: 0321 },
                                size: 42,
                                content: RecursiveChunkRef {
                                    chunk_id: ChunkId::from([42; 32]),
                                    depth: 29,
                                },
                            },
                        ),
                    ]
                    .into_iter()
                    .collect(),
                },
            )]
            .into_iter()
            .collect(),
        };

        let child_diff = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 47, gid: 49 },
            permissions: FilesystemPermissions { mode: 0660 },
            children: vec![(
                "dir1".to_string(),
                TreeNode::Directory {
                    ownership: FilesystemOwnership { uid: 46, gid: 50 },
                    permissions: FilesystemPermissions { mode: 0560 },
                    children: vec![
                        ("file2".to_string(), TreeNode::Deleted),
                        (
                            "file42".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1970,
                                ownership: FilesystemOwnership { uid: 55, gid: 60 },
                                permissions: FilesystemPermissions { mode: 0321 },
                                size: 42,
                                content: RecursiveChunkRef {
                                    chunk_id: ChunkId::from([42; 32]),
                                    depth: 29,
                                },
                            },
                        ),
                    ]
                    .into_iter()
                    .collect(),
                },
            )]
            .into_iter()
            .collect(),
        };

        let mut child_full = child_diff.clone();
        integrate_node_in_place(&mut child_full, &parent);

        let expected_child_full = TreeNode::Directory {
            ownership: FilesystemOwnership { uid: 47, gid: 49 },
            permissions: FilesystemPermissions { mode: 0660 },
            children: vec![(
                "dir1".to_string(),
                TreeNode::Directory {
                    ownership: FilesystemOwnership { uid: 46, gid: 50 },
                    permissions: FilesystemPermissions { mode: 0560 },
                    children: vec![
                        (
                            "file1".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1996,
                                ownership: FilesystemOwnership { uid: 54, gid: 59 },
                                permissions: FilesystemPermissions { mode: 0311 },
                                size: 42,
                                content: RecursiveChunkRef {
                                    chunk_id: ChunkId::from([37; 32]),
                                    depth: 2,
                                },
                            },
                        ),
                        (
                            "file42".to_string(),
                            TreeNode::NormalFile {
                                mtime: 1970,
                                ownership: FilesystemOwnership { uid: 55, gid: 60 },
                                permissions: FilesystemPermissions { mode: 0321 },
                                size: 42,
                                content: RecursiveChunkRef {
                                    chunk_id: ChunkId::from([42; 32]),
                                    depth: 29,
                                },
                            },
                        ),
                    ]
                    .into_iter()
                    .collect(),
                },
            )]
            .into_iter()
            .collect(),
        };

        assert_eq!(child_full, expected_child_full);
    }

    pub fn example_file() -> TreeNode {
        TreeNode::NormalFile {
            mtime: 424242,
            ownership: FilesystemOwnership {
                uid: 1042,
                gid: 1043,
            },
            permissions: FilesystemPermissions { mode: 0o760 },
            size: 42,
            content: RecursiveChunkRef {
                chunk_id: ChunkId::from([0u8; 32]),
                depth: 0,
            },
        }
    }

    pub fn example_dir(
        file1: Option<(&str, TreeNode)>,
        file2: Option<(&str, TreeNode)>,
    ) -> TreeNode {
        let mut map = BTreeMap::new();
        if let Some((name, file)) = file1 {
            map.insert(name.to_owned(), file);
        }
        if let Some((name, file)) = file2 {
            map.insert(name.to_owned(), file);
        }
        TreeNode::Directory {
            ownership: FilesystemOwnership {
                uid: 1042,
                gid: 1043,
            },
            permissions: FilesystemPermissions { mode: 0o770 },
            children: map,
        }
    }

    #[test]
    pub fn test_exists() {
        let file = example_file();
        assert!(file.exists(&[]));
        assert!(!file.exists(&["anything"]));

        let subdir = example_dir(Some(("fetchmailrc", example_file())), None);
        let dir = example_dir(Some(("boot.img", example_file())), Some(("etc", subdir)));
        assert!(dir.exists(&[]));
        assert!(dir.exists(&["boot.img"]));
        assert!(dir.exists(&["etc", "fetchmailrc"]));
        assert!(!dir.exists(&["bin"]));
        assert!(!dir.exists(&["etc", "resolv.conf"]));
        assert!(!dir.exists(&["boot.img", "hehe"]));
    }
}
