use eyre::{bail, Context as EyreContext};
use serde::{Deserialize, Serialize};
use sha2::Digest;
use std::fmt::{Debug, Display, Formatter};
use std::io;
use std::io::SeekFrom;
use std::pin::Pin;
use std::task::{Context, Poll};
use tokio::io::{AsyncRead, AsyncSeek, AsyncWrite, ReadBuf};
use yama_midlevel_crypto::sym_stream::SymStreamKey;
use yama_wormfile::WormFileWriter;

pub struct HashedWormWriter<W: WormFileWriter> {
    inner: W,
    hasher: sha2::Sha256,
}

#[derive(Copy, Clone, Serialize, Deserialize, Ord, PartialOrd, Eq, PartialEq, Hash)]
#[serde(transparent)]
pub struct Sha256(pub [u8; 32]);

impl TryFrom<&str> for Sha256 {
    type Error = eyre::Error;

    fn try_from(value: &str) -> Result<Self, Self::Error> {
        if value.len() != 64 {
            bail!("sha256 hexlength not 64 chars");
        }
        let decoded = hex::decode(value).context("failed to decode hex")?;

        if decoded.len() != 32 {
            bail!("wrong number of decoded bytes");
        }

        let mut bytes = [0u8; 32];
        bytes.copy_from_slice(&decoded);
        Ok(Sha256(bytes))
    }
}

impl Debug for Sha256 {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "Sha256({})", hex::encode(&self.0))
    }
}

impl Display for Sha256 {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", hex::encode(&self.0))
    }
}

impl<W: WormFileWriter> HashedWormWriter<W> {
    pub fn new(writer: W) -> Self {
        Self {
            inner: writer,
            hasher: sha2::Sha256::new(),
        }
    }
}

impl<W: WormFileWriter + Unpin> AsyncWrite for HashedWormWriter<W> {
    #[inline]
    fn poll_write(
        mut self: Pin<&mut Self>,
        cx: &mut Context<'_>,
        buf: &[u8],
    ) -> Poll<Result<usize, io::Error>> {
        let result = Pin::new(&mut self.inner).poll_write(cx, buf);
        if let Poll::Ready(Ok(num_bytes_written)) = result {
            // Once a write is complete, add the written bytes to the hasher.
            self.hasher.update(&buf[0..num_bytes_written]);
        }
        result
    }

    #[inline]
    fn poll_flush(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Result<(), io::Error>> {
        Pin::new(&mut self.inner).poll_flush(cx)
    }

    #[inline]
    fn poll_shutdown(
        mut self: Pin<&mut Self>,
        cx: &mut Context<'_>,
    ) -> Poll<Result<(), io::Error>> {
        Pin::new(&mut self.inner).poll_shutdown(cx)
    }
}

impl<W: WormFileWriter> HashedWormWriter<W> {
    /// Finish hashing. Returns the hash and gives back the writer.
    pub fn finalise(self) -> (W, Sha256) {
        let mut output = [0; 32];
        output.copy_from_slice(&self.hasher.finalize()[..]);
        (self.inner, Sha256(output))
    }
}

// TODO We should consider buffering writes so we don't waste encryptions. But that would make it
//      a little more complex, so will save that for later...
pub struct SymStreamWriter<W: AsyncWrite> {
    inner: W,
    offset: u64,
    sym_stream_key: SymStreamKey,
}

impl<W: AsyncWrite> SymStreamWriter<W> {
    pub fn new(inner: W, sym_stream_key: SymStreamKey) -> Self {
        Self {
            inner,
            offset: 0,
            sym_stream_key,
        }
    }

    pub fn finish(self) -> W {
        self.inner
    }

    pub fn offset(&self) -> u64 {
        self.offset
    }
}

impl<W: AsyncWrite + Unpin> AsyncWrite for SymStreamWriter<W> {
    fn poll_write(
        mut self: Pin<&mut Self>,
        cx: &mut Context<'_>,
        buf: &[u8],
    ) -> Poll<Result<usize, io::Error>> {
        let mut enc_buf = buf.to_vec();
        // Safety: Deny use of unencrypted `buf` from here on.
        #[allow(unused)]
        let buf = ();
        let offset = self.offset;
        self.sym_stream_key.apply_xor(offset, &mut enc_buf);

        let result = Pin::new(&mut self.inner).poll_write(cx, &enc_buf);

        if let Poll::Ready(Ok(num_bytes_written)) = result {
            // Once a write is complete, add the offset to our internally tracked offset.
            self.offset += num_bytes_written as u64;
        }

        result
    }

    #[inline]
    fn poll_flush(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Result<(), io::Error>> {
        Pin::new(&mut self.inner).poll_flush(cx)
    }

    #[inline]
    fn poll_shutdown(
        mut self: Pin<&mut Self>,
        cx: &mut Context<'_>,
    ) -> Poll<Result<(), io::Error>> {
        Pin::new(&mut self.inner).poll_shutdown(cx)
    }
}

pub struct SymStreamReader<R: AsyncRead> {
    inner: R,
    offset: u64,
    sym_stream_key: SymStreamKey,
}

impl<R: AsyncRead> SymStreamReader<R> {
    pub fn new(inner: R, sym_stream_key: SymStreamKey) -> Self {
        Self {
            inner,
            offset: 0,
            sym_stream_key,
        }
    }

    pub fn finish(self) -> R {
        self.inner
    }

    pub fn offset(&self) -> u64 {
        self.offset
    }
}

impl<R: AsyncRead + Unpin> AsyncRead for SymStreamReader<R> {
    fn poll_read(
        mut self: Pin<&mut Self>,
        cx: &mut Context<'_>,
        buf: &mut ReadBuf<'_>,
    ) -> Poll<std::io::Result<()>> {
        let old_filled = buf.filled().len();

        let result = Pin::new(&mut self.inner).poll_read(cx, buf);

        if result.is_ready() {
            let filled = buf.filled_mut();
            let new_filled = filled.len();
            let to_decrypt = &mut filled[old_filled..new_filled];

            let offset = self.offset;
            // eprintln!("read {} @ {offset}", to_decrypt.len());
            self.sym_stream_key.apply_xor(offset, to_decrypt);
            self.offset += to_decrypt.len() as u64;
        }

        result
    }
}

impl<R: AsyncRead + AsyncSeek + Unpin> AsyncSeek for SymStreamReader<R> {
    fn start_seek(mut self: Pin<&mut Self>, position: SeekFrom) -> std::io::Result<()> {
        // eprintln!("SS {position:?}");
        Pin::new(&mut self.inner).start_seek(position)
    }

    fn poll_complete(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<std::io::Result<u64>> {
        let result = Pin::new(&mut self.inner).poll_complete(cx);
        if let Poll::Ready(Ok(new_offset)) = result {
            // eprintln!("sought {new_offset}");

            self.offset = new_offset;
        }
        result
    }
}
