use crate::paths::{WormPath, WormPathBuf};
use crate::{WormFileMeta, WormFileProvider, WormFileReader, WormFileWriter};
use async_trait::async_trait;
use std::error::Error;
use std::fmt::{Debug, Display, Formatter};
use std::ops::DerefMut;
use std::pin::Pin;

pub struct BoxErr(Box<dyn Error + Send + Sync>);

impl Debug for BoxErr {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        Debug::fmt(&self.0, f)
    }
}

impl Display for BoxErr {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        Display::fmt(&self.0, f)
    }
}

impl Error for BoxErr {}

impl BoxErr {
    pub fn new(err: impl Error + Sync + Send + 'static) -> Self {
        BoxErr(Box::new(err))
    }
}
#[async_trait]
trait BoxableWormFileProvider: Debug + Send + Sync {
    async fn is_dir_b(&self, path: &WormPath) -> eyre::Result<bool>;
    async fn is_regular_file_b(&self, path: &WormPath) -> eyre::Result<bool>;
    async fn list_b(&self, path: &WormPath) -> eyre::Result<Vec<WormPathBuf>>;
    async fn list_meta_b(&self, path: &WormPath) -> eyre::Result<Vec<(WormPathBuf, WormFileMeta)>>;
    async fn read_b(&self, path: &WormPath) -> eyre::Result<Pin<Box<dyn WormFileReader>>>;
    async fn write_b(&self) -> eyre::Result<Pin<Box<dyn WormFileWriter>>>;
    async fn delete_b(&self, path: &WormPath) -> eyre::Result<()>;
}

#[async_trait]
impl<T: WormFileProvider> BoxableWormFileProvider for T {
    async fn is_dir_b(&self, path: &WormPath) -> eyre::Result<bool> {
        self.is_dir(path).await
    }

    async fn is_regular_file_b(&self, path: &WormPath) -> eyre::Result<bool> {
        self.is_regular_file(path).await
    }

    async fn list_b(&self, path: &WormPath) -> eyre::Result<Vec<WormPathBuf>> {
        self.list(path).await
    }

    async fn list_meta_b(&self, path: &WormPath) -> eyre::Result<Vec<(WormPathBuf, WormFileMeta)>> {
        self.list_meta(path).await
    }

    async fn read_b(&self, path: &WormPath) -> eyre::Result<Pin<Box<dyn WormFileReader>>> {
        self.read(path)
            .await
            .map(|wfr| Box::pin(wfr) as Pin<Box<dyn WormFileReader>>)
    }

    async fn write_b(&self) -> eyre::Result<Pin<Box<dyn WormFileWriter>>> {
        self.write()
            .await
            .map(|wfw| Box::pin(wfw) as Pin<Box<dyn WormFileWriter>>)
    }

    async fn delete_b(&self, path: &WormPath) -> eyre::Result<()> {
        self.delete(path).await
    }
}

#[derive(Debug)]
pub struct BoxedWormFileProvider {
    inner: Box<dyn BoxableWormFileProvider>,
}

impl BoxedWormFileProvider {
    pub fn new(inner: impl WormFileProvider + 'static) -> BoxedWormFileProvider {
        Self {
            inner: Box::new(inner),
        }
    }
}

#[async_trait]
impl WormFileProvider for BoxedWormFileProvider {
    type WormFileReader = Pin<Box<dyn WormFileReader>>;
    type WormFileWriter = Pin<Box<dyn WormFileWriter>>;

    async fn is_dir(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<bool> {
        let path = path.as_ref();
        self.inner.is_dir_b(path).await
    }

    async fn is_regular_file(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<bool> {
        let path = path.as_ref();
        self.inner.is_regular_file_b(path).await
    }

    async fn list(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<Vec<WormPathBuf>> {
        let path = path.as_ref();
        self.inner.list_b(path).await
    }

    async fn list_meta(
        &self,
        path: impl AsRef<WormPath> + Send,
    ) -> eyre::Result<Vec<(WormPathBuf, WormFileMeta)>> {
        let path = path.as_ref();
        self.inner.list_meta_b(path).await
    }

    async fn read(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<Self::WormFileReader> {
        let path = path.as_ref();
        self.inner.read_b(path).await
    }

    async fn write(&self) -> eyre::Result<Self::WormFileWriter> {
        self.inner.write_b().await
    }

    async fn delete(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<()> {
        let path = path.as_ref();
        self.inner.delete_b(path).await
    }
}

#[async_trait]
impl WormFileReader for Pin<Box<dyn WormFileReader>> {}

#[async_trait]
impl WormFileWriter for Pin<Box<dyn WormFileWriter>> {
    async fn finalise(&mut self, target_path: &WormPath, replace: bool) -> std::io::Result<()> {
        self.deref_mut().finalise(target_path, replace).await
    }
}

// pub struct BoxedWormFileWriter(Pin<Box<dyn WormFileWriter>>)
//
// #[async_trait]
// impl WormFileWriter for BoxedWormFileWriter {
//     async fn finalise(self, target_path: &WormPath, replace: bool) -> std::io::Result<()> {
//         self.0.finalise(ztarget_path, replace).await
//     }
// }
