use crate::paths::{WormPath, WormPathBuf};
use async_trait::async_trait;
use std::fmt::Debug;
use std::io;
use tokio::io::{AsyncRead, AsyncSeek, AsyncWrite};

pub mod boxed;
pub mod paths;

#[async_trait]
pub trait WormFileProvider: Debug + Send + Sync {
    type WormFileReader: WormFileReader;
    type WormFileWriter: WormFileWriter;
    // sad because no context:
    //type Error: Error + Send + Sync + 'static;

    /// Tests whether the path is a directory.
    /// Does not fail if the path does not exist, even if the parent path doesn't exist — returns
    /// false in that case.
    /// Beware! Some backends do not support the concept of a directory and will happily report
    /// `true` for any path.
    async fn is_dir(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<bool>;

    /// Tests whether the path is a regular file.
    /// Does not fail if the path does not exist, even if the parent path doesn't exist — returns
    /// false in that case.
    async fn is_regular_file(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<bool>;

    /// Lists all the files and directories in the specified path.
    ///
    /// If the path does not exist, gives an error.
    /// TODO a streaming version of this might be beneficial.
    async fn list(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<Vec<WormPathBuf>>;

    /// Lists all the files and directories in the specified path, with metadata.
    ///
    /// If the path does not exist, gives an error.
    /// TODO a streaming version of this might be beneficial.
    async fn list_meta(
        &self,
        path: impl AsRef<WormPath> + Send,
    ) -> eyre::Result<Vec<(WormPathBuf, WormFileMeta)>>;

    /// Reads a file.
    ///
    /// Fails if the file does not exist or is not a regular file.
    async fn read(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<Self::WormFileReader>;

    /// Writes to a file.
    ///
    /// No path is supplied here because this interface provides atomic writing by first writing
    /// to a temporary file and moving it in place once ready.
    ///
    /// When applicable, the file is first created in the `tmp` directory during writing and then
    /// moved into place afterwards.
    async fn write(&self) -> eyre::Result<Self::WormFileWriter>;

    /// Deletes the given file.
    async fn delete(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<()>;
}

#[derive(Clone, Debug)]
pub struct WormFileMeta {
    pub file_size: u64,
}

pub trait WormFileReader: AsyncRead + AsyncSeek + Debug + Send + Sync + Unpin + 'static {}

#[async_trait]
pub trait WormFileWriter: AsyncWrite + Debug + Send + Sync + Unpin + 'static {
    /// Finish writing the file.
    /// Moves the file atomically to `target_path`.
    /// If `replace` is false, will not overwrite a file. (May be best-effort depending on backend;
    /// intended as a sanity check rather than a flawless safeguard.)
    ///
    /// Must only be called once, but can't consume the WormFileWriter due to object safety...
    async fn finalise(&mut self, target_path: &WormPath, replace: bool) -> io::Result<()>;
}
