use std::borrow::Borrow;
use std::fmt::{Debug, Formatter};

/// Simplified version of `Path` for use in WormFile situations.
/// The Path is guaranteed to remain within the root and does not contain any `.` or `..` elements.
#[repr(transparent)]
pub struct WormPath {
    inner: str,
}

impl Debug for WormPath {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "<{}>", &self.inner)
    }
}

impl WormPath {
    pub fn new(path_str: &str) -> Option<&WormPath> {
        if path_str
            .split('/')
            .any(|component| component.is_empty() || component == "." || component == "..")
        {
            None
        } else {
            Some(unsafe { Self::new_unchecked(path_str) })
        }
    }

    unsafe fn new_unchecked(path_str: &str) -> &WormPath {
        &*(path_str as *const str as *const WormPath)
    }

    pub fn as_str(&self) -> &str {
        &self.inner
    }

    pub fn join(&self, extension: impl AsRef<str>) -> Option<WormPathBuf> {
        let extension = extension.as_ref();

        if extension.starts_with('/')
            || extension
                .split('/')
                .any(|component| component.is_empty() || component == "." || component == "..")
        {
            return None;
        }

        let mut owned = self.inner.to_owned();
        owned.reserve_exact(extension.len() + 1);
        owned.push('/');
        owned.push_str(extension);
        // TODO maybe convert to new_unchecked in the future.
        Some(WormPathBuf::new(owned).expect("already checked upfront"))
    }
}

impl ToOwned for WormPath {
    type Owned = WormPathBuf;

    fn to_owned(&self) -> Self::Owned {
        WormPathBuf {
            inner: self.inner.to_owned(),
        }
    }
}

impl AsRef<WormPath> for WormPath {
    fn as_ref(&self) -> &WormPath {
        self
    }
}

/// Simplified version of `PathBuf` for use in WormFile situations.
/// Owned form of `WormPath`.
#[derive(Clone)]
#[repr(transparent)]
pub struct WormPathBuf {
    inner: String,
}

impl Debug for WormPathBuf {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "<{}>", self.inner)
    }
}

impl WormPathBuf {
    pub fn new(path_string: String) -> Option<WormPathBuf> {
        if path_string
            .split('/')
            .any(|component| component.is_empty() || component == "." || component == "..")
        {
            None
        } else {
            Some(WormPathBuf { inner: path_string })
        }
    }
}

impl AsRef<WormPath> for WormPathBuf {
    fn as_ref(&self) -> &WormPath {
        unsafe { WormPath::new_unchecked(&self.inner) }
    }
}

impl Borrow<WormPath> for WormPathBuf {
    fn borrow(&self) -> &WormPath {
        unsafe { WormPath::new_unchecked(&self.inner) }
    }
}
