use async_trait::async_trait;
use eyre::Context as EyreContext;
use std::fmt::{Debug, Formatter};
use std::io;
use std::io::{ErrorKind, SeekFrom};
use std::os::unix::fs::MetadataExt;
use std::path::PathBuf;
use std::pin::Pin;
use std::task::{Context, Poll};
use tokio::fs::{File, OpenOptions};
use tokio::io::{AsyncRead, AsyncSeek, AsyncWrite, AsyncWriteExt, ReadBuf};
use yama_wormfile::paths::{WormPath, WormPathBuf};
use yama_wormfile::{WormFileMeta, WormFileProvider, WormFileReader, WormFileWriter};

/// WormFileProvider that uses the local filesystem, in a given root directory.
#[derive(Debug)]
pub struct LocalWormFilesystem {
    /// The root directory.
    root_dir: PathBuf,
}

impl LocalWormFilesystem {
    pub fn new(root_dir: impl Into<PathBuf>) -> io::Result<LocalWormFilesystem> {
        let root_dir = root_dir.into();
        if !root_dir.is_dir() {
            return Err(io::Error::new(
                ErrorKind::Other,
                "root for LocalWormFilesystem is not a directory!",
            ));
        }
        Ok(LocalWormFilesystem { root_dir })
    }

    pub fn resolve_real_path(&self, path: impl AsRef<WormPath>) -> PathBuf {
        let relpath = path.as_ref();
        self.root_dir.join(relpath.as_str())
    }
}

#[async_trait]
impl WormFileProvider for LocalWormFilesystem {
    type WormFileReader = FileWormReader;
    type WormFileWriter = FileWormWriter;

    async fn is_dir(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<bool> {
        let path = self.resolve_real_path(path.as_ref());
        Ok(tokio::fs::metadata(path).await?.is_dir())
    }

    async fn is_regular_file(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<bool> {
        let path = self.resolve_real_path(path.as_ref());
        Ok(tokio::fs::metadata(path).await?.is_file())
    }

    async fn list(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<Vec<WormPathBuf>> {
        let worm_path = path.as_ref();
        let real_path = self.resolve_real_path(worm_path);
        let mut dir_reader = match tokio::fs::read_dir(real_path).await {
            Ok(ok) => ok,
            Err(e) if e.kind() == ErrorKind::NotFound => {
                return Ok(Vec::new());
            }
            Err(other) => return Err(other.into()),
        };
        let mut out = Vec::new();
        while let Some(next_ent) = dir_reader.next_entry().await? {
            if let Some(name_str) = next_ent.file_name().to_str() {
                out.push(worm_path.join(name_str).unwrap());
            }
        }
        Ok(out)
    }

    async fn list_meta(
        &self,
        path: impl AsRef<WormPath> + Send,
    ) -> eyre::Result<Vec<(WormPathBuf, WormFileMeta)>> {
        let worm_path = path.as_ref();
        let real_path = self.resolve_real_path(worm_path);
        let mut dir_reader = match tokio::fs::read_dir(real_path).await {
            Ok(ok) => ok,
            Err(e) if e.kind() == ErrorKind::NotFound => {
                return Ok(Vec::new());
            }
            Err(other) => return Err(other.into()),
        };
        let mut out = Vec::new();
        while let Some(next_ent) = dir_reader.next_entry().await? {
            if let Some(name_str) = next_ent.file_name().to_str() {
                let metadata = next_ent.metadata().await?;
                out.push((
                    worm_path.join(name_str).unwrap(),
                    WormFileMeta {
                        file_size: metadata.size(),
                    },
                ));
            }
        }
        Ok(out)
    }

    async fn read(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<Self::WormFileReader> {
        let worm_path = path.as_ref();
        let real_path = self.resolve_real_path(worm_path);
        let file = OpenOptions::new().read(true).open(&real_path).await?;
        Ok(FileWormReader {
            path: real_path,
            file,
        })
    }

    async fn write(&self) -> eyre::Result<Self::WormFileWriter> {
        let tmp_dir = self.root_dir.join("tmp");
        if !tokio::fs::try_exists(&tmp_dir)
            .await
            .context("can't check <tmp> exists")?
        {
            tokio::fs::create_dir(&tmp_dir)
                .await
                .context("can't create <tmp>")?;
        }

        let (tmp_path, file) = loop {
            let rand_num: u32 = rand::random();
            let pid = std::process::id();

            let try_fn = format!("pid{pid}-{rand_num:08X}.writing");
            let try_path = tmp_dir.join(try_fn);
            match OpenOptions::new()
                .write(true)
                .create_new(true)
                .open(&try_path)
                .await
            {
                Ok(file) => break (try_path, file),
                Err(err) => {
                    if err.kind() == ErrorKind::AlreadyExists {
                        continue;
                    } else {
                        return Err(err).context("can't create_new in <tmp>");
                    }
                }
            }
        };

        Ok(FileWormWriter {
            temp_path: tmp_path,
            file,
            root_dir: self.root_dir.clone(),
        })
    }

    async fn delete(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<()> {
        let worm_path = path.as_ref();
        let real_path = self.resolve_real_path(worm_path);
        tokio::fs::remove_file(&real_path)
            .await
            .with_context(|| format!("failed to remove_file({real_path:?})"))?;
        Ok(())
    }
}

pub struct FileWormReader {
    path: PathBuf,
    file: File,
}

impl Debug for FileWormReader {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "FileWormReader({:?})", self.path)
    }
}

impl AsyncRead for FileWormReader {
    fn poll_read(
        mut self: Pin<&mut Self>,
        cx: &mut Context<'_>,
        buf: &mut ReadBuf<'_>,
    ) -> Poll<io::Result<()>> {
        Pin::new(&mut self.file).poll_read(cx, buf)
    }
}

impl AsyncSeek for FileWormReader {
    fn start_seek(mut self: Pin<&mut Self>, position: SeekFrom) -> io::Result<()> {
        Pin::new(&mut self.file).start_seek(position)
    }

    fn poll_complete(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<io::Result<u64>> {
        Pin::new(&mut self.file).poll_complete(cx)
    }
}

impl WormFileReader for FileWormReader {}

pub struct FileWormWriter {
    temp_path: PathBuf,
    file: File,
    root_dir: PathBuf,
}

impl Debug for FileWormWriter {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "FileWormWriter({:?})", self.temp_path)
    }
}

impl AsyncWrite for FileWormWriter {
    fn poll_write(
        mut self: Pin<&mut Self>,
        cx: &mut Context<'_>,
        buf: &[u8],
    ) -> Poll<Result<usize, io::Error>> {
        Pin::new(&mut self.file).poll_write(cx, buf)
    }

    fn poll_flush(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Result<(), io::Error>> {
        Pin::new(&mut self.file).poll_flush(cx)
    }

    fn poll_shutdown(
        mut self: Pin<&mut Self>,
        cx: &mut Context<'_>,
    ) -> Poll<Result<(), io::Error>> {
        Pin::new(&mut self.file).poll_shutdown(cx)
    }
}

#[async_trait]
impl WormFileWriter for FileWormWriter {
    async fn finalise(&mut self, target_path: &WormPath, replace: bool) -> io::Result<()> {
        self.flush().await?;

        let FileWormWriter {
            root_dir,
            temp_path,
            ..
        } = self;

        let worm_path = target_path;
        let real_path = root_dir.join(worm_path.as_str());

        // Directories will be created as needed.
        let parent = real_path.parent().unwrap();
        tokio::fs::create_dir_all(parent).await?;

        // Avoid allowing a replacement if not intended.
        // But this is currently not atomic, so it's just a sanity check rather than a foolproof
        // safeguard!
        if !replace && tokio::fs::try_exists(&real_path).await? {
            return Err(io::Error::new(
                ErrorKind::AlreadyExists,
                "finalise()ing a writer: dest already exists and replace = false",
            ));
        }

        // Perform the move, atomically.
        tokio::fs::rename(&temp_path, &real_path).await?;
        Ok(())
    }
}
