extern crate core;

use async_trait::async_trait;
use eyre::{bail, Context as EyreContext, ContextCompat};
use openssh::{KnownHosts, RemoteChild, Session, SessionBuilder, Stdio};
use openssh_sftp_client::error::SftpErrorKind;
use openssh_sftp_client::file::{File, TokioCompatFile};
use openssh_sftp_client::fs::{DirEntry, Fs};
use openssh_sftp_client::Error::SftpError;
use openssh_sftp_client::{Sftp, SftpOptions};
use ouroboros::self_referencing;
use std::fmt::{Debug, Formatter};
use std::io;
use std::io::{ErrorKind, SeekFrom};
use std::num::NonZeroUsize;
use std::path::{Path, PathBuf};
use std::pin::Pin;
use std::sync::Arc;
use std::task::{Context, Poll};
use std::time::{Duration, Instant};
use tokio::io::{AsyncRead, AsyncSeek, AsyncWrite, AsyncWriteExt, ReadBuf};
use tokio_stream::StreamExt;
use tracing::debug;
use yama_wormfile::paths::{WormPath, WormPathBuf};
use yama_wormfile::{WormFileMeta, WormFileProvider, WormFileReader, WormFileWriter};

/// WormFileProvider that uses an SFTP connection, in a given root directory.
#[derive(Debug)]
pub struct SftpWormFilesystem {
    conn: Arc<SftpConn>,

    /// The root directory.
    root_dir: PathBuf,
}

#[self_referencing]
struct SftpConn {
    /// The SSH session.
    ssh: Session,

    /// The SSH child process
    #[borrows(ssh)]
    #[covariant]
    ssh_child: RemoteChild<'this>,

    /// The SFTP client.
    #[borrows(mut ssh_child)]
    sftp: Sftp,

    root_dir: PathBuf,
    // #[borrows(sftp)]
    // #[covariant]
    // fs: Fs<'this>,
}

impl SftpConn {
    pub async fn create(ssh_connect: &str, root_dir: impl Into<PathBuf>) -> eyre::Result<Self> {
        let root_dir = root_dir.into();
        let mut session_builder = SessionBuilder::default();
        if let Ok(ssh_key_path) = std::env::var("YAMA_SSH_KEY") {
            session_builder.keyfile(&ssh_key_path);
        }
        session_builder.known_hosts_check(KnownHosts::Add);
        let session = session_builder.connect(ssh_connect).await?;

        let res = SftpConnAsyncTryBuilder {
            ssh: session,
            ssh_child_builder: |ssh| {
                Box::pin(async move {
                    ssh.subsystem("sftp")
                        .stdin(Stdio::piped())
                        .stdout(Stdio::piped())
                        .spawn()
                        .await
                        .map_err(|e| eyre::Error::from(e))
                })
            },
            sftp_builder: |ssh_child| {
                Box::pin(async move {
                    let sftp_opts = SftpOptions::new()
                        // Don't buffer infinitely when writing files: 32 MiB per file in flight
                        // will do.
                        .tokio_compat_file_write_limit(
                            NonZeroUsize::new(32 * 1024 * 1024).unwrap(),
                        );
                    Sftp::new(
                        ssh_child.stdin().take().unwrap(),
                        ssh_child.stdout().take().unwrap(),
                        sftp_opts,
                    )
                    .await
                    .map_err(|e| eyre::Error::from(e))
                })
            },
            // fs_builder: |sftp| Box::pin(async move {
            //     let mut fs = sftp.fs();
            //     fs.set_cwd(&root_dir);
            //     Ok(fs)
            // })
            root_dir,
        }
        .try_build()
        .await?;
        Ok(res)
    }

    pub fn get_fs(&self) -> Fs {
        let mut fs = self.borrow_sftp().fs();
        fs.set_cwd(&self.borrow_root_dir());
        fs
    }

    async fn create_dir_all(&self, worm_path_as_pathbuf: PathBuf) -> eyre::Result<()> {
        // Try twice to try and shake out race conditions if another worker is doing this exactly
        // now...
        if self
            .create_dir_all_impl_once(worm_path_as_pathbuf.clone())
            .await
            .is_ok()
        {
            return Ok(());
        }
        tokio::time::sleep(Duration::from_secs(3)).await;
        self.create_dir_all_impl_once(worm_path_as_pathbuf).await
    }

    async fn create_dir_all_impl_once(&self, worm_path_as_pathbuf: PathBuf) -> eyre::Result<()> {
        let mut fs = self.get_fs();
        let mut stack = vec![];

        let mut at = Some(worm_path_as_pathbuf);
        while let Some(at_path) = at {
            match fs.metadata(&at_path).await {
                Ok(_) => {
                    break;
                }
                Err(SftpError(SftpErrorKind::NoSuchFile, _)) => {
                    at = at_path.parent().map(Path::to_owned);
                    stack.push(at_path);
                }
                Err(sftp_err) => {
                    return Err(sftp_err).context("other sftp err when create_dir_all (main loop)");
                }
            }
        }

        while let Some(path) = stack.pop() {
            fs.create_dir(path)
                .await
                .context("failed to create dir in create_dir_all")?;
        }

        Ok(())
    }
}

impl Debug for SftpConn {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "<SftpConn>")
    }
}

impl SftpWormFilesystem {
    pub async fn new(
        ssh_connect: &str,
        root_dir: impl Into<PathBuf>,
    ) -> eyre::Result<SftpWormFilesystem> {
        let root_dir = root_dir.into();

        let conn = Arc::new(SftpConn::create(ssh_connect, &root_dir).await?);

        if !conn
            .get_fs()
            .metadata(".")
            .await?
            .file_type()
            .unwrap()
            .is_dir()
        {
            bail!("{root_dir:?} is not a dir on SFTP remote.");
        }

        Ok(SftpWormFilesystem { conn, root_dir })
    }

    fn get_fs(&self) -> Fs {
        let mut fs = self.conn.borrow_sftp().fs();
        fs.set_cwd(&self.root_dir);
        fs
    }
}

#[async_trait]
impl WormFileProvider for SftpWormFilesystem {
    type WormFileReader = SftpWormReader;
    type WormFileWriter = SftpWormWriter;

    #[async_backtrace::framed]
    async fn is_dir(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<bool> {
        let path = path.as_ref().as_str();
        let mut fs = self.get_fs();
        match fs.metadata(path).await {
            Ok(meta) => Ok(meta.file_type().unwrap().is_dir()),
            Err(SftpError(SftpErrorKind::NoSuchFile, _)) => Ok(false),
            Err(sftp_err) => Err(sftp_err.into()),
        }
    }

    #[async_backtrace::framed]
    async fn is_regular_file(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<bool> {
        let path = path.as_ref().as_str();
        let mut fs = self.get_fs();
        match fs.metadata(path).await {
            Ok(meta) => Ok(meta.file_type().unwrap().is_file()),
            Err(SftpError(SftpErrorKind::NoSuchFile, _)) => Ok(false),
            Err(sftp_err) => Err(sftp_err.into()),
        }
    }

    #[async_backtrace::framed]
    async fn list(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<Vec<WormPathBuf>> {
        let worm_path = path.as_ref();
        let path = worm_path.as_str();
        let mut fs = self.get_fs();

        let remote_dir = match fs.open_dir(path).await {
            Ok(ok) => ok,
            Err(openssh_sftp_client::Error::SftpError(SftpErrorKind::NoSuchFile, _msg)) => {
                return Ok(Vec::new());
            }
            Err(other) => {
                return Err(other.into());
            }
        };
        let dir_reader: Vec<DirEntry> = remote_dir
            .read_dir()
            .collect::<Result<Vec<DirEntry>, _>>()
            .await?;

        Ok(dir_reader
            .iter()
            .filter_map(|entry| {
                if let Some(name_str) = entry.filename().as_os_str().to_str() {
                    if name_str.is_empty() || name_str == "." || name_str == ".." {
                        None
                    } else {
                        Some(worm_path.join(name_str).expect("pre-checked"))
                    }
                } else {
                    None
                }
            })
            .collect())
    }

    #[async_backtrace::framed]
    async fn list_meta(
        &self,
        path: impl AsRef<WormPath> + Send,
    ) -> eyre::Result<Vec<(WormPathBuf, WormFileMeta)>> {
        let worm_path = path.as_ref();
        let path = worm_path.as_str();
        let mut fs = self.get_fs();

        let remote_dir = match fs.open_dir(path).await {
            Ok(ok) => ok,
            Err(openssh_sftp_client::Error::SftpError(SftpErrorKind::NoSuchFile, _msg)) => {
                return Ok(Vec::new());
            }
            Err(other) => {
                return Err(other.into());
            }
        };
        let dir_reader: Vec<DirEntry> = remote_dir
            .read_dir()
            .collect::<Result<Vec<DirEntry>, _>>()
            .await?;

        Ok(dir_reader
            .iter()
            .filter_map(|entry| {
                if let Some(name_str) = entry.filename().as_os_str().to_str() {
                    if name_str.is_empty() || name_str == "." || name_str == ".." {
                        None
                    } else {
                        Some((
                            worm_path.join(name_str).expect("pre-checked"),
                            WormFileMeta {
                                file_size: entry.metadata().len().expect("no size on SFTP file?"),
                            },
                        ))
                    }
                } else {
                    None
                }
            })
            .collect())
    }

    #[async_backtrace::framed]
    async fn read(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<Self::WormFileReader> {
        let real_path = self.root_dir.join(path.as_ref().as_str());

        let real_path2 = real_path.clone();

        let mut file: File = self.conn.borrow_sftp().open(real_path).await?;
        let file_length = file
            .metadata()
            .await?
            .len()
            .context("no len in SFTP file metadata!")?;
        let file = Some(Box::pin(TokioCompatFile::new(file)));

        Ok(SftpWormReader {
            path: real_path2,
            file,
            length: file_length,
        })
    }

    #[async_backtrace::framed]
    async fn write(&self) -> eyre::Result<Self::WormFileWriter> {
        if !self
            .is_dir(WormPath::new("tmp").unwrap())
            .await
            .context("can't check <tmp> exists")?
        {
            self.get_fs()
                .create_dir(Path::new("tmp"))
                .await
                .context("failed to make <tmp>")?;
        }
        loop {
            let rand_num: u32 = rand::random();
            let pid = std::process::id();

            let try_path = format!("tmp/pid{pid}-{rand_num:08X}.writing");

            let real_path = self.root_dir.join(try_path.as_str());
            // the `Send` in the below line is very important...

            if self
                .is_regular_file(WormPath::new(&try_path).unwrap())
                .await
                .context("can't check that try_path doesn't exist")?
            {
                continue;
            }

            let file = Some(Box::pin(TokioCompatFile::new(
                self.conn.borrow_sftp().create(real_path).await?,
            )));

            break Ok(SftpWormWriter {
                temp_path: WormPathBuf::new(try_path).unwrap(),
                file,
                conn: self.conn.clone(),
                unflushed_written_bytes: 0,
            });
        }
    }

    #[async_backtrace::framed]
    async fn delete(&self, path: impl AsRef<WormPath> + Send) -> eyre::Result<()> {
        let worm_path = path.as_ref();
        let path = worm_path.as_str();
        let mut fs = self.get_fs();
        fs.remove_file(path).await?;
        Ok(())
    }
}

pub struct SftpWormReader {
    path: PathBuf,
    file: Option<Pin<Box<TokioCompatFile>>>,
    length: u64,
}

impl Debug for SftpWormReader {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "FileWormReader({:?})", self.path)
    }
}

impl AsyncRead for SftpWormReader {
    fn poll_read(
        mut self: Pin<&mut Self>,
        cx: &mut Context<'_>,
        buf: &mut ReadBuf<'_>,
    ) -> Poll<io::Result<()>> {
        Pin::new(self.file.as_mut().unwrap()).poll_read(cx, buf)
    }
}

impl AsyncSeek for SftpWormReader {
    fn start_seek(mut self: Pin<&mut Self>, mut position: SeekFrom) -> io::Result<()> {
        if let SeekFrom::End(pos) = position {
            // SFTP doesn't support seek from the end, so implement it manually...
            position = SeekFrom::Start((self.length as i64 + pos).try_into().map_err(|_| {
                io::Error::new(ErrorKind::InvalidInput, "SeekFrom::End out of bounds")
            })?);
        }
        Pin::new(self.file.as_mut().unwrap()).start_seek(position)
    }

    fn poll_complete(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<io::Result<u64>> {
        Pin::new(self.file.as_mut().unwrap()).poll_complete(cx)
    }
}

impl WormFileReader for SftpWormReader {}

pub struct SftpWormWriter {
    temp_path: WormPathBuf,
    file: Option<Pin<Box<TokioCompatFile>>>,
    conn: Arc<SftpConn>,
    unflushed_written_bytes: u64,
}

impl Debug for SftpWormWriter {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "FileWormWriter({:?})", self.temp_path)
    }
}

/// Maximum number of unflushed bytes to have written, before we force-schedule a flush.
const MAX_UNFLUSHED_WRITTEN_BYTES: u64 = 32 * 1024 * 1024;

impl AsyncWrite for SftpWormWriter {
    fn poll_write(
        mut self: Pin<&mut Self>,
        cx: &mut Context<'_>,
        buf: &[u8],
    ) -> Poll<Result<usize, io::Error>> {
        // Workaround for openssh-sftp-client bug that means that writes are infinitely buffered:
        // only allow a certain number of bytes to be written before flushing.
        if self.unflushed_written_bytes > MAX_UNFLUSHED_WRITTEN_BYTES {
            // Flush and only continue if a flush was a no-op/immediately performed.
            //ready!(self.as_mut().poll_flush(cx))?;
        }
        let r = Pin::new(self.file.as_mut().unwrap()).poll_write(cx, buf);
        if r.is_ready() {
            self.unflushed_written_bytes += buf.len() as u64;
        }
        r
    }

    fn poll_flush(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Result<(), io::Error>> {
        let r = Pin::new(self.file.as_mut().unwrap()).poll_flush(cx);
        if r.is_ready() {
            self.unflushed_written_bytes = 0;
        }
        r
    }

    fn poll_shutdown(
        mut self: Pin<&mut Self>,
        cx: &mut Context<'_>,
    ) -> Poll<Result<(), io::Error>> {
        Pin::new(self.file.as_mut().unwrap()).poll_shutdown(cx)
    }
}

#[async_trait]
impl WormFileWriter for SftpWormWriter {
    #[async_backtrace::framed]
    async fn finalise(&mut self, target_path: &WormPath, replace: bool) -> io::Result<()> {
        let flush_start = Instant::now();
        self.flush().await?;

        let mkdirs_start = Instant::now();
        let SftpWormWriter {
            temp_path,
            file,
            conn,
            ..
        } = self;

        let file = file.take().unwrap();
        let file: TokioCompatFile = TokioCompatFile::clone(&file);
        // This looks wrong but might be OK. Hack mostly needed due to the very awkward Pinning.
        file.into_inner()
            .close()
            .await
            .map_err(|e| io::Error::new(ErrorKind::Other, e))?;

        let mut fs = conn.get_fs();

        // Directories will be created as needed.
        if let Some(parent) = PathBuf::from(target_path.as_str()).parent() {
            conn.create_dir_all(parent.to_owned())
                .await
                .map_err(|e| io::Error::new(ErrorKind::Other, e.to_string()))?;
        }
        // Avoid allowing a replacement if not intended.
        // But this is currently not atomic, so it's just a sanity check rather than a foolproof
        // safeguard!
        if !replace {
            match fs.metadata(target_path.as_str()).await {
                Ok(_) => {
                    return Err(io::Error::new(
                        ErrorKind::AlreadyExists,
                        "finalise()ing a writer: dest already exists and replace = false",
                    ));
                }
                Err(SftpError(SftpErrorKind::NoSuchFile, _)) => {
                    // ideal. nop.
                }
                Err(sftp_err) => {
                    return Err(io::Error::new(ErrorKind::Other, sftp_err.to_string()));
                }
            }
        }

        let move_start = Instant::now();
        // Perform the move.
        fs.rename(&temp_path.as_ref().as_str(), target_path.as_str())
            .await
            .map_err(|e| io::Error::new(ErrorKind::Other, e))?;

        let end = Instant::now();
        debug!(
            "Saved SFTP file to {target_path:?} (flush={:?}, mkdirs={:?}, move={:?})",
            mkdirs_start - flush_start,
            move_start - mkdirs_start,
            end - move_start
        );
        Ok(())
    }
}
