use crate::asym_keyx::Exchanged;
use crate::byte_layer::ByteLayer;
use crate::key_derivation::DerivedKey;
use chacha20::cipher::{KeyIvInit, StreamCipher};
use chacha20::XChaCha20;
use poly1305::universal_hash::KeyInit;
use poly1305::Poly1305;
use rand::Rng;
use serde::{Deserialize, Serialize};
use std::borrow::Cow;
use std::marker::PhantomData;

pub const SYMBOX_NONCE_LENGTH: usize = 24;
pub const SYMBOX_MAC_LENGTH: usize = 16;
pub const SYMBOX_FOOTER_LENGTH: usize = SYMBOX_MAC_LENGTH + SYMBOX_NONCE_LENGTH;

#[derive(Clone, Serialize, Deserialize)]
#[serde(transparent)]
pub struct SymBox<'bytes, T> {
    // payload || nonce || mac
    bytes: Cow<'bytes, [u8]>,

    #[serde(skip)]
    phantom: PhantomData<T>,
}

#[derive(Clone)]
pub struct SymKey {
    xchacha20: [u8; 32],
    poly1305: poly1305::Key,
}

impl SymKey {
    fn from_64_bytes(input: &[u8]) -> Self {
        let (xchacha20_bytes, poly1305_bytes) = input.split_at(32);

        SymKey {
            xchacha20: xchacha20_bytes.try_into().unwrap(),
            poly1305: *poly1305::Key::from_slice(poly1305_bytes),
        }
    }
}

impl From<Exchanged> for SymKey {
    fn from(exchanged: Exchanged) -> Self {
        SymKey::from_64_bytes(&exchanged.0)
    }
}

impl From<DerivedKey> for SymKey {
    fn from(derived: DerivedKey) -> Self {
        SymKey::from_64_bytes(&derived.0)
    }
}

impl<'bytes, T> SymBox<'bytes, T> {
    pub fn as_bytes(&self) -> &[u8] {
        self.bytes.as_ref()
    }

    pub fn new_from_raw(bytes: &'bytes [u8]) -> Self {
        Self {
            bytes: Cow::Borrowed(bytes),
            phantom: Default::default(),
        }
    }
}

impl<'bytes, T> ByteLayer for SymBox<'bytes, T> {
    fn from_byte_vec(bytes: Vec<u8>) -> Self {
        Self {
            bytes: Cow::Owned(bytes),
            phantom: Default::default(),
        }
    }

    fn into_byte_vec(self) -> Vec<u8> {
        self.bytes.into_owned()
    }
}

impl<T> SymBox<'static, T> {
    pub fn into_vec(self) -> Vec<u8> {
        match self.bytes {
            Cow::Borrowed(b) => b.to_vec(),
            Cow::Owned(o) => o,
        }
    }
}

impl<'bytes, T: ByteLayer> SymBox<'bytes, T> {
    // TODO error
    pub fn unlock(&self, symkey: &SymKey) -> Option<T> {
        let blen = self.bytes.len();

        if blen < SYMBOX_FOOTER_LENGTH {
            return None;
        }

        let (ciphertext_then_nonce, mac) = self.bytes.split_at(blen - SYMBOX_MAC_LENGTH);
        let (ciphertext, nonce) = ciphertext_then_nonce.split_at(blen - SYMBOX_FOOTER_LENGTH);

        let poly1305_mac = Poly1305::new(&symkey.poly1305).compute_unpadded(&ciphertext_then_nonce);
        if poly1305_mac.as_slice() != mac {
            // TODO Should this pedantically be a constant-time equality check?
            //      I don't think it matters in any exploitable way for Yama though...
            return None;
        }

        let mut out_buf = Vec::<u8>::with_capacity(ciphertext.len());
        out_buf.extend_from_slice(&ciphertext);
        let mut xchacha20: XChaCha20 = XChaCha20::new(&symkey.xchacha20.into(), nonce.into());
        xchacha20.apply_keystream(&mut out_buf);
        let decrypted: Vec<u8> = out_buf;

        Some(T::from_byte_vec(decrypted))
    }
}

impl<'bytes, T: ByteLayer> SymBox<'bytes, T> {
    // TODO error
    pub fn new(contents: T, symkey: &SymKey) -> Option<Self> {
        let unencrypted = contents.into_byte_vec();
        let mut rng = rand::thread_rng();
        let nonce = rng.gen::<[u8; SYMBOX_NONCE_LENGTH]>();

        let mut out_buf = Vec::<u8>::with_capacity(unencrypted.len() + SYMBOX_FOOTER_LENGTH);
        out_buf.extend_from_slice(&unencrypted);
        let mut xchacha20: XChaCha20 = XChaCha20::new(&symkey.xchacha20.into(), &nonce.into());
        xchacha20.apply_keystream(&mut out_buf);

        out_buf.extend_from_slice(&nonce);

        let poly1305_mac = Poly1305::new(&symkey.poly1305).compute_unpadded(&out_buf);
        out_buf.extend_from_slice(poly1305_mac.as_slice());

        Some(SymBox {
            bytes: Cow::Owned(out_buf),
            phantom: Default::default(),
        })
    }
}
